package sd;

import java.awt.BorderLayout;
import java.awt.Frame;
import java.awt.Rectangle;
import java.awt.event.WindowAdapter;
import java.awt.event.WindowEvent;

import Jcg.geometry.Point_2;
import Jcg.geometry.Point_3;
import Jcg.mesh.MeshLoader;
import Jcg.polyhedron.Polyhedron_3;
import Jcg.polyhedron.Vertex;

import jdg.graph.AdjacencyListGraph;
import jdg.io.GraphReader;
import jdg.io.GraphReader_Edges;
import jdg.io.GraphReader_MTX;
import jdg.io.GraphReader_OFF;

import processing.core.PApplet;
import sd.layout.Alexa;
import sd.layout.FR91Layout;
import sd.layout.FastFR91Layout;
import sd.layout.IterativeRelaxation;
import sd.layout.Layout3D;
import sd.layout.SphericalFR91;
import sd.layout.SphericalLayout;
import sd.util.MeshAlgorithms;
import sd.viewer.ArcBall;
import sd.viewer.ConfigSphericalDrawing;
import sd.viewer.MeshRepresentation;
import sd.viewer.SphericalLayoutControlBar;
import sd.viewer.SurfaceMeshRendering;

/**
 * This class implements the computation of the spherical layout for a planar triangle mesh
 * 
 * @author Luca Castelli Aleardi, Ecole Polytechnique
 * @version august 2017
 */
public class SphericalLayoutComputation extends PApplet {

	int renderType=0; // choice of type of rendering
	int renderModes=3; // number of rendering modes

	public SurfaceMeshRendering meshRendering; // 3d rendering for surface meshes and spherical layouts
	public SphericalLayoutControlBar controlBar; // control bar for reading input parameters and dealing with button actions

	/** store all information concerning the input graph and input mesh */
	static public MeshRepresentation input;

	/** an array storing all Layout methods for the spherical case */
	public SphericalLayout[] sphericalLayouts;
	/** an array storing all Layout methods for the euclidean case */
	public Layout3D[] euclideanLayouts;

	public SphericalLayout selectedSphericalLayout=null; // choice of the Layout algorithm to run
	public Layout3D selectedEuclideanLayout=null; // choice of the Layout algorithm to run

	public void setup() {
		size(1200,800,P3D);
		//size(1600,1000,P3D);
		
		this.meshRendering=new SurfaceMeshRendering(this, input);

		if(this.input.mesh!=null) { // the input file is in OFF format
			MeshAlgorithms.resetMeshIndices(input.mesh);
		}
		else if(this.input.graph!=null) { // the input file is a graph in MTX format (without combinatorial embedding)
			//Layout3D.setRandomPoints3D(this.input.graph, 1., 1., 0.); // place vertices at random in a 3d cube
			SphericalLayout.setRandomPointsOnSphere(this.input.graph);
		}

		//this.setLayout(new BorderLayout());
		this.setLayout(null);
		this.controlBar=new SphericalLayoutControlBar(this, 280, this.height);
		this.controlBar.setBounds(0, 0, 280, this.height);
		this.add("West", controlBar);

		ArcBall arcball = new ArcBall(this); // for interaction with mouse events and 3D rendering
		//this.ortho(-width/2, width/2, -height/2, height/2);
		
		this.meshRendering.showSphericalLayout(); // by default, draw the spherical layout
		this.meshRendering.updateScaleFactor(); // compute the scale factor for the original 3d mesh
		this.initLayouts();
		
		this.controlBar.panelSphericalLayout.showCurrentSpringEmbedderParameters();
		
	}
	
	/** For closing the processing frame */	
	public void windowClosing(WindowEvent e) {
	    System.exit(0);
	}

	/**
	 * Initialize the parameters of all layout algorithms
	 */	
	public void initLayouts() {
		this.sphericalLayouts=new SphericalLayout[3];
		this.euclideanLayouts=new Layout3D[1]; // FR91
		this.meshRendering.showSphericalLayout();

		// fixed parameters
		double coordPrecision=0.0000000000000001; // numeric precision of coordinates computations

		//double threshold=this.input.threshold; // read from input config file

		// read input parameters from user interface
		//double tolerance = 0.0000000000001;
		double tolerance=Double.parseDouble(this.controlBar.panelSphericalLayout.fieldNumericTolerance.getText()); // numeric tolerance for convergence
		double coarseningThreshold=Double.parseDouble(this.controlBar.panelSphericalLayout.fieldCoarseningThreshold.getText()); // for multi-level layouts
		double C=Double.parseDouble(this.controlBar.panelSphericalLayout.fieldRepulsiveStrenght.getText()); // relative strength of repulsive forces
		double energyThreshold=tolerance;
		double theta=1.2;
		double minT=Double.parseDouble(this.controlBar.panelSphericalLayout.fieldMinimalTemperature.getText()); // minimal temperature

		// initialize spherical layouts
		this.sphericalLayouts[0]=new IterativeRelaxation(this.input.graph, this.input.mesh, energyThreshold, tolerance); // Gauss-Seidel iterative relaxation
		this.sphericalLayouts[1]=new Alexa(this.input.graph, this.input.mesh, energyThreshold,tolerance); // Alexa method
		this.sphericalLayouts[2]=new SphericalFR91(this.input.graph, this.input.mesh, energyThreshold, coordPrecision); // force-directed method (Fruchterman Reingold)
		/*this.sphericalLayouts[2]=new GotsmanSheffer2005bis(this.input.graph,this.input.mesh,0.1, tolerance,0.000001); // Non-linear optimization method
		this.sphericalLayouts[4]=new SphericalFR91Approx(this.input.graph, this.input.mesh, theta, energyThreshold, coordPrecision); // FR91 with octree (theta=1.2)
		this.sphericalLayouts[5]=new SphericalWalshaw(this.input.graph, coarseningThreshold, C);*/

		SphericalFR91 sFR91=(SphericalFR91)this.sphericalLayouts[2];
		sFR91.setC(C);
		
		//SphericalFR91 sfastFR91=(SphericalFR91)this.sphericalLayouts[4];
		//SphericalWalshaw sphericalWalshaw=(SphericalWalshaw)this.sphericalLayouts[5];
		
		// set input parameters
		//sfastFR91.setC(C);

		// initialize euclidean force-directed layouts (FR91)
		this.euclideanLayouts[0]=new FR91Layout(this.input.graph, 2., 2., 2.); // force-directed method (Fruchterman Reingold 91)

		// initialize euclidean multi-leve force-directed layouts (Walshaw03)
		//this.euclideanLayouts[1]=new WalshawLayout(this.input.graph, coarseningThreshold, C, 2., 2., 2.); // force-directed method (Walshaw 03)

		// choose by default the first iterative method (Gauss-Seidel relaxation)
		this.selectedSphericalLayout=this.sphericalLayouts[0];
		this.selectedEuclideanLayout=this.euclideanLayouts[0];
	}

	/**
	 * Select a layout algorithm
	 * 
	 * @param LayoutName  name of the layout algorithm
	 */	
	public void updateSelectedLayoutMethod(String layoutName) {
		if(layoutName.equals("Gauss-Seidel relaxation")) {
			this.selectedSphericalLayout=this.sphericalLayouts[0];
			this.meshRendering.showSphericalLayout();
		}
		else if(layoutName.equals("Alexa's method")) {
			this.selectedSphericalLayout=this.sphericalLayouts[1];
			this.meshRendering.showSphericalLayout();		
		}
		else if(layoutName.equals("Spherical FR91")) {
			this.selectedSphericalLayout=this.sphericalLayouts[2];
			this.meshRendering.showSphericalLayout();		
		}
		/*else if(layoutName.equals("Non-linear optimization")) {
			this.selectedSphericalLayout=this.sphericalLayouts[3];
			this.meshRendering.showSphericalLayout();
			this.selectedEuclideanLayout=null;			
		}
		else if(layoutName.equals("Approx spherical FR91 (octree)")) {
			this.selectedSphericalLayout=this.sphericalLayouts[4];
			this.meshRendering.showSphericalLayout();
			this.selectedEuclideanLayout=null;			
		}*/
		/*else if(layoutName.equals("Approx spherical FR91 (octree)")) {
			this.selectedSphericalLayout=this.sphericalLayouts[4];
			this.meshRendering.showSphericalLayout();
			this.selectedEuclideanLayout=null;			
		}*/
		else if(layoutName.equals("Standard euclidean FR91")) {
			this.meshRendering.showEuclidean3DLayout();
			this.selectedEuclideanLayout=this.euclideanLayouts[0];			
		}
		/*else if(layoutName.equals("Multi-level (Walshaw)")) {
			this.selectedSphericalLayout=null;
			this.meshRendering.showEuclidean3DLayout();
			this.selectedEuclideanLayout=this.euclideanLayouts[1];
		}*/
		else {
			System.err.println("Warning: the selected layout is not supported");
			return;
		}

		if(this.selectedSphericalLayout!=null)
			System.out.println("Selected (spherical) layout: "+this.selectedSphericalLayout.getName());
		if(this.selectedEuclideanLayout!=null)
			System.out.println("Selected (euclidean) layout: "+this.selectedEuclideanLayout.getName());
	}

	public void draw() {			
		background(255);
		//this.lights();
		/*directionalLight(101, 204, 255, -1, 0, 0);
		directionalLight(51, 102, 126, 0, -1, 0);
		directionalLight(51, 102, 126, 0, 0, -1);
		directionalLight(102, 50, 126, 1, 0, 0);
		directionalLight(51, 50, 102, 0, 1, 0);
		directionalLight(101, 204, 255, 0, 0, 1);*/
		
		//directionalLight(101, 204, 255, -1, 0, 0);
		//directionalLight(101, 204, 126, 0, -1, 0);
		//directionalLight(101, 204, 255, 0, 0, -1);
		directionalLight(102, 50, 126, 1, 0, 0);
		directionalLight(51, 102, 102, 0, 1, 0);
		ambientLight(51, 102, 102, 0, 0, 1);


		translate(width/2.f,height/2.f,-1*height/2.f);
		this.strokeWeight(1);
		stroke(150,150,150);

		this.meshRendering.draw(renderType);
	}

	/**
	 * Get mouse coordinates on the screen
	 */
	public void mouseClicked() {
		this.meshRendering.selectVertexOnScreen(mouseX, mouseY);
	}

	/**
	 * For running the PApplet as Java application <p>
	 * 
	 * Usage: java -jar PolyhedronViewer3D inputmesh.off
	 */
	public static void main(String args[]) {
		System.out.println("Spherical Layout Computation (Ecole Polytechnique, 2017)");
		if(args.length!=1) {
			System.out.println("Error: wrong arguments, one parameter required: mesh.off");
			//System.out.println("Usage example:\t ");

			System.exit(0);
		}		

		String filename=args[0];
		GraphReader reader=null;
		Polyhedron_3<Point_3> mesh=null;
		if(filename.endsWith(".txt") || filename.endsWith(".edges"))
			reader=new GraphReader_Edges();
		else if(filename.endsWith(".off") || filename.endsWith(".OFF")) {
			mesh=MeshLoader.getSurfaceMesh(filename);
			mesh.isValid(false); // check combinatorial validity

			reader=new GraphReader_OFF();
		}
		else if(filename.endsWith(".mtx"))
			reader=new GraphReader_MTX();
		else {
			System.out.println("Error: input format not supported");
			System.out.println("Supported formats are: edge list, OFF, MTX");
			System.exit(0);
		}

		AdjacencyListGraph g=reader.read(filename); // read input network from file
		MeshRepresentation input=new MeshRepresentation(g, mesh); // store the input mesh and graph

		ConfigSphericalDrawing.readInputParameters("config_SphericalLayout.txt");
		ConfigSphericalDrawing.setParameters();

		SphericalLayoutComputation.input=input; // set the input graph and input mesh
		PApplet.main(new String[] { "sd.SphericalLayoutComputation" });
	}

}
