package sd.graphalgorithms;

import Jcg.polyhedron.*;
import Jcg.schnyderwoods.EdgeOrientation;

import java.util.LinkedList;

import Jcg.geometry.Point_;
import Jcg.util.*;

/**
 * @author Luca Castelli Aleardi (Ecole Polytechnique, 2017)
 * 
 * This class allows to compute a "Rectangular Canonical Labeling" of a planar triangulation, having a 'rectangular boundary':
 * the boundary of the region defining the triangulation consists of: <p>
 * -) two vertical sides, each defined by a 'vertical' edge<p>
 * -) a bottom boundary: one of the three disjoint paths p0, p1 or p2 (the first and last vertex are not included)<p>
 * -) an upper boundary: teh 'cut border' defined by the boundary edges of the 'river'<p>
 * <p>
 * Remarks <p>
 * It works for planar (genus 0) triangle meshes without boundary.<p>
 * Faces are assumed to be ccw oriented (only for combinatorics)
 */
public class BalancedRectangularCanonicalLabeling extends RectangularCanonicalLabeling {
    	
    /** the queue storing all candidate vertices to be removed (free vertices) */
    private FreeVertices freeVerticesArray;
    private int steps=0, avoid=0;
    //private LinkedList<DListNode<Halfedge>> freeVertices;
    public static double totalDuration=0; // for evaluating time performances
    
    /**
     * Initialize all data structures and informations involved in the computation of the rectangular canonical labeling.
     * 
     * @param outerCycle  the list of halfedges defining the 'cut-border'
     * @param isOnPath  a boolean array of size E, storing for each halfedge, whether it belongs to one of the three disjoint paths p0, p1 or p2
     * @param reverseCanonicalOrdering  an array of size 'n', storing for each vertex the 'vertex conquest' order (reverse of canonical ordering)
     * @param steps  number of preliminary steps to compute a 'balanced' canonical labeling
     * @param avoid  number of extremal vertices (on the cut-border) to avoid to remove during the conquest
     */
    public BalancedRectangularCanonicalLabeling(Polyhedron_3 mesh, 
    									EdgePath outerCycle, 
    									boolean[] isEdgeOnPath, int[] xStretch, Halfedge[] left, Halfedge[] right,
    									int steps, int avoid) {
    	System.out.print("Initializing rectangular patch...");
    	if(mesh==null)
    		throw new Error("error: null polyhedron");
    	
    	this.steps=steps;
    	this.avoid=avoid;
    
    	int nE=mesh.sizeOfHalfedges();
    	int nV=mesh.sizeOfVertices();
    	
    	this.reverseCanonicalOrdering=new int[nV];
    	for(int i=0;i<nV;i++)
    		this.reverseCanonicalOrdering[i]=-1;

    	this.polyhedron=mesh;
    	this.outerCycle=outerCycle;
    	this.isOnPath=new boolean[nV];
    	this.xStretch=xStretch;
    	this.left=left;
    	this.right=right;
    	
    	// initialize edge colors: for computing the Schnyder wood // LCA: not needed for the 'Shift Algorithm'
    	this.edgeColor=new byte[nE];
    	for(int i=0;i<edgeColor.length;i++)
    		this.edgeColor[i]=-1;
     	this.isWellOriented=new boolean[nE];
   	
    	//LCA this.isChord=new boolean[nE];
    	this.isOnOuterCycle=new boolean[nV];
    	
    	// mark all edges lying on the cut-border
    	DListNode<Halfedge> pNode=outerCycle.edges.getFirst();
    	while(pNode.getElement()!=null) {
    		Halfedge e=pNode.getElement();
    		this.isOnOuterCycle[e.getVertex().index]=true;
    		this.isOnOuterCycle[e.getOpposite().getVertex().index]=true;
    		
    		pNode=pNode.getNext();
    	}
    	
    	// mark halfedges which are 'chords' and mark vertices on the bottom path
    	for(Halfedge e: this.polyhedron.halfedges) {    		
    		// mark vertices on the bottom path // LCA: this can made more efficient
    		if(isEdgeOnPath[e.index]==true) {
    			this.isOnPath[e.getVertex().index]=true; // mark the vertex on the bottom path
    		}
    	}
    	
    	//this.initFreeVertices();
    	System.out.println("done ("+this.outerCycle.edges.size()+" edges on the cut-border)");
    }
 
    /**
     * It removes a vertex 'u' from the outer cycle (vertex conquest) <p>
     * It updates the outer cycle, stting the following informations: <p> 
     * -) assign color to edges <p>
     * -) set the x-stretch of edges <p>
     * -) set the reverse canonical ordering for removed vertices <p>
     * -) store, for each removed vertex, the left and right edges incident to 'u': (u, w) and (v, u) <p>
     * 
     * @param node  The node in the double linked list storing the half-edge (v, u) (whose vertex 'u' will be removed)
     */
    public void vertexRemoval(DListNode<Halfedge> node, boolean insertInQueue) {
    	//System.out.println("vertex removal");
    	
    	
    	if(node==null) {
    		//throw new Error("Error: node not defined");
    		return;
    	}
    	/*if(node.getElement()==null) {
    		throw new Error("Error: halfedge not defined");
    	}
    	if(node.getPrev().getElement()==null) {
    		throw new Error("Error: halfedge not defined");
    	}
    	*/
    	int uIndex=node.getElement().getVertex().index; // index of the vertex 'u'
    	if(this.reverseCanonicalOrdering[uIndex]!=-1) { // vertices already removed cannot be processed
    		//throw new Error("Error: the vertex has already be removed");
    		return;
    	}
    	
    	//if(node.getPrev().getElement()==null) {
    	//	throw new Error("null reference: leftEdge");
    	//}
    	
    	Halfedge<Point_> rightEdge=node.getElement(); // the vertex v of rightEdge will be removed
    	if(this.isOnPath[rightEdge.getVertex().index]==true) // the vertex cannot be removed: it belongs to the bottom path
    		return;
    	
    	Halfedge<Point_> leftEdge=node.getPrev().getElement();
    	
    	if(rightEdge.getNext()==leftEdge) { // the vertex is incident to a triangle: it can be removed
    		this.storeReverseCanonicalOrdering(rightEdge.getVertex());
    		triangleRemoval(node, insertInQueue);
    		return;
    	}

    	if(this.hasIncidentChords(node)==true) // the vertex cannot be removed
    		return;
    	
    	// The vertex is free (it is not on the bottom path, no incident chords)
    	// general case: the vertex is incident to more than one triangle
    	
    	// process left and right edges incident to v on the cut-border
    	this.right[uIndex]=rightEdge;
    	this.left[uIndex]=leftEdge;
    	
    	this.setOutgoingEdge1(this.right[uIndex]);
    	this.setOutgoingEdge0(this.left[uIndex]);
    	
    	//this.boundary.remove(i);
    	
    	// v is not anymore on the cut-border
    	this.isOnOuterCycle[uIndex]=false;
    	
    	// add new edges to the cut-border
    	//this.addToCutBorder(rightEdge.getPrev().getOpposite(), i-1);
    	this.setToCutBorder(this.right[uIndex].getPrev().getOpposite());
    	node.setElement(this.right[uIndex].getPrev().getOpposite());
    	
    	if(insertInQueue==true) {
    		this.freeVerticesArray.addLast(node);
    		//this.freeVertices.addLast(node); // the new vertex is added to the cut-border, and it could be free // LCA: performance could be improved with a test
    	}
    	this.xStretch[node.getElement().index]=1+this.xStretch[rightEdge.index]; // set the x-stretch of the edge below the edge (v, u)
    	this.xStretch[node.getElement().getOpposite().index]=1+this.xStretch[rightEdge.index]; // set the x-stretch of the opposite halfedge
    	
    	DListNode<Halfedge> previousNode=node.getPrev(); // the previous boundary edge
   	
    	Halfedge pEdge=this.right[uIndex].getNext().getOpposite();
    	while(pEdge!=this.left[uIndex].getOpposite()) {
    		//this.addToCutBorder(pEdge.getPrev().getOpposite(), i-1); // old version
    		this.setToCutBorder(pEdge.getPrev().getOpposite());
    		this.outerCycle.edges.insertAfter(previousNode, pEdge.getPrev().getOpposite());
    		
    		if(insertInQueue==true) {
    			this.freeVerticesArray.addLast(previousNode.getNext());
    			//this.freeVertices.addLast(previousNode.getNext()); // add the new node to the potential free vertices // LCA: performance could be improved with a test
    		}
    		this.setIngoingEdge2(pEdge);
    		pEdge=pEdge.getNext().getOpposite();
    	}
    	DListNode<Halfedge> result=previousNode.getNext();
    	this.outerCycle.edges.delete(previousNode);
    	
    	this.xStretch[leftEdge.getNext().getOpposite().index]=1+this.xStretch[leftEdge.index]; // set the x-stretch of the edge below the edge (u, w)
    	this.xStretch[leftEdge.getNext().index]=1+this.xStretch[leftEdge.index]; // set the x-stretch of the opposite halfedge
    	
    	this.storeReverseCanonicalOrdering(rightEdge.getVertex());
    }

    /**
     * It removes a vertex 'u' incident to a triangle 'wuv' from the boundary cycle <p>
     * <p>
     * Remark: <p>
     * -) 'w' is the left neighbor of 'u' (on the cut-border)
     * -) 'v' is the right neighbor of 'u' (on the cut-border)
     * 
     * @param node the list node storing the halfedge whose vertex 'u' will be removed
     */
    private DListNode<Halfedge> triangleRemoval(DListNode<Halfedge> node, boolean insertInQueue) {
    	int uIndex=node.getElement().getVertex().index; // index of vertex 'u'
    	this.right[uIndex]=node.getElement(); // edge (v, u)
    	this.left[uIndex]=node.getPrev().getElement(); // edge (u, w)

    	this.setOutgoingEdge1(this.right[uIndex]);
    	this.setOutgoingEdge0(this.left[uIndex]);

    	Halfedge e=this.right[uIndex].getPrev().getOpposite();
    	node.setElement(e);
    	this.outerCycle.edges.delete(node.getPrev());

    	this.isOnOuterCycle[this.right[uIndex].getVertex().index]=false;
    	
    	this.edgeColor[e.index]=3;
    	this.edgeColor[e.getOpposite().index]=3;
    	this.isOnOuterCycle[e.getVertex().index]=true;
    	this.isOnOuterCycle[e.getOpposite().getVertex().index]=true;
    	this.setToCutBorder(e);
    	
    	// the removal of the triangle 'wuv' leads to the removal of the chord 'wv'
    	// vertices 'w' and 'v' could be free after this removal
    	Vertex w=e.getVertex();
    	if(insertInQueue==true && this.isOnPath[w.index]==false) {
    		this.freeVerticesArray.addLast(node); // the corresponding halfedge has vertex 'w' ('w' could potentially be a free vertex)
    	}
    	
    	Vertex v=e.getOpposite().getVertex();
    	if(insertInQueue==true && this.isOnPath[v.index]==false) // LCA: the second condition is redundant and could be removed
    		this.freeVerticesArray.addLast(node.getNext()); // the corresponding halfedge has vertex 'v' (a potential free vertex)

    	// set the x-stretch of the edge 'e=(wv)'
    	this.xStretch[e.index]=this.xStretch[this.left[uIndex].index]+this.xStretch[this.right[uIndex].index]+2;
    	this.xStretch[e.getOpposite().index]=this.xStretch[this.left[uIndex].index]+this.xStretch[this.right[uIndex].index]+2;
    	
    	return node;
    }

    /**
     * Initialize free vertices: initialize the queue storing all free vertices on the 'cut-border' at the beginning (before starting the conquest)<p>
     * <p>
     * Remark: the cut-border is assumed to be already initialized
     * 
     */
    public void initFreeVertices(DListNode<Halfedge>[] vertices) {
    	if(this.outerCycle.edges==null || this.outerCycle.edges.isEmpty()) {
    		System.out.println("Error: cut-border not defined or empty");
    	}
    	
    	int n=this.polyhedron.sizeOfVertices();
    	this.freeVerticesArray=new FreeVertices(n); // create the queue storing all possible free vertices (array-based implementation)
    	
    	for(DListNode<Halfedge> pNode: vertices)
    		this.freeVerticesArray.addLast(pNode);
    	
    	//System.out.print("Initializing vertices to be removed on cut-border...");
    	//System.out.println("done ("+this.freeVerticesArray.size()+" free vertices to remove)");
    }
    
    /**
     * Compute and return all (current) free vertices on the 'cut-border', stored in an array <p>
     * <p>
     * Remark: the cut-border is assumed to be already initialized
     * 
     */
    public DListNode<Halfedge>[] getAllFreeVertices() {
    	//System.out.print("Computing all free vertices on cut-border...");
    	
    	if(this.outerCycle.edges==null || this.outerCycle.edges.isEmpty()) {
    		System.out.println("Error: cut-border not defined or empty");
    	}
    	
    	int k=this.outerCycle.edges.size();
    	DListNode<Halfedge>[] temp=new DListNode[k];
    	
    	// visit all nodes on the cut-border
    	int count=0; // counter for free vertices
    	DListNode<Halfedge> pNode=outerCycle.edges.getFirst();
    	while(pNode.getElement()!=null) {
    		if(this.isFree(pNode)==true) {
    			temp[count]=pNode;
    			count++;
    		}
    		pNode=pNode.getNext();
    	}
    	
    	DListNode<Halfedge>[] result=new DListNode[count];
    	for(int i=0;i<count;i++)
    		result[i]=temp[i];
    	
    	//System.out.println("done ("+count+" free vertices on the cut-border)");
    	
    	return result;
    }
    
    /**
     * Compute and return a sub-set of free vertices on the 'cut-border', stored in an array <p>
     * <p>
     * Remark: the cut-border is assumed to be already initialized
     * 
     */
    public DListNode<Halfedge>[] getSubsetOfFreeVertices(int last) {
    	//System.out.print("Computing all free vertices on cut-border...");
    	
    	if(this.outerCycle.edges==null || this.outerCycle.edges.isEmpty()) {
    		System.out.println("Error: cut-border not defined or empty");
    	}
    	
    	int k=this.outerCycle.edges.size();
    	DListNode<Halfedge>[] temp=new DListNode[k];
    	
    	// visit all nodes on the cut-border
    	int count=0; // counter for free vertices
    	DListNode<Halfedge> pNode=outerCycle.edges.getFirst();
    	while(pNode.getElement()!=null) {
    		if(this.isFree(pNode)==true) {
    			temp[count]=pNode;
    			count++;
    		}
    		pNode=pNode.getNext();
    	}
    	
    	DListNode<Halfedge>[] result;
    	if(2*last>=count) {
    		result=new DListNode[count];
    		for(int i=0;i<count;i++)
    			result[i]=temp[i];
    	}
    	else {
    		result=new DListNode[count-2*last];
    		for(int i=last;i<count-last;i++)
    			result[i-last]=temp[i];
    	}
    	
    	// check correctness
    	for(int i=0;i<result.length;i++)
    		if(result[i]==null)
    			throw new Error("Error: wrong initialization of free vertices");
    	
    	//System.out.println("done ("+count+" free vertices on the cut-border)");
    	
    	return result;
    }

    /**
     * Compute and return a sub-set of free vertices on the 'cut-border', stored in an array <p>
     * <p>
     * Remark: the cut-border is assumed to be already initialized
     * 
     */
    public DListNode<Halfedge>[] getSubsetOfFreeVertices(int last, int delta) {
    	//System.out.print("Computing all free vertices on cut-border...");
    	
    	if(this.outerCycle.edges==null || this.outerCycle.edges.isEmpty()) {
    		System.out.println("Error: cut-border not defined or empty");
    	}
    	
    	int k=this.outerCycle.edges.size();
    	DListNode<Halfedge>[] temp=new DListNode[k];
    	
    	// visit all nodes on the cut-border
    	int count=0; // counter for free vertices
    	DListNode<Halfedge> pNode=outerCycle.edges.getFirst();
    	while(pNode.getElement()!=null) {
    		if(this.isFree(pNode)==true) {
    			temp[count]=pNode;
    			count++;
    		}
    		pNode=pNode.getNext();
    	}
    	
    	DListNode<Halfedge>[] result;
    	if(2*last>=count) {
    		result=new DListNode[count];
    		for(int i=0;i<count;i++)
    			result[i]=temp[i];
    		
    		return result;
    	}
    	else if(count-2*last<delta) {
    		result=new DListNode[count-2*last];
    		for(int i=last;i<count-last;i++)
    			result[i-last]=temp[i];
    		
    		return result;
    	}
    	else {
    		int n=(count-2*last)/delta;
    		result=new DListNode[n+1];
    		int counter=0;
    		for(int i=last;i<count-last;i=i+delta) {
    			result[counter]=temp[i];
    			counter++;
    		}
    	}
    	
    	// check correctness
    	boolean check=true;
    	for(int i=0;i<result.length;i++) {
    		if(result[i]==null) {
    			//throw new Error("Error: wrong initialization of free vertices, n="+result.length);
    			check=false;
    		}
    	}
    	
    	if(check==false) {
    		result=new DListNode[count-2*last];
    		for(int i=last;i<count-last;i++)
    			result[i-last]=temp[i];
    		
    		return result;    		
    	}
    	
    	//System.out.println("done ("+count+" free vertices on the cut-border)");
    	
    	return result;
    }
    
    /**
     * Perform all steps of the vertex conquest, computing the Schnyder wood
     */
    public void performTraversal() {
    	System.out.print("Computing 'Balanced Rectangular Canonical Labeling'...");
    	//System.out.println("Cut border: \n"+this.toString());
    	long startTime=System.nanoTime(), endTime; // for evaluating time performances
    	
    	//System.out.println("free vertices: "+this.freeVertices.size());
    	
    	
    	DListNode<Halfedge>[] vertices; // array storing a subset of free vertices on the cut-border
    	
    	// first pass
    	int countFirstPass=0; // count the number of vertex removals during the first pass
    	for(int i=0;i<this.steps;i++) {
    		int k=this.avoid; // number of vertices on the cut-border to exclude (closest to the extremities)
    		int skip=(this.steps-i)+1; // only a free vertex over 'skip' is taken
    		vertices=this.getSubsetOfFreeVertices(k, skip); // exclude the first 'k' and the last 'k' vertices on the cut-border
    		this.initFreeVertices(vertices);
    		DListNode<Halfedge> node;
    		while(this.freeVerticesArray.isEmpty()==false) {
    			node=this.freeVerticesArray.poll();
    			this.vertexRemoval(node, false); // do not update the queue storing the free vertices
    			countFirstPass++;
    		}
    	}

    	// second pass
    	int countSecondPass=0; // count the number of vertex removals during the first pass
    	for(int i=0;i<this.steps;i++) {
    		int k=this.avoid; // number of vertices on the cut-border to exclude (closest to the extremities)
    		int skip=(this.steps-i)+1; // only a free vertex over 'skip' is taken
    		vertices=this.getSubsetOfFreeVertices(k, skip); // exclude the first 'k' and the last 'k' vertices on the cut-border
    		this.initFreeVertices(vertices);
    		DListNode<Halfedge> node;
    		while(this.freeVerticesArray.isEmpty()==false) {
    			node=this.freeVerticesArray.poll();
    			this.vertexRemoval(node, true); // update the queue storing the free vertices
    			countSecondPass++;
    		}
    	}

    	// last pass :perform a standard conquest (all free vertices on the cut-border are candidates to be removed)
    	int countLastPass=0; // count the number of vertex removals during the first pass
    	vertices=this.getAllFreeVertices();
    	this.initFreeVertices(vertices);
    	while(this.freeVerticesArray.isEmpty()==false) {
    		DListNode<Halfedge> node=this.freeVerticesArray.poll();
    		this.vertexRemoval(node, true);
    		countLastPass++;
    	}

    	// set the color of the root edge, which is of color 0 (oriented toward v0)
        
    	endTime=System.nanoTime();
        double duration=(double)(endTime-startTime)/1000000000.;
        totalDuration=totalDuration+duration;
        
        DListNode<Halfedge> node=this.getFreeVertex(); // for checking correctness
        if(node!=null) {
        	throw new Error("Error: a free vertex still remaining after the conquest");
        }
        
        int totCount=countFirstPass+countSecondPass+countLastPass;
    	System.out.print("ok");
    	System.out.println(" ("+duration+" seconds)");
    	System.out.print("\t [cut-border: "+this.outerCycle.edges.size()+"]");
    	System.out.print(" - vertex removals: "+this.countVertexRemovals+", "+totCount+" total polls");
    	System.out.println(" - ("+countFirstPass+", "+countSecondPass+", "+countLastPass+")");
    }
        
    //----------------------------------------------
    //--- Methods for visualizing the cut-border ---
    //----------------------------------------------
    
    public String toString() {
    	String result="outer cycle size: "+this.outerCycle.edges.size();
    	
    	result=result+"\n boundary edges: [";
		DListNode<Halfedge> n=this.outerCycle.edges.getFirst();
		while(n!=this.outerCycle.edges.getLast()) {
			result=result+n.getElement().getVertex().index+" ";
			n=n.getNext();
		}
		result=result+n.getElement().getVertex().index+" ";
		return result+"]";
    }
   
}
