package sd.graphalgorithms;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Random;

import Jcg.geometry.Point_3;
import Jcg.polyhedron.Face;
import Jcg.polyhedron.Halfedge;
import Jcg.polyhedron.Polyhedron_3;
import Jcg.polyhedron.Vertex;
import sd.util.MeshAlgorithms;
import tc.TC;

/**
 * This class provides graph algorithms for computing shortest paths, disjoints paths, graph decompositions, ...
 * 
 * @author Luca Castelli Aleardi, Ecole Polytechnique
 * @version april 2017
 */
public class DisjointPaths {

	/** Random generator */	
	static Random generator;
	
	public DisjointPaths() {
		this.generator=new Random(1); // initialize random generator
	}
	
    /**
     * Compute three shortest vertex-disjoint paths p1, p2, p3, going from a face f1 to a face f2<p>
     * The faces are subdivided by inserting a degree 3 vertex inside <p>
     * 
     * <p>
     * Remark: the vertices are assumed to have an index, between 0..n-1<p>
     * 
     * @param f1  first face
     * @param f2  second face
     * @return  an array storing the three vertex-disjoint paths p1, p2, p3
     */		   
    public static VertexPath[] getThreeShortestDisjointPaths(Polyhedron_3 mesh, Face f1, Face f2, int minDistance){
    	if(f1==null || f2==null)
    		return null;
    	
    	VertexPath[] disjointPaths; // result to return
    	
    	boolean checkAdjacency=MeshAlgorithms.areAdjacent(f1, f2, mesh); // check whether the two faces are adjacent
    	if(checkAdjacency==true) {
    		throw new Error("Error: the two faces are adjacent");
    	}
    	
    	mesh.createCenterVertex(f1, (Point_3)f1.getBarycenter());
    	mesh.createCenterVertex(f2, (Point_3)f2.getBarycenter());
    	mesh.isValid(false);
    	
    	// the vertices 'u' and 'v' are assumed to be the last two inserted
    	Vertex u=(Vertex)mesh.vertices.get(mesh.sizeOfVertices()-2);
    	Vertex v=(Vertex)mesh.vertices.get(mesh.sizeOfVertices()-1); // last vertex inserted
    	
		//disjointPaths=getThreeDisjointPaths(mesh, u, v); // approximate version
		disjointPaths=getThreeShortestDisjointPaths(mesh, u, v, minDistance);// deterministic version
		
		return disjointPaths;
    }

    /**
     * Save the three disjoint paths to an output file<p>
     * 
     * <p>
     * Remark: the 3 paths are assumed to have the same origin and destination vertex
     * <p>
     * 
     * @param paths  3 vertex paths from 'u' to 'v'
     * @param filename  output file
     */		   
    public static void savePathsToFile(VertexPath[] paths, String filename){
    	System.out.print("Saving disjoint paths to "+filename+"...");
    	if(paths==null || paths.length!=3) {
    		System.out.println("Warning: paths not defined");
    		return;
    	}
    	
    	TC.ecritureDansNouveauFichier(filename);
    	for(int i=0;i<3;i++) {
    		for(Vertex v: paths[i].getVertices()) {
    			TC.print(v.index+" ");
    		}
    		TC.println();
    	}
    	
    	TC.ecritureSortieStandard();
    	System.out.println("done");
    }

    /**
     * Load the three disjoint paths from an file and ad two new vertices (south and north poles)
     * by splitting the south and north face <p>
     * 
     * <p>
     * Remark: the 3 paths are assumed to have the same origin and destination vertex
     * <p>
     * 
     * @param mesh	an input triangular mesh (planar and without boundaries)
     * @param filename  input file storing the 3 vertex disjoint paths
     * @return 3 vertex paths from 'u' to 'v'
     */		   
    public static VertexPath[] getThreeDisjointPathsFromFile(Polyhedron_3 mesh, String filename){
    	System.out.println("Loading disjoint paths from "+filename+"...");
    	if(filename==null || filename.length()<3) {
    		System.out.println("Warning: wrong input file: "+filename);
    	}
    	
    	VertexPath[] result=new VertexPath[3];
    	
    	TC.lectureDansFichier(filename);
    	
    	// read all vertices at the exception of the first and the last one
    	for(int i=0;i<3;i++) {
    		result[i]=new VertexPath();
    		String[] indices=TC.motsDeChaine(TC.lireLigne());
    		for(int j=1;j<indices.length-1;j++) {
    			Vertex v=(Vertex)mesh.vertices.get(Integer.parseInt(indices[j]));
    			result[i].add(v);
    		}
    	}
    	
    	TC.lectureEntreeStandard();
    	System.out.println("done");
    	
    	// find the south face
    	Face south=null;
    	Vertex v0, v1, v2; // vertices incident to the south face
    	v0=result[0].getVertices().getFirst();
    	v1=result[1].getVertices().getFirst();
    	v2=result[2].getVertices().getFirst();
    	System.out.println("South face: ("+v0.index+", "+v1.index+", "+v2.index+")");
    	ArrayList<Halfedge> edges=mesh.halfedges;
    	for(Halfedge e: edges) {
    		if(e.getOpposite().getVertex()==v0 && e.getVertex()==v1 && e.getNext().getVertex()==v2)
    			south=e.getFace();
    	}
    	if(south==null)
    		throw new Error("Error: wrong input vertex paths (south face not found)");
    	
    	// find the north face
    	Face north=null;
    	Vertex u0, u1, u2; // vertices incident to the north face. Warning: in the north face the vertices are oriented in opposite direction
    	u0=result[2].getVertices().getLast();
    	u1=result[1].getVertices().getLast();
    	u2=result[0].getVertices().getLast();
    	System.out.println("North face: ("+u0.index+", "+u1.index+", "+u2.index+")");
    	for(Halfedge e: edges) {
    		if(e.getOpposite().getVertex()==u0 && e.getVertex()==u1 && e.getNext().getVertex()==u2)
    			north=e.getFace();
    	}
    	if(north==null)
    		throw new Error("Error: wrong input vertex paths (north face not found)");
    	
    	boolean checkAdjacency=MeshAlgorithms.areAdjacent(south, north, mesh); // check whether the two faces are adjacent
    	if(checkAdjacency==true) {
    		throw new Error("Error: the two faces are adjacent");
    	}
    	
    	System.out.println("\tsouth face: f"+south.index);
    	System.out.println("\tnorth face: f"+north.index);
    	
    	// add two degree 3 vertices (by subdividing the south and north face)
    	int n=mesh.vertices.size();
    	mesh.createCenterVertex(south, (Point_3)south.getBarycenter());
    	Vertex southPole=(Vertex)mesh.vertices.get(n);
    	
    	mesh.createCenterVertex(north, (Point_3)north.getBarycenter());
    	Vertex northPole=(Vertex)mesh.vertices.get(n+1);
    	
    	mesh.isValid(false);
    	
    	// add the south and north poles to the 3 vertex paths
    	result[0].addFirst(southPole);
    	result[1].addFirst(southPole);
    	result[2].addFirst(southPole);
    	result[0].add(southPole);
    	result[1].add(southPole);
    	result[2].add(southPole);
    	
    	return result;
    }

    /**
     * Compute three shortest vertex-disjoint paths p1, p2, p3, going from vertex 'u' to vertex 'v'
     * <p>
     * Remark: the vertices are assumed to have an index, between 0..n-1<p>
     * 
     * @param u  the starting vertex
     * @param v  the destination
     * @return  an array storing the three vertex-disjoint paths p1, p2, p3
     */		   
    public static VertexPath[] getThreeShortestDisjointPaths(Polyhedron_3 mesh, Vertex u, Vertex v, int minDistance){
    	System.out.print("Computing 3 disjoint paths...");
    	if(u==null || v==null) {
    		System.err.println("Warning: south/north poles not defined");
    		return null;
    	}
    	
    	long startTime=System.nanoTime(), endTime; // for evaluating time performances

		ArrayList<VertexPath> disjointPaths=new ArrayList<VertexPath>();
		VertexPath[] result=new VertexPath[3];
		
		// compute path p0
		//System.out.println("Computing shortest path p0");
		VertexPath p1=MeshAlgorithms.computeShortestPath(mesh, u, v);
		p1.isSimplePath(mesh);
		disjointPaths.add(p1);
		
		int cycleSize=p1.getVertices().size(); // size of the shortest path
		//System.out.println(" [|p0|= "+cycleSize+"] ");
		
		Vertex medium1=p1.getVertices().get(cycleSize/2);
		LinkedList<Vertex> seedVertices=new LinkedList<Vertex>();
		seedVertices.add(medium1);
		
		int[] distances=MeshAlgorithms.getDistancesFromVertexSet(mesh, seedVertices);
		VertexPath verticesToAvoid=new VertexPath();
		for(Vertex x: (List<Vertex>)mesh.vertices) {
			if(distances[x.index]<minDistance)
				verticesToAvoid.add(x);
		}
		disjointPaths.add(verticesToAvoid);
		
		// compute path p2
		//System.out.println("Computing shortest path p1 avoiding p0");
		VertexPath p2=MeshAlgorithms.computeShortestDisjointPath(mesh, u, v, disjointPaths);
		disjointPaths.add(p2);		
		p2.isSimplePath(mesh);
		
		int cycleSize2=p2.getVertices().size(); // size of the shortest path
		//System.out.println(" [|p1|= "+cycleSize+"] ");
		
		Vertex medium2=p2.getVertices().get(cycleSize2/2);
		LinkedList<Vertex> seedVertices2=new LinkedList<Vertex>();
		seedVertices2.add(medium2);
		
		int[] distances2=MeshAlgorithms.getDistancesFromVertexSet(mesh, seedVertices2);
		VertexPath verticesToAvoid2=new VertexPath();
		for(Vertex x: (List<Vertex>)mesh.vertices) {
			if(distances2[x.index]<minDistance)
				verticesToAvoid2.add(x);
		}
		disjointPaths.add(verticesToAvoid2);

				
		//System.out.println("Computing shortest path p2 avoiding p0 and p1");
		VertexPath p3=MeshAlgorithms.computeShortestDisjointPath(mesh, u, v, disjointPaths);
		p3.isSimplePath(mesh);
		
		result[0]=p1;
		result[1]=p2;
		result[2]=p3;
		
    	endTime=System.nanoTime();
        double duration=(double)(endTime-startTime)/1000000000.;
    	System.out.println("done ("+duration+" seconds)");
		
		return result;
    }

    /**
     * Compute a simple separating cycle consisting of two paths P1 and P2: P1 from 'u' to 'v', and P2 from 'v' to 'u'
     * <p>
     * Remark: the vertices are assumed to have an index, between 0..n-1<p>
     * 
     * @param u  the starting vertex
     * @param v  the destination
     * @param minDistance  parameter for computing a well balanced edge separators
     * 
     * @return  the list of halfedges defining the cycle (halfedges must all be oriented in the same direction, in the cycle)
     */		   
    public static LinkedList<Halfedge> getSeparatingCycle(Polyhedron_3 mesh, Vertex u, Vertex v, int minDistance){
    	if(u==null || v==null)
    		return null;

    	long startTime=System.nanoTime(), endTime; // for evaluating time performances
    	System.out.print("Computing separating cycle (edge separator)...");
    	
		ArrayList<VertexPath> disjointPaths=new ArrayList<VertexPath>();
		VertexPath[] result=new VertexPath[2];
		
		// compute path p1 from south to north poles
		VertexPath p1=MeshAlgorithms.computeShortestPath(mesh, u, v);
		p1.isSimplePath(mesh);
		disjointPaths.add(p1);
		
		// get the first
		Vertex u1=p1.getVertices().get(1);
		Halfedge firstEdgeOnPath=null; // edge from u0 to u1
		Halfedge e=u.getHalfedge().getOpposite();
		if(e.getVertex()==u1) {
			firstEdgeOnPath=e;
		}
		else {
			Halfedge pEdge=e.getPrev().getOpposite();
			while(pEdge!=e && firstEdgeOnPath==null) {
				if(pEdge.getVertex()==u1)
					firstEdgeOnPath=pEdge;
				pEdge=pEdge.getPrev().getOpposite();
			}
		}

		// get the vertex incident to v on the path
		Vertex v1=p1.getVertices().get(p1.getVertices().size()-2); // get the vertex before v on the path from u to v
		Halfedge lastEdgeOnPath=null; // edge from v1 to v
		e=v.getHalfedge().getOpposite();
		if(e.getVertex()==v1) {
			lastEdgeOnPath=e;
		}
		else {
			Halfedge pEdge=e.getPrev().getOpposite();
			while(pEdge!=e && lastEdgeOnPath==null) {
				if(pEdge.getVertex()==v1)
					lastEdgeOnPath=pEdge;
				pEdge=pEdge.getPrev().getOpposite();
			}
		}

		int cycleSize=p1.getVertices().size(); // size of the shortest path
		//System.out.print(" [|p1|= "+cycleSize+"] ");
		
		Vertex medium1=p1.getVertices().get(cycleSize/2);
		//Vertex medium2=p1.getVertices().get(cycleSize/3*2);
		LinkedList<Vertex> seedVertices=new LinkedList<Vertex>();
		seedVertices.add(medium1);
		//seedVertices.add(medium2);
		
		int[] distances=MeshAlgorithms.getDistancesFromVertexSet(mesh, seedVertices);
		
		VertexPath verticesToAvoid=new VertexPath();
		
		if(minDistance>0) {
			int n=mesh.sizeOfVertices();
			for(Vertex x: (List<Vertex>)mesh.vertices) {
				if(distances[x.index]<minDistance)
					verticesToAvoid.add(x);
			}


			// two neighbors of u0 and u1 are to be avoided, in order to get a chord-free cycle
			int dU=mesh.vertexDegree(u);
			int dV=mesh.vertexDegree(v);
			if(dU<4 || dV<4)
				throw new Error("Warning: the source or destination vertices have degree 3");
			verticesToAvoid.add(firstEdgeOnPath.getPrev().getOpposite().getVertex());
			verticesToAvoid.add(firstEdgeOnPath.getOpposite().getNext().getVertex());
			verticesToAvoid.add(lastEdgeOnPath.getPrev().getOpposite().getVertex());
			verticesToAvoid.add(lastEdgeOnPath.getOpposite().getNext().getVertex());

			disjointPaths.add(verticesToAvoid);
		}
		
		// compute path p2
		//System.out.println("Computing shortest path p1 avoiding p0");
		VertexPath p2=MeshAlgorithms.computeShortestDisjointPath(mesh, u, v, disjointPaths);
		p2.isSimplePath(mesh);
		
		LinkedList<Halfedge> firstPath=(LinkedList<Halfedge>)p1.getEdgesOnPath(mesh); // get the first part of the cycle (halfedges from 'u' to 'v')
		LinkedList<Halfedge> secondPath=(LinkedList)p2.getEdgesOnPath(mesh); // get the second part of the cycle (the halfedges oriented from 'v' to 'u')
		
		DisjointPaths.checkChordsBetweenPaths(mesh, firstPath, secondPath);
		
		// first remove chords on the left of the 'first path' (inside the cycle)
		/*LinkedList<Halfedge>[] subPaths=DisjointPaths.removeChordsOnLeftSide(mesh, secondPath, firstPath, firstPath.size()/2);
		
		LinkedList<Halfedge> leftPath, rightPath;
		if(subPaths!=null) {
			leftPath=subPaths[0];
			rightPath=subPaths[1];
			System.out.println("Chords removed inside");
		}
		else {
			leftPath=secondPath;
			rightPath=firstPath;
			System.out.println("\t no chords inside");
		}*/
		
		// finally remove the chords on the left of the 'left path' (outside the cycle)
		/*subPaths=DisjointPaths.removeChordsOnLeftSide(mesh, rightPath, leftPath, leftPath.size()/2);
		if(subPaths!=null) {
			firstPath=subPaths[0];
			secondPath=subPaths[1];
			System.out.println("Chords removed outside");
		}*/

		
		// reverse the orientation of the halfedges of the second path
		firstPath=DisjointPaths.mergePaths(mesh, secondPath, firstPath);
		
    	endTime=System.nanoTime();
        double duration=(double)(endTime-startTime)/1000000000.;
    	System.out.println("done ("+duration+" seconds)");
		
		boolean check=DisjointPaths.checkSeparatingCycle(mesh, firstPath, true);
		if(check==false)
			return null;
		
		return firstPath;
    }
    
    /**
     * Given two disjoint paths P1 and P2 from 'u' to 'v' (bottom and top paths respectively) 
     * compute a separating cycle 'C' defined by the concatenation of: eRight + P2 + eLeft + P1 
     * <p>
     * Remark: the vertices are assumed to have an index, between 0..n-1<p>
     * 
     * @param bottomPath  the bottom path
     * @param topPath  the top path
     * @return  the list of halfedges defining the cycle (halfedges must all be oriented in the same direction, in the cycle)
     */		   
    public static LinkedList<Halfedge> mergePaths(Polyhedron_3 mesh, LinkedList<Halfedge> leftPath, LinkedList<Halfedge> rightPath) {
    	if(leftPath==null || rightPath==null)
    		return null;

    	System.out.print("Merging paths into a cycle...");
		// reverse the orientation of the halfedges of the second path (append the halfedges at the end of the cycle)
		Halfedge[] halfCycle=new Halfedge[leftPath.size()];
		int count=0;
		for(Halfedge e: leftPath) {
			halfCycle[count]=e.getOpposite(); // reverse the orientation of the edge
			count++;
		}
		
		// add the second path P2 (in reverse order)
		int n2=halfCycle.length;
		for(int i=n2-1;i>=0;i--)
			rightPath.add(halfCycle[i]);
		
		System.out.println("done ["+rightPath.size()+" vertices]");
		
		return rightPath;
    }


    /**
     * Given two disjoint paths P1 and P2 from 'u' to 'v' (bottom and top paths respectively) 
     * compute a separating cycle 'C' defined by the concatenation of: eRight + P2 + eLeft + P1 
     * <p>
     * Remark: the vertices are assumed to have an index, between 0..n-1<p>
     * 
     * @param bottomPath  the bottom path
     * @param topPath  the top path
     * @return  the list of halfedges defining the cycle (halfedges must all be oriented in the same direction, in the cycle)
     */		   
    public static LinkedList<Halfedge> getSeparatingCycle(Polyhedron_3 mesh, LinkedList<Halfedge> bottomPath, LinkedList<Halfedge> topPath){
    	if(bottomPath==null || topPath==null)
    		return null;

    	System.out.print("Computing a separating cycle defined by two disjoint cycles...");
    	long startTime=System.nanoTime(), endTime; // for evaluating time performances
    	
    	boolean existChords=DisjointPaths.checkChordsBetweenPaths(mesh, bottomPath, topPath);
    	
		LinkedList<Halfedge> cycle=new LinkedList<Halfedge>(); // the separating cycle to compute and return
		
		Halfedge eRight=bottomPath.getFirst().getOpposite().getPrev().getOpposite(); // edge connecting P1 to P2 (oriented toward P2)
		cycle.add(eRight);
		
		int count=0;
		int topSize=topPath.size(); // number of edges in the top path
		for(Halfedge e: topPath) {
			if(count!=0 && count!=topSize-1) // add all edges, at the exception of the first and the last one (incident to 'u' and 'v')
				cycle.add(e);
			count++;
		}
		
		Halfedge eLeft=topPath.getLast().getPrev().getOpposite(); // vertical edge, connecting P2 and P1 (oriented downward, toward to P1)
		cycle.add(eLeft);
		
		// reverse the orientation of the halfedges of the bottom path (append the halfedges at the end)
		Halfedge[] reverseBottomPath=new Halfedge[bottomPath.size()-2];
		count=0;
		for(Halfedge e: bottomPath) {
			if(count!=0 && count!=bottomPath.size()-1) {
				reverseBottomPath[count-1]=e.getOpposite(); // reverse the orientation of the edges
			}
			count++;
		}
		
		// add the bottom path P1 (in reverse order)
		int bottomSize=reverseBottomPath.length;
		for(int i=bottomSize-1;i>=0;i--)
			cycle.add(reverseBottomPath[i]);
		
    	// evaluate time performances
    	endTime=System.nanoTime();
    	double duration=(double)(endTime-startTime)/1000000000.;
    	System.out.print("done ("+duration+" seconds)");

		System.out.println("["+cycle.size()+" vertices]");
		
		return cycle;
    }

    /**
     * Check the existence of chords between two paths (bottom and top paths, sharing just their extremities) 
     * <p>
     * Remark: the vertices are assumed to have an index, between 0..n-1<p>
     * 
     * @param bottomPath  the bottom path
     * @param topPath  the top path
     * @return  true if there are chords between the two paths
     */		   
    public static boolean checkChordsBetweenPaths(Polyhedron_3 mesh, LinkedList<Halfedge> bottomPath, LinkedList<Halfedge> topPath){
    	int nV=mesh.sizeOfVertices();
    	int nE=mesh.sizeOfHalfedges();
    	
    	System.out.print("Checking paths...");

    	boolean[] isEdgeOnTop=new boolean[nE]; // mark the edges on the top path
    	boolean[] isVertexOnPaths=new boolean[nV]; // mark the edges on 
    	for(Halfedge e: topPath) {
    		Vertex x=e.getVertex();
    		isEdgeOnTop[e.index]=true;
    		isEdgeOnTop[e.getOpposite().index]=true;
    		isVertexOnPaths[x.index]=true;
    		isVertexOnPaths[e.getOpposite().getVertex().index]=true;
    	}

    	boolean[] isEdgeOnBottom=new boolean[nE]; // mark the edges on the bottom path
    	for(Halfedge e: bottomPath) {
    		Vertex x=e.getVertex();
    		isEdgeOnBottom[e.index]=true;
    		isEdgeOnBottom[e.getOpposite().index]=true;
    		isVertexOnPaths[x.index]=true;
    		isVertexOnPaths[e.getOpposite().getVertex().index]=true;
    	}

    	int count=0; // count chords
    	for(Halfedge e: (List<Halfedge>)mesh.halfedges) {
    		if(isEdgeOnTop[e.index]==false && isEdgeOnBottom[e.index]==false && isVertexOnPaths[e.getVertex().index]==true && isVertexOnPaths[e.getOpposite().getVertex().index]==true) {
    			count++;
    		}
    	}

		System.out.println("["+count+" chords]");
		
		if(count==0)
			return false;
		
		return true;
    }

    /**
     * Check the existence of chords between two paths (bottom and top paths, sharing just their extremities) 
     * <p>
     * Remark: the vertices are assumed to have an index, between 0..n-1<p>
     * 
     * @param bottomPath  the bottom path
     * @param topPath  the top path
     * @return  true if there are chords between the two paths
     */		   
    public static LinkedList<Halfedge>[] removeChordsOnLeftSide(Polyhedron_3 mesh, LinkedList<Halfedge> bottomPath, LinkedList<Halfedge> topPath, int medium){
    	int nV=mesh.sizeOfVertices();
    	int nE=mesh.sizeOfHalfedges();
    	
    	System.out.println("Computing chords on left side...");

    	boolean[] isEdgeOnTop=new boolean[nE]; // mark the edges on the top path
    	boolean[] isEdgeOnBottom=new boolean[nE]; // mark the edges on the bottom path
    	boolean[] isChord=new boolean[nE]; // mark the edges on the bottom path
    	boolean[] isVertexOnPaths=new boolean[nV]; // mark the edges on 
    	
    	// store the halfedges in two arrays, for the sake of efficiency and simplicity
    	Halfedge[] pTop=new Halfedge[topPath.size()];
    	Halfedge[] pBottom=new Halfedge[bottomPath.size()];
    	
    	int countTop=0;
    	for(Halfedge e: topPath) {
    		Vertex x=e.getVertex();
    		isEdgeOnTop[e.index]=true;
    		isEdgeOnTop[e.getOpposite().index]=true;
    		isEdgeOnBottom[e.getOpposite().index]=true;
    		isVertexOnPaths[x.index]=true;
    		isVertexOnPaths[e.getOpposite().getVertex().index]=true;
    		pTop[countTop]=e;
    		countTop++;
    	}

    	int countBottom=0;
    	for(Halfedge e: bottomPath) {
    		Vertex x=e.getVertex();
    		isEdgeOnBottom[e.index]=true;
    		isEdgeOnBottom[e.getOpposite().index]=true;
    		isVertexOnPaths[x.index]=true;
    		isVertexOnPaths[e.getOpposite().getVertex().index]=true;
    		pBottom[countBottom]=e;
    		countBottom++;
    	}
    	
    	// mark all chords
    	int countChords=0;
    	for(Halfedge e: (List<Halfedge>)mesh.halfedges) {
    		if(isEdgeOnTop[e.index]==false && isEdgeOnBottom[e.index]==false && isVertexOnPaths[e.getVertex().index]==true && isVertexOnPaths[e.getOpposite().getVertex().index]==true) {
    			isChord[e.index]=true;
    			countChords++;
    		}
    	}
    	System.out.println("\t chords: "+countChords);
    	
    	Halfedge nextChord=null;
    	int i=medium;
    	while(i<pTop.length-1 && nextChord==null) { // remark: the last vertex on the path cannot have incident chords (by construction)
    		Halfedge e=pTop[i];
    		
    		// visit all edges on the left of 'e.getVertex', in cw direction
    		Halfedge pEdge=e.getNext().getOpposite();
    		while(pEdge!=pTop[i+1].getOpposite()) {
    			if(isChord[pEdge.index]==true)
    				nextChord=pEdge.getOpposite();
    			
    			pEdge=pEdge.getNext().getOpposite();
    		}
    		
    		if(nextChord==null)
    			i++;
    	}
    	System.out.println("\t next chord: e"+nextChord+", i="+i+" |p|="+pTop.length);
		
    	Halfedge prevChord=null;
    	int j=medium;
    	while(j>0 && prevChord==null) { // remark: the last vertex on the path cannot have incident chords (by construction)
    		Halfedge e=pBottom[j];
    		
    		// visit all edges on the left of 'e.getVertex', in ccw direction
    		Halfedge pEdge=e.getPrev().getOpposite();
    		while(pEdge!=pBottom[j-1].getOpposite()) {
    			if(isChord[pEdge.index]==true)
    				prevChord=pEdge.getOpposite();
    			
    			pEdge=pEdge.getPrev().getOpposite();
    		}
    		
    		if(prevChord==null)
    			j--;
    	}
    	
    	System.out.println("\t prev chord:  e"+prevChord+", j="+j+" |p|="+pTop.length);
    	
    	return DisjointPaths.removeChordsOnLeftSide(mesh, pBottom, pTop, prevChord, nextChord, j, i);
    }

    /**
     * Compute two new paths, removing the chords on the 'left' of the right path
     * <p>
     * Remark: the vertices are assumed to have an index, between 0..n-1<p>
     * 
     * @param bottomPath  the bottom path
     * @param topPath  the top path
     * @return  true if there are chords between the two paths
     */		   
    public static LinkedList<Halfedge>[] removeChordsOnLeftSide(Polyhedron_3 mesh, Halfedge[] leftPath, Halfedge[] rightPath, Halfedge firstChord, Halfedge lastChord, int first, int last){
    	System.out.print("Removing chords on left side...");
    	
    	if(firstChord==null && lastChord==null) {
    		System.out.println("no removed chords");
    		return null;
    	}
    	
    	int nV=mesh.sizeOfVertices();
    	int nE=mesh.sizeOfHalfedges();
    	LinkedList<Halfedge> rightSubPath=new LinkedList<Halfedge>();
    	LinkedList<Halfedge> leftSubPath=new LinkedList<Halfedge>();
    	

    	// first get the start and end positions of the right sub-path
    	int rightStart, rightEnd;
    	if(firstChord==null)
    		rightStart=0;
    	else
    		rightStart=first;
    	
    	if(lastChord==null)
    		rightEnd=rightPath.length-1;
    	else
    		rightEnd=last;
    	
    	// construct the new right sub-path: it will contain the first and last chord
    	rightSubPath.add(firstChord);
    	for(int i=rightStart;i<=rightEnd;i++)
    		rightSubPath.add(rightPath[i]);
    	rightSubPath.add(lastChord);
    	
    	// --- construct left sub-path ---
    	int leftStart;
    	if(firstChord==null) {
    		leftStart=0;
    	}
    	else {
    		int firstLeftVertex=firstChord.getVertex().index;
    		
    		int i=0;
    		while(i<leftPath.length && leftPath[i].getOpposite().getVertex().index!=firstLeftVertex)
    			i++;
    		
    		leftStart=i;
    		
    	}
    	
    	int leftEnd;
    	if(lastChord==null) {
    		leftEnd=leftPath.length-1;
    	}
    	else {
    		int lastLeftVertex=lastChord.getVertex().index;
    		
    		int i=0;
    		while(i<leftPath.length && leftPath[i].getOpposite().getVertex().index!=lastLeftVertex)
    			i++;
    		
    		leftEnd=i;
    	}
    	
    	System.out.println("left start: "+leftStart+", leftEnd: "+leftEnd);
    	for(int i=leftStart;i<=leftEnd;i++)
    		leftSubPath.add(leftPath[i]);
    	
    	LinkedList<Halfedge>[] result=new LinkedList[2];
    	result[0]=leftSubPath;
    	result[1]=rightSubPath;
    	
    	System.out.print("done");
    	System.out.println("[left path: "+leftSubPath.size()+", rightpath: "+rightSubPath.size());
    	return result;
    }

    /**
     * Given a collection of halfedges defining an oriented cycle, return a cycle with opposite orientation <p>
     * <p>
     * Remark: the vertices are assumed to have an index, between 0..n-1<p>
     * 
     * @return  the list of (opposite) halfedges defining the cycle (oriented in opposite direction)
     */		   
    public static LinkedList<Halfedge> reverseCycle(Polyhedron_3 mesh, LinkedList<Halfedge> cycle){
    	if(mesh==null || cycle==null)
    		throw new Error("Error: mesh or cycle not defined");

    	LinkedList<Halfedge> oppositeCycle=new LinkedList<Halfedge>();
    	
    	for(Halfedge e: cycle) {
    		oppositeCycle.addFirst(e.getOpposite()); // add the 'opposite' halfedge, in reverse order
    	}
    	
    	return oppositeCycle;
    }

    /**
     * Compute three vertex-disjoint paths p1, p2, p3, going from vertex 'u' to vertex 'v'
     * <p>
     * Remark: the vertices are assumed to have an index, between 0..n-1<p>
     * 
     * @param u  the starting vertex
     * @param v  the destination
     * @return  an array storing the three vertex-disjoint paths p1, p2, p3
     */		   
    public static VertexPath[] getThreeDisjointPaths(Polyhedron_3 mesh, Vertex u, Vertex v){
    	if(u==null || v==null)
    		return null;

		ArrayList<VertexPath> disjointPaths=new ArrayList<VertexPath>();
		VertexPath[] result=new VertexPath[3];
		
		// compute path p1
		//System.out.println("Computing shortest path p1");
		VertexPath p1=MeshAlgorithms.computeShortestPath(mesh, u, v);
		p1.isSimplePath(mesh);
		disjointPaths.add(p1);
		
		// compute path p2
		Vertex farthestVertexFromP1=p1.getFarthestVertexFromPath(mesh, 0.2);
		//System.out.println("Computing shortest path p2a avoiding p1");
		VertexPath p2a=MeshAlgorithms.computeShortestDisjointPath(mesh, u, farthestVertexFromP1, disjointPaths);
		
		disjointPaths.add(p2a);
		//System.out.println("Computing shortest path p2b avoiding p1");
		VertexPath p2b=MeshAlgorithms.computeShortestDisjointPath(mesh, farthestVertexFromP1, v, disjointPaths);
		p2a.concatenate(mesh, p2b);
		disjointPaths.add(p2a);
		p2a.isSimplePath(mesh);
		
		// compute path p3
		VertexPath p1AndP2=new VertexPath();
		for(Vertex v1: p1.getVertices())
			p1AndP2.add(v1);
		int count=0;
		for(Vertex v2: p2a.getVertices()) {
			if(count!=0)
				p1AndP2.add(v2);
			count++;
		}
		
		Vertex farthestVertexFromP1AndP2=p1AndP2.getFarthestVertexFromPath(mesh, 0.2);
		//farthestVertexFromP1AndP2.tag=2;
		//System.out.println("Computing shortest path p3a avoiding p1 and p2");
		VertexPath p3a=MeshAlgorithms.computeShortestDisjointPath(mesh, u, farthestVertexFromP1AndP2, disjointPaths);
		//System.out.println("Computing shortest path p3b avoiding p1 and p2");
		VertexPath p3b=MeshAlgorithms.computeShortestDisjointPath(mesh, farthestVertexFromP1AndP2, v, disjointPaths);
		
		//VertexPath p3a=MeshAlgorithms.computeShortestDisjointPath(mesh, u, v, disjointPaths);

		p3a.concatenate(mesh, p3b);
		p3a.isSimplePath(mesh);
		
		result[0]=p1;
		result[1]=p2a;
		result[2]=p3a;
		
		return result;
    }

    /**
     * Given a collection of paths p1, p2, ..., pk, check whether they define a valid set of disjoint chord-free and simple paths <p>
     * <p>
     * Perform the following tests: <p>
     * -) check whether they have the same source vertex 'u' and the same destination vertex 'v'
     * -) Check whether they are simple (no auto-intersections), and they just share the first and last vertex 'u' and 'v'
     * -) check whether they are chord-free
     * <p>
     * Remark: the vertices are assumed to have an index, between 0..n-1<p>
     * 
     * @param u  the starting vertex
     * @param v  the destination
     * @param showMsg  choose to print errors or not
     * @return  true if the paths are valid (disjoint, chord-free and simple)
     */		   
    public static boolean checkDisjointPaths(Polyhedron_3 mesh, VertexPath[] paths, boolean showMsg){
    	if(mesh==null || paths==null)
    		throw new Error("Error: mesh or paths not defined");

    	int n=mesh.sizeOfVertices(); // number of vertices in the mesh
    	int k=paths.length; // number of paths
    	boolean result=true;
    	System.out.print("- Checking "+k+" disjoint paths");
    	if(showMsg==true)
    		System.out.println();
    	
    	Vertex u, v; // first and last vertex
    	u=paths[0].getVertices().getFirst();
    	v=paths[0].getVertices().getLast();
    	for(int i=1;i<k;i++) {
    		if(paths[i].getVertices().getFirst()!=u) {
    			if(showMsg==true)
    				System.out.println("path "+i+" has wrong first vertex");
    			result=false;
    		}
    		if(paths[i].getVertices().getLast()!=v) {
    			if(showMsg==true)
    				System.out.println("path "+i+" has wrong last vertex");
    			result=false;
    		}
    		if(paths[i].getVertices().size()<0) {
    			if(showMsg==true)
    				System.out.println("path "+i+" is empty");
    		}
    		if(paths[i].getVertices().size()==1) {
    			if(showMsg==true)
    				System.out.println("path "+i+" has only one vertex");
    		}
    	}
    	
    	short[][] visited=new short[k][n]; // the set of vertices already visited
    	for(int i=0;i<k;i++) {
    		for(Vertex x: paths[i].getVertices()) {
    			visited[i][x.index]++; // increment the counter (count the number of occurrences of 'x')
    			if(visited[i][x.index]>1) {
    				if(showMsg==true)
    					System.out.println("The path "+i+" is not simple");
    				result=false;
    			}
    		}
    	}
    	
    	int[] multiplicity=new int[n]; // count the multiplicity of vertices
    	for(int i=0;i<k;i++) {
    		for(Vertex x: paths[i].getVertices()) {
    			multiplicity[x.index]=multiplicity[x.index]+visited[i][x.index]; // increment the counter (count the number of occurrences of 'x')
    			if(multiplicity[x.index]>1 && x!=u && x!=v) {
    				if(showMsg==true)
    					System.out.println("Warning: vertex "+x.index+" (paths are crossing)");
    				result=false;
    			}
    		}
    	}
    	
    	for(int i=0;i<k;i++) {
    		result=result && paths[i].isChordFree(mesh);
    	}

    	if(result==true)
    		System.out.println("ok (the paths are simple, disjoint and chord free)");
    	else
    		System.out.println("warning: the paths are not valid");
    	return result;
    }

    /**
     * Given a collection of halfedges, check whether they define a valid separating cycle <p>
     * <p>
     * Remark: the vertices are assumed to have an index, between 0..n-1<p>
     * 
     * @param showMsg  choose to print errors or not
     * @return  true is the cycle is valid (simple and separating)
     */		   
    public static boolean checkSeparatingCycle(Polyhedron_3 mesh, LinkedList<Halfedge> cycle, boolean showMsg){
    	if(mesh==null || cycle==null)
    		throw new Error("Error: mesh or cycle not defined");

    	if(cycle.size()<2)
    		throw new Error("Error: the cycle has length 0 or 1");

    	int n=mesh.sizeOfVertices(); // number of vertices in the mesh
    	int E=mesh.sizeOfHalfedges(); // number of halfedges in the mesh
    	boolean result=true;
    	System.out.print("- Checking separating cycle ["+cycle.size()+" edges]...");
    	
    	Halfedge eFirst=cycle.getFirst(); // first edge in the cycle
    	Halfedge eLast=cycle.getLast();
    	Vertex vFirst, vLast; // first and last vertex: they should coincide
    	
    	vFirst=eFirst.getOpposite().getVertex();
    	vLast=eLast.getVertex();
    	
    	if(vFirst!=vLast) {
    		System.err.println("Error: the path is not a close cycle: first and last vertices do not coincide");
    		System.err.println("\t ---- try again with different parameters ---- ");
    		return false;
    	}
    	
    	Halfedge ePrevious=eLast;
    	for(Halfedge e: cycle) {
    		Vertex source=e.getOpposite().getVertex();
    		if(source!=ePrevious.getVertex()) {
    			result=false;
    			if(showMsg==true) {
    				System.out.println("Error: consecutive edges in the cycle do not share the same vertex");
    			}
    		}
    		
    		ePrevious=e;
    	}
    	
    	short[] visited=new short[n]; // the set of vertices already visited
    	for(Halfedge e: cycle) {
    		Vertex x=e.getVertex();
    		visited[x.index]++; // increment the counter (count the number of occurrences of 'x')
    		if(visited[x.index]>1) {
    			result=false;
    			if(showMsg==true){
    				//System.out.println("The cycle is not simple");
    				throw new Error("Error: the separating cycle is not simple");
    			}
    		}
    	}

    	boolean[] isEdgeOnCycle=new boolean[E]; // mark the edges on the cycle
    	boolean[] isVertexOnCycle=new boolean[n]; // mark the edges on
    	
    	System.out.println("Setting edges on cycle: "+cycle.size());
    	for(Halfedge e: cycle) {
    		isEdgeOnCycle[e.index]=true;
    		isEdgeOnCycle[e.getOpposite().index]=true;
    		isVertexOnCycle[e.getVertex().index]=true;
    		isVertexOnCycle[e.getOpposite().getVertex().index]=true;
    	}

    	for(Halfedge e: (List<Halfedge>)mesh.halfedges) {
    		if(isEdgeOnCycle[e.index]==false && isVertexOnCycle[e.getVertex().index]==true && isVertexOnCycle[e.getOpposite().getVertex().index]==true) {
    			System.out.print("Warning: the cycle is not chord-free: ");
    			System.out.println("e"+e.index+", "+e.getOpposite().getVertex().index+"-"+e.getVertex().index);
    			//throw new Error("Warning: the cycle is not chord-free");
    		}
    	}

    	if(result==true)
    		System.out.println("ok (the separating cycle is valid (simple and chord-free)");
    	else
    		System.out.println("warning: the cycle is not valid");
    	return result;
    }

    /**
     * Check whether a partition of G into G1 and G2, defined by a separating cycle 'C', is valid <p>
     * <p>
     * Remark: the vertices are assumed to have an index, between 0..n-1<p>
     * 
     * @return  true if the partition (G1, G2) defined by a separating cycle 'C' is valid
     */		   
    public static boolean checkPartition(Polyhedron_3 mesh, LinkedList<Halfedge> cycle,
    									boolean[] innerVertices, boolean[] outerVertices, 
    									boolean[] innerEdges, boolean[] outerEdges){
    	
    	System.out.print("- Checking partition...");
    	
    	if(mesh==null || cycle==null)
    		throw new Error("Error: mesh or cycle not defined");

    	if(cycle.size()<2)
    		throw new Error("Error: the cycle has length 0 or 1");

    	int n=mesh.sizeOfVertices(); // number of vertices in the mesh
    	int nE=mesh.sizeOfHalfedges();
    	boolean result=true;
    	
    	
    	// first check the partition of vertices
    	int countInnerVertices=0, countOuterVertices=0;
    	boolean[] isVertexOnCycle=new boolean[n]; // mark vertices on the cycle
    	for(Halfedge e: cycle)
    		isVertexOnCycle[e.getVertex().index]=true;
    	
    	int[] countVertices=new int[n]; // each vertex must appear exactly, in G1 or G2 or 'C'
    	for(Vertex v: (List<Vertex>)mesh.vertices) {
    		if(innerVertices[v.index]==true) {
    			countVertices[v.index]++;
    			countInnerVertices++;
    		}
    		if(outerVertices[v.index]==true) {
    			countVertices[v.index]++;
    			countOuterVertices++;
    		}
    		if(isVertexOnCycle[v.index]==true)
    			countVertices[v.index]++;
    		
    		if(countVertices[v.index]>1) {
    			result=false;
    			System.out.println("\n Warning: vertex v"+v.index+" appears "+countVertices[v.index]+" times in the partition");
    		}
    		if(countVertices[v.index]==0) {
    			result=false;
    			System.out.println("\n Warning: vertex v"+v.index+" does not appear in the partition");
    		}
    	}
    	
    	// check the partition of edges
    	int countInnerEdges=0, countOuterEdges=0;
    	boolean[] isEdgeOnCycle=new boolean[nE]; // mark vertices on the cycle
    	for(Halfedge e: cycle) {
    		isEdgeOnCycle[e.index]=true;
    		isEdgeOnCycle[e.getOpposite().index]=true;
    	}
    	
    	int[] countEdges=new int[nE]; // each halfedge must appear exactly, in G1 or G2 or 'C'
    	for(Halfedge e: (List<Halfedge>)mesh.halfedges) {
    		if(innerEdges[e.index]==true) {
    			countEdges[e.index]++;
    			countInnerEdges++;
    		}
    		if(outerEdges[e.index]==true) {
    			countEdges[e.index]++;
    			countOuterEdges++;
    		}
    		if(isEdgeOnCycle[e.index]==true)
    			countEdges[e.index]++;
    		
    		if(countEdges[e.index]>1) {
    			result=false;
    			System.out.println("Warning: edge e"+e.index+" appears "+countEdges[e.index]+" times in the partition");
    		}
    		if(countEdges[e.index]==0) {
    			result=false;
    			System.out.println("Warning: edge e"+e.index+" does not appear in the partition");
    		}
    	}
    	    	
    	if(result==true)
    		System.out.println("ok, the partition is valid");
    	else
    		System.out.println("warning: the partition is not valid");
    	
    	System.out.println("\t("+countInnerVertices+" inner vertices, "+countOuterVertices+" outer vertices, "+cycle.size()+" boundary vertices)");
    	System.out.println("\t("+(countInnerEdges/2)+" inner edges, "+(countOuterEdges/2)+" outer edges, "+cycle.size()+" boundary edges)");
    	
    	return result;
    }

    /**
     * Print a list of halfedges
     */		   
    public static void printEdges(LinkedList<Halfedge> edges){
    	if(edges==null) {
    		System.out.println("Warning: edges not defined");
    		return;
    	}
    	
    	System.out.println("List of "+edges.size()+" edges:");
     	for(Halfedge e: edges) {
     		int i0=e.getOpposite().getVertex().index;
     		int i1=e.getVertex().index;
     		System.out.print("("+i0+", "+i1+"), ");
    	}
     	System.out.println();
    }

}
