package sd.graphalgorithms;

import Jcg.geometry.Point_;
import Jcg.polyhedron.Halfedge;
import Jcg.polyhedron.Polyhedron_3;
import Jcg.polyhedron.Vertex;
import Jcg.schnyderwoods.EdgeOrientation;
import Jcg.util.DListNode;

public abstract class RectangularCanonicalLabeling extends EdgeOrientation {
    
	/** it stores the 'reverse canonical ordering'" for each vertex it stores the order of 'vertex conquest'<p>
	 * e.g.: the first removed vertex has 'reverse order' 0*/
	public int[] reverseCanonicalOrdering;

	/** it stores for each vertex the left and right edges on the cut-border (removed during the conquest) */
	public Halfedge[] left, right;

    /** store, for each edge, the corresponding stretch in the x-direction (for the shift algorithm) */
    public int[] xStretch; 
    
    // information defining the rectangular region
    /** edges defining the upper boundary of the region*/
    public EdgePath outerCycle;
    
    /** say whether a vertex is lying on one of the three disjoint paths p0, p1 or p2 */
    public boolean[] isOnPath;

	/** counter for vertex removals */
	public int countVertexRemovals=0;

    // auxiliary information needed for computing a Schnyder wood
    
    /** say whether a vertex is lying on the cut-border */
    public boolean[] isOnOuterCycle; 
     
    /**
     * Check whether a vertex on the cut-border is free
     * 
     * @param node  a node on the cut-border
     */
    public boolean isFree(DListNode<Halfedge> node) {
    	//System.out.println("vertex removal");
    	
    	if(node==null || node.getElement()==null) {
    		return false;
    	}

    	if(node==this.outerCycle.edges.getFirst()) {
    		return false;
    	}

    	Halfedge<Point_> rightEdge=node.getElement(); // the vertex v of rightEdge will be removed
    	
    	Halfedge<Point_> leftEdge=node.getPrev().getElement();
    	if(leftEdge==null) {
    		throw new Error("null reference: leftEdge not defined");
    	}
    	
    	// the vertex cannot be removed: the halfedge belongs to the bottom path
    	if(this.isOnPath[rightEdge.getVertex().index]==true) 
    		return false;
    	
    	// the vertex is incident to a triangle
    	if(rightEdge.getNext()==leftEdge) {
    		return true;
    	}

    	if(this.hasIncidentChords(node)==true)
    		return false;
    	
    	return true;
    }
    
    /**
     * Check whether a vertex has incident chords
     * In that case it cannot be removed from the outer cycle
     */
    public boolean hasIncidentChords(DListNode<Halfedge> node) {
    	//System.out.print("Checking incident chord of v"+node.getElement().getVertex().index);
    	Halfedge rightEdge=node.getElement();
    	Halfedge leftEdge=node.getPrev().getElement();

    	Halfedge pEdge=rightEdge.getNext().getOpposite();
    	while(pEdge!=leftEdge.getOpposite()) {
    		Vertex<Point_> v=pEdge.getOpposite().getVertex();
    		//System.out.println("\t v"+v.index+" is chord: "+this.isOnOuterCycle[v.index]);
    		if(this.isOnOuterCycle[v.index]==true)
    			return true;
    		pEdge=pEdge.getNext().getOpposite();
    	}
    	return false;
    }

    public void setIngoingEdge2(Halfedge e) {
    	this.edgeColor[e.index]=2;
    	this.edgeColor[e.getOpposite().index]=2;
    	
    	e.tag=3;
    	e.getOpposite().tag=3;
    	
    	this.isWellOriented[e.index]=true;
    	this.isWellOriented[e.getOpposite().index]=false;
    }

    public void setOutgoingEdge1(Halfedge e) {
    	this.edgeColor[e.index]=1;
    	this.edgeColor[e.getOpposite().index]=1;
    	
    	e.tag=2; // LCA: for rendering
    	e.opposite.tag=2; // LCA: for rendering

    	this.isWellOriented[e.index]=false;
    	this.isWellOriented[e.getOpposite().index]=true;
    }

    public void setOutgoingEdge0(Halfedge e) {
    	this.edgeColor[e.index]=0;
    	this.edgeColor[e.getOpposite().index]=0;
    	
    	e.tag=1; // LCA: for rendering
    	e.opposite.tag=1; // LCA: for rendering
    	
    	this.isWellOriented[e.index]=true;
    	this.isWellOriented[e.getOpposite().index]=false;
    }

    /**
     * Add a half-edge to the outer cycle, at a given position (in the cycle)
     * Update all concerned information (colors, existent chords, boundary vertices, ...)
     */
    public void setToCutBorder(Halfedge e) {
    	this.edgeColor[e.index]=3;
    	this.edgeColor[e.getOpposite().index]=3;
    	e.tag=5; // LCA: for rendering and debugging
    	e.opposite.tag=5; // LVA: for rendering and debugging
    	
    	//LCA this.isChord[e.index]=false;
    	//LCA this.isChord[e.getOpposite().index]=false;
    	
    	this.isOnOuterCycle[e.getVertex().index]=true;
    	this.isOnOuterCycle[e.getOpposite().getVertex().index]=true;
    }

    /**
     * Sore the 'vertex conquest order' ('reverse canonical ordering') for a vertex 'v'
     */
    public void storeReverseCanonicalOrdering(Vertex v) {
    	if(this.reverseCanonicalOrdering[v.index]!=-1) {
    		throw new Error("Error: vertex v"+v.index+" has already been removed");
    	}
    	
		this.reverseCanonicalOrdering[v.index]=this.countVertexRemovals; //  set the 'reverse canonical ordering' for the removed vertex
		this.countVertexRemovals++; // update the number of vertex removals
    }

    /**
     * Perform all steps of the vertex conquest, computing the Schnyder wood/Canonical labeling
     */
    public abstract void performTraversal();
        
    /**
     * Return an array of indices representing the original vertex ordering
     */
    public int[] getOriginalVertexOrdering() {
    	int[] result=new int[this.polyhedron.sizeOfVertices()];
    	
    	int i=0;
    	for(Object o: this.polyhedron.vertices) {
    		Vertex<Point_> v=(Vertex<Point_>)o;
    		result[i]=v.index;
    		i++;
    	}
    	return result;    	
    }
	
    //----------------------------------------------
    //--- Methods for visualizing the cut-border ---
    //----------------------------------------------
    
    public String[] originalVertexOrderingToString() {
    	String[] result=new String[this.polyhedron.sizeOfVertices()];
    	
    	int i=0;
    	for(Object o: this.polyhedron.vertices) {
    		Vertex<Point_> v=(Vertex<Point_>)o;
    		result[i]=""+v.index;
    		i++;
    	}
    	return result;    	
    }

    public String toString() {
    	String result="outer cycle size: "+this.outerCycle.edges.size();
    	
    	result=result+"\n boundary edges: [";
		DListNode<Halfedge> n=this.outerCycle.edges.getFirst();
		while(n!=this.outerCycle.edges.getLast()) {
			result=result+n.getElement().getVertex().index+" ";
			n=n.getNext();
		}
		result=result+n.getElement().getVertex().index+" ";
		return result+"]";
    }
   
    //----------------------------------------------
    //-- Methods for debugging the vertex conquest -
    //----------------------------------------------
    
    public boolean checkCutBorder() {
    	// mark all edges lying on the cut-border
    	DListNode<Halfedge> pNode=outerCycle.edges.getFirst();
    	Vertex u=pNode.getElement().getVertex();
    	Vertex v=pNode.getElement().getOpposite().getVertex();
 
    	if(this.isOnPath[u.index]==false) {
    		throw new Error("Error: the left most vertex on the cut border is not on the bottom path");
    	}

    	if(this.isOnOuterCycle[u.index]==false) {
    		throw new Error("Error: the left most vertex on the cut border is not marked");
    	}
    	
    	pNode=pNode.getNext(); // traverse the second node on the cut-border (from the left)
    	while(pNode.getElement()!=null) {
    		Halfedge e=pNode.getElement();
    		u=pNode.getElement().getVertex(); 
    		v=pNode.getPrev().getElement().getOpposite().getVertex();
    		if(u!=v) {
    			throw new Error("Error: the cut-border is not connected");
    		}
    		
    		pNode=pNode.getNext();
    	}
    	
    	v=pNode.getPrev().getElement().getOpposite().getVertex();
    	if(this.isOnPath[v.index]==false) {
    		throw new Error("Error: the right most vertex on the cut border is not on the bottom path");
    	}

    	if(this.isOnOuterCycle[v.index]==false) {
    		throw new Error("Error: the right most vertex on the cut border is not marked");
    	}
    	
    	return true;
    }

    public DListNode<Halfedge> getFreeVertex() {
    	// mark all edges lying on the cut-border
    	DListNode<Halfedge> pNode=outerCycle.edges.getFirst();
     	
    	while(pNode.getElement()!=null) {
    		if(this.isFree(pNode)==true)
    			return pNode;
    		
    		pNode=pNode.getNext();
    	}
    	
    	return null;
    }

}
