package sd.graphalgorithms;

import java.util.LinkedList;
import java.util.List;

import Jcg.polyhedron.Face;
import Jcg.polyhedron.Halfedge;
import Jcg.polyhedron.Polyhedron_3;
import Jcg.polyhedron.Vertex;
import sd.util.MeshAlgorithms;

/**
 * This class perform the computation of the 'river' (path in the dual mesh) <p>
 * 
 * @author Luca Castelli Aleardi, Ecole Polytechnique
 * @version april 2017
 */
public class RiverComputation {

	/** The three disjoint paths p0, p1, p2 (encountered in cw order around vertex u) */
	public LinkedList<Halfedge> path0, path1, path2;
	
	/** The vertex source and vertex destination */
	public Vertex u, v;
	
	/** The three faces incident to the destination vertex*/
	public Face d0, d1, d2;

	/** The three faces incident to the source vertex */
	public Face s0, s1, s2;
	
	/** says whether a face belong to a river: result of the computation */
	public boolean[] isOnRiver;

	/** says whether an edge is on one the three paths p0, p1 or p2: these are 'wall' edges (such edges cannot be traversed) */
	public boolean[] isOnPath;

	/** auxiliary table for performing a BFS traversal of the mesh (for computing the 'river') */
	private boolean[] visited;
	
	public static double totalDuration=0; // for evaluating time performances

	/**
	 * Initialize the three paths <p>
	 * Remark: the three paths are assumed to be simple, disjoint and chord-free <p>
	 * Remark: 'u' and 'v' are assumed to be degree 3 vertices
	 */
	public RiverComputation(Polyhedron_3 mesh, VertexPath[] vertexPaths) {
		this.u=vertexPaths[0].getVertices().getFirst(); // vertex in the south face
		this.v=vertexPaths[0].getVertices().getLast(); // vertex in the north face
		if(mesh.vertexDegree(u)!=3 || mesh.vertexDegree(v)!=3)
			throw new Error("Error: wrong vertex degree");
		
		// check path validity
		DisjointPaths.checkDisjointPaths(mesh, vertexPaths, false);
		
		// convert from vertex paths to edge paths (halfedges are outgoing from u)
		LinkedList<Halfedge> edgePath0=(LinkedList)vertexPaths[0].getEdgesOnPath(mesh);
		LinkedList<Halfedge> edgePath1=(LinkedList)vertexPaths[1].getEdgesOnPath(mesh);
		LinkedList<Halfedge> edgePath2=(LinkedList)vertexPaths[2].getEdgesOnPath(mesh);
		
		// order paths according to the cyclic order of edges around vertices: in cw order we encounter p0, p1, p2
		path0=edgePath0; // take an arbitrary first path p0
		Halfedge e0=edgePath0.getFirst();
		Halfedge e1=edgePath1.getFirst();
		
		if(e0.getOpposite().getNext()==e1) { // only works if 'u' has degree 3
			path1=edgePath1;
			path2=edgePath2;
		}
		else {
			path1=edgePath2;
			path2=edgePath1;
		}
		
		// set the three source and destination faces for computing the three 'rivers' (three paths in the dual mesh)
		this.s0=path0.getFirst().getFace();
		this.d0=path0.getLast().getFace();
		this.s1=path1.getFirst().getFace();
		this.d1=path1.getLast().getFace(); // LCA: we can avoid to perform getLast()
		this.s2=path2.getFirst().getFace();
		this.d2=path2.getLast().getFace(); // LCA: we can avoid to perform getLast()
		
		int f=mesh.sizeOfFacets();
		this.isOnRiver=new boolean[f];
		this.visited=new boolean[f];
		this.isOnPath=new boolean[mesh.sizeOfHalfedges()];
		
		for(Halfedge e: this.path0) {
			this.isOnPath[e.index]=true;
			this.isOnPath[e.getOpposite().index]=true;
		}
		for(Halfedge e: this.path1) {
			this.isOnPath[e.index]=true;
			this.isOnPath[e.getOpposite().index]=true;
		}
		for(Halfedge e: this.path2) {
			this.isOnPath[e.index]=true;
			this.isOnPath[e.getOpposite().index]=true;
		}
	}
	
    /**
     * Compute a (BFS) shortest path in the dual graph, between two faces
     * It performs a BFS visit of the (dual) mesh starting from the face 'source':
     * the edges on the paths p0, p1 and p2 cannot be traversed (they are 'walls')<p>
     * <p>
     * During the traversal the faces belonging to the river are marked (see array isOnRiver[])
     * <p>
     * Remark: the vertices are assumed to have an index, between 0..n-1
     * 
     * @param source  the starting face
     * @param destination  the destination face
     * @return  an array storing for each face the previous reference (in the dual path)
     */		   
    public Face[] computeShortestPathOnDual(Polyhedron_3 mesh, Face source, Face destination){
    	if(source==null || destination==null)
    		return null;
    	
    	System.out.print("Computing river (dual shortest path) from f"+source.index+"...");
    	long startTime=System.nanoTime(), endTime; // for evaluating time performances
    	
    	int nFaces=mesh.sizeOfFacets(); // number of vertices in the mesh
    	
    	LinkedList<Face> queue=new LinkedList<Face>(); // stack containing the face to visit
    	Face[] previous=new Face[nFaces]; // used to retrieve the shortest path (in the dual mesh)
    	    	
    	queue.add(source);
    	this.visited[source.index]=true;
    	previous[source.index]=source; // the source face has itself as previous
    	
    	while(queue.isEmpty()==false) {
    		Face f=queue.poll(); // get and removes the node in the head of the stack
    		//f.tag=3; // LCA: only for rendering (remove it for measuring time performances)
    		
    		if(f==destination) {
    			//System.out.println("ok: destination reached");
    			// evaluate time performances
    		}
    		
    		else{
    			Halfedge[] neighbors=new Halfedge[3]; // the three halfedges of the face
    			neighbors[0]=f.getEdge();
    			neighbors[1]=neighbors[0].getNext();
    			neighbors[2]=neighbors[1].getNext();
    			
    			for(Halfedge e: neighbors) { // visit the three neighboring faces of 'f'
    				Face neighbor=e.getOpposite().getFace(); // neighboring face, sharing edge 'e' with 'f'
    				if(this.visited[neighbor.index]==false && this.isOnPath[e.index]==false) { // do not cross 'wall' edges
    					this.visited[neighbor.index]=true; // mark the face as visited
    					previous[neighbor.index]=f;
    					//f.tag=3;
    					queue.add(neighbor);
    				}
    			}
    		}
    	}
    	
    	// reconstruct the shortest path in the dual mesh (using 'previous' references)
    	Face f=destination; // start from last node (destination face)
    	while(previous[f.index]!=f) { // traverse the path in reverse order
    		this.isOnRiver[f.index]=true; // mark the face: it belongs to the river
    		f.tag=4; // LCA: only for rendering
    		f=previous[f.index];
    	}
    	source.tag=2;

    	// evaluate time performances
    	endTime=System.nanoTime();
    	double duration=(double)(endTime-startTime)/1000000000.;
    	totalDuration=totalDuration+duration;
    	System.out.println("done ("+duration+" seconds)");

    	return previous;
    }

    /**
     * Compute a 'river' from the source to the destination, which is close to the 'top path': it contains
     * all faces incident to the top path<p>
     * <p>
     * Warning: this method works only for chord-free top paths <p>
     * <p>
     * During the traversal the faces belonging to the river are marked (see array isOnRiver[])
     * <p>
     * Remark: the vertices are assumed to have an index, between 0..n-1
     * 
     * @param bottomPath  the bottom path consisting of the halfedges delimiting the top boundary of the region
     * @param topPath  the top path consisting of the halfedges delimiting the top boundary of the region
     * 
     * @return  an array storing for each face the previous reference (in the dual path)
     */		   
    public Face[] computeRiverClosestToTopPath(Polyhedron_3 mesh, LinkedList<Halfedge> bottomPath, LinkedList<Halfedge> topPath){
    	Face source=topPath.getFirst().getFace();
    	Face destination=topPath.getLast().getFace();
    	
    	if(source==null || destination==null)
    		return null;	
    	if(bottomPath.getFirst().getOpposite().getFace()!=source) // check that the source face is incident to both the bottom and top paths
    		throw new Error("Error: bottom and top paths do not share the same source face");
    	if(bottomPath.getLast().getOpposite().getFace()!=destination) // check that the destination face is incident to both the bottom and top paths
    		throw new Error("Error: bottom and top paths do not share the same destination face");
    	
    	System.out.print("Computing river (closest to top path) from f"+source.index+"...");
    	long startTime=System.nanoTime(), endTime; // for evaluating time performances
    	
    	int nFaces=mesh.sizeOfFacets(); // number of vertices in the mesh
    	
    	Face[] previous=new Face[nFaces]; // used to retrieve the shortest path (in the dual mesh)
    	    	
    	//this.visited[source.index]=true;
    	previous[source.index]=source; // the source face has itself as previous
    	Halfedge startEdge=topPath.getFirst().getNext().getOpposite(); // the first half-edge (oriented toward the top path)
    	Halfedge lastEdge=topPath.getLast().getPrev(); // the last half-edge to stop the visit (inside the destination face)
    	
    	Halfedge pEdge=startEdge;
    	Face previousFace=source; // the first 'inner' face into the river has the source as 'previous face'
    	
    	while(pEdge!=lastEdge) {
    		Face currentFace=pEdge.getFace();
    		previous[currentFace.index]=previousFace;
    		
    		// move to the next neighboring face (incident to the top path): we must distinguish two cases
    		if(this.isOnPath[pEdge.getNext().index]==false) { // case 1: the next edge is not on the top path
    			pEdge=pEdge.getNext().getOpposite();
    		}
    		else { // case 2: the next edge (in the current face) is on the top path (IMPORTANT: the top path is assumed to be chord-free)
    			pEdge=pEdge.getNext().getNext().getOpposite();
    		}
    		
    		previousFace=currentFace; // update the previous face
    	}
    	
    	// deal with the last face (destination)
    	previous[destination.index]=previousFace;
    	
    	// reconstruct the shortest path in the dual mesh (using 'previous' references)
    	Face f=destination; // start from last node (destination face)
    	while(previous[f.index]!=f) { // traverse the path in reverse order
    		this.isOnRiver[f.index]=true; // mark the face: it belongs to the river
    		f.tag=4; // LCA: only for rendering
    		f=previous[f.index];
    	}
    	source.tag=2;

    	// evaluate time performances
    	endTime=System.nanoTime();
    	double duration=(double)(endTime-startTime)/1000000000.;
    	totalDuration=totalDuration+duration;
    	System.out.println("done ("+duration+" seconds)");

    	return previous;
    }

    /**
     * Given a face-path (a set of faces defining in the dual graph) from source to the destination, compute and return
     * the bottom edge-path consisting of the halfedges defining the bottom border of the face-path
     * <p>
     * Remark: we assume given a valid 'river'. The information concerning faces inside the river is stored in the
     * boolean array 'isOnRiver'
     * <p>
     * Remark: the vertices are assumed to have an index, between 0..n-1
     * 
     * @param bottomPath  bottom edge path in the rectangle region (one of the three paths p0, p1 or p2)
     * @return and egde-path defining the bottom boundary cycle
     */		   
    public EdgePath getBottomBoundary(Polyhedron_3 mesh, LinkedList<Halfedge> bottomPath){
    	if(bottomPath==null) {
    		return null;
    	}
    	System.out.print("Computing bottom boundary...");
    	long startTime=System.nanoTime(), endTime; // for evaluating time performances
    	
    	EdgePath result=new EdgePath();
    	Face source=bottomPath.getFirst().getOpposite().getFace();
    	Face destination=bottomPath.getLast().getOpposite().getFace();
    	
    	destination.tag=5;
    	
    	Vertex vLeft=bottomPath.getLast().getOpposite().getVertex(); // bottom left vertex (of the rectangular frame)
    	Vertex vRight=bottomPath.getFirst().getVertex(); // bottom right vertex (of the rectangular frame)
    	
    	// get the starting and ending half-edge
    	Halfedge eStart=bottomPath.getLast().getOpposite().getNext().getOpposite(); // halfedge on the left vertical side
    	Halfedge eEnd=bottomPath.getFirst().getOpposite().getPrev(); // halfedge on the right vertical side
    	
    	eStart.tag=5;
    	eStart.getOpposite().tag=5;
    	eEnd.tag=5;
    	eEnd.getOpposite().tag=5;
    	
    	Halfedge pEdge=eStart; // start the traversal of the river from the left vertical side
    	while(pEdge!=eEnd) { // the traversal ends at the right vertical side
    		Face f=pEdge.getFace();
    		// distinguish 3 cases
    		if(this.isOnPath[pEdge.index]==true) { // the edge belong to the bottom horizontal path of the frame
    			// only for rendering
    			pEdge.tag=5;
    			pEdge.getOpposite().tag=5;
    			
    			result.edges.add(pEdge);
    			pEdge=pEdge.getOpposite().getNext().getOpposite(); // move to the next edge in the river
    		}
    		else if(this.isOnRiver[f.index]==true) { // the halfedge is inside the 'river'
    			pEdge=pEdge.getNext().getOpposite(); // move to next halfedge in cw order
    		}
    		else { // the halfedge is on the boundary of the river ('outside' halfedge)
    			// only for rendering
    			pEdge.tag=5;
    			pEdge.getOpposite().tag=5;
    			
    			result.edges.add(pEdge);
    			pEdge=pEdge.getOpposite().getNext().getOpposite(); // move to the next halfedge inside the river
    		}
    	}

    	endTime=System.nanoTime();
        double duration=(double)(endTime-startTime)/1000000000.;
        totalDuration=totalDuration+duration;
        System.out.println("done ("+duration+" seconds)");

        // check validity: the that the leftmost and rightmost vertices on the bottom path are correct
        Vertex checkLeft=result.edges.getFirst().getElement().getVertex();
        Vertex checkRight=result.edges.getLast().getElement().getOpposite().getVertex();
        if(vLeft!=checkLeft)
        	System.out.println("Error: wrong left bottom vertex");
        if(vRight!=checkRight)
        	System.out.println("Error: wrong right bottom vertex");
        
        return result;
    }

        /**
         * Given a face-path (a set of faces defining in the dual graph) from source to the destination, compute and return
         * the top edge-path consisting of the halfedges defining the top border of the face-path
         * <p>
         * Remark: we assume given a valid 'river'. The information concerning faces inside the river is stored in the
         * boolean array 'isOnRiver' <p>
         * Important remark: the halfedges on the paths p0, p1 and p2 are oriented from the source to the destination, while on the top 'cut-border'
         * must be oriented from the destination to the source. <p>
         * <p>
         * Remark: the vertices are assumed to have an index, between 0..n-1
         * 
         * @param topPath  bottom edge path in the rectangle region (one of the three paths p0, p1 or p2)
         * @return and egde-path defining the bottom boundary cycle
         */		   
        public EdgePath getTopBoundary(Polyhedron_3 mesh, LinkedList<Halfedge> topPath){
        	if(topPath==null) {
        		return null;
        	}
        	System.out.print("Computing top boundary...");
        	long startTime=System.nanoTime(), endTime; // for evaluating time performances
        	
        	EdgePath result=new EdgePath();
        	Face source=topPath.getFirst().getFace();
        	Face destination=topPath.getLast().getFace();
        	
        	destination.tag=5;
        	
        	Vertex vLeft=topPath.getLast().getOpposite().getVertex(); // top left vertex (of the rectangular frame)
        	Vertex vRight=topPath.getFirst().getVertex(); // top right vertex (of the rectangular frame)
        	
        	// get the starting and ending half-edge
        	Halfedge eStart=topPath.getFirst().getNext().getOpposite(); // halfedge on the right vertical side (inside the river, directed upward)
        	Halfedge eEnd=topPath.getLast().getPrev(); // halfedge on the left vertical side (outside the river, directed upward)
        	
        	eStart.tag=5;
        	eStart.getOpposite().tag=5;
        	eEnd.tag=5;
        	eEnd.getOpposite().tag=5;
        	
        	Halfedge pEdge=eStart; // start the traversal of the river from the right vertical side
        	while(pEdge!=eEnd) { // the traversal ends at the right vertical side
        		Face f=pEdge.getFace();
        		// distinguish 3 cases
        		if(this.isOnPath[pEdge.index]==true) { // the edge belong to the top horizontal path of the frame
        			// only for rendering
        			pEdge.tag=5;
        			pEdge.getOpposite().tag=5;
        			
        			result.edges.add(pEdge);
        			pEdge=pEdge.getOpposite().getNext().getOpposite(); // move to the next edge in the river (turning in cw direction)
        		}
        		else if(this.isOnRiver[f.index]==true) { // the halfedge is inside the 'river'
        			pEdge=pEdge.getNext().getOpposite(); // move to next halfedge in cw order
        		}
        		else { // the halfedge is on the boundary of the river ('outside' halfedge)
        			// only for rendering
        			pEdge.tag=5;
        			pEdge.getOpposite().tag=5;
        			
        			result.edges.add(pEdge);
        			pEdge=pEdge.getOpposite().getNext().getOpposite(); // move to the next halfedge inside the river
        		}
        	}

    	endTime=System.nanoTime();
        double duration=(double)(endTime-startTime)/1000000000.;
        totalDuration=totalDuration+duration;
        System.out.println("done ("+duration+" seconds)");

        // check validity: the leftmost and rightmost vertices on the bottom path are correct
        Vertex checkRight=result.edges.getFirst().getElement().getVertex();
        Vertex checkLeft=result.edges.getLast().getElement().getOpposite().getVertex();
        if(vLeft!=checkLeft)
        	System.out.println("Error: wrong left top vertex");
        if(vRight!=checkRight)
        	System.out.println("Error: wrong right top vertex");
        
        return result;
    }

    /**
     * Check whether the 'river' (path in the dual graph) is valid<p>
     * Remark: a river is valid if: <p>
     * -) it defines an outerplanar graph, with a simple boundary (auto-intersection)
     * <p>
     * Remark: the vertices are assumed to have an index, between 0..n-1
     * 
     * @param destinayion  the destination face
     * @param previous  an array storing for each face the previous reference (in the river)
     */		   
    public boolean checkRiver(Polyhedron_3 mesh, Face[] previous, Face destination){
    	if(destination==null)
    		throw new Error("Error: destination face not defined");
    	
    	System.out.print("- Checking river: destination f"+destination.index+"...");
    	boolean result=true;
    	
    	// check whether the river defines a connected path in the dual mesh
    	Face f=destination; // start from last node (destination face)
    	int count=0;
    	while(previous[f.index]!=f) { // traverse the path in the river in reverse order
    		if(MeshAlgorithms.getCommonHalfedge(f, previous[f.index])==null) { // check whether two consecutive faces in the path are adjacent in the mesh
    			System.out.println("\t Warning: the river is not face-connected");
    			return false;
    		}
    		f=previous[f.index];
    		count++;
    	}
    	//System.out.println("The river has size: "+count);

    	// check the boundary of the river is simple
    	int n=mesh.sizeOfVertices();
    	short[] encountered=new short[n]; // store the number of occurrences of boundary vertices on the river (each vertex should be added once)
    	Face nextFace=destination; // start from last node (destination face)
    	f=previous[nextFace.index]; // face preceding the destination
    	while(previous[f.index]!=f) { // traverse the path in the river in reverse order
    		Halfedge common=MeshAlgorithms.getCommonHalfedge(f, nextFace); // the common edge shared by 'f' and its neighboring face: 'common' is inside 'f'
    		if(common==null) {
    			throw new Error("Error: the two neighboring faces do not share any edge");
    		}
    		
    		Vertex thirdVertex=common.getNext().getVertex(); // the third vertex of the face 'f'
    		encountered[thirdVertex.index]++; // increment the counter
    		
    		if(encountered[thirdVertex.index]>1) { // check whether the third vertex is already on the boundary (in that case the 'river' is not simple)
    			System.out.println("\t Warning: the boundary of the river is not simple: vertex v"+thirdVertex.index+" appears more than once");
    			result=false;
    		}
    		
    		nextFace=f; // update the neighboring face
    		f=previous[f.index];
    	}

    	if(result==true)
    		System.out.println("ok ("+count+" faces)");
    	else
    		System.out.println("error: river not valid");
    		
    	return result;
    }

}
