package sd.graphalgorithms;

import java.util.ArrayList;
import java.util.LinkedList;
import java.util.List;
import java.util.Random;

import Jcg.geometry.Point_3;
import Jcg.polyhedron.Face;
import Jcg.polyhedron.Halfedge;
import Jcg.polyhedron.Polyhedron_3;
import Jcg.polyhedron.Vertex;
import sd.util.MeshAlgorithms;

/**
 * A class providing implementation of algorithms involving vertex paths on meshes
 * 
 * @author Luca Castelli Aleardi, Ecole Polytechnique
 * @version april 2017
 */
public class VertexPath {

	/** Random generator */	
	static Random generator = new Random(1); // initialize random generator

	/** vertices on the path*/
	private LinkedList<Vertex> vertices;
	
	public VertexPath() {
		this.vertices=new LinkedList<Vertex>(); // create an empty path
	}
	
	public void add(Vertex v) {
		this.vertices.add(v);
	}
	
	public void addFirst(Vertex v) {
		this.vertices.addFirst(v);
	}

	public LinkedList<Vertex> getVertices() {
		return this.vertices;
	}

    /**
     * Check whether the input path is simple
     * <p>
     * Remark: the vertices are assumed to have an index, between 0..n-1
     * 
     * @return true if the input path is simple (no vertex appears twice or more times)
     */		   
    public boolean isSimplePath(Polyhedron_3 mesh){
    	if(vertices==null)
    		throw new Error("Error: vertex path not defined");
    	
    	int n=mesh.sizeOfVertices(); // number of vertices in the mesh
    	//System.out.print("Checking path ...");
    	
    	short[] visited=new short[n]; // the set of vertices already visited
    	for(Vertex v: this.vertices) {
    		visited[v.index]++; // increment the counter (count the number of occurrences of 'v')
    		if(visited[v.index]>1) {
    			System.out.println("error: the path is not simple: vertex "+v.index+" appears twice");
    			return false;
    		}
    	}
    	
    	//System.out.println("ok the path is simple ("+vertices.size()+" vertices)");
    	return true;
    }

    /**
     * Check whether the input path is chord free
     * <p>
     * Remark: the vertices are assumed to have an index, between 0..n-1
     * 
     * @return true if the input path is chord free (no edge having its two extremities on the path)
     */		   
    public boolean isChordFree(Polyhedron_3 mesh){
    	if(vertices==null)
    		throw new Error("Error: vertex path not defined");
    	
    	boolean result=true;
    	int n=mesh.sizeOfVertices(); // number of vertices in the mesh
    	System.out.print("Checking for chords...");
    	
    	boolean[] onPath=new boolean[n]; // the set of vertices on the
    	for(Vertex v: this.vertices) {
    		onPath[v.index]=true; // mark vertices which are on the path
    	}
    	
    	List<Halfedge> pathEdges=this.getEdgesOnPath(mesh);
    	boolean[] edgesOnPath=new boolean[mesh.sizeOfHalfedges()];
    	for(Halfedge e: pathEdges) { // mark all edges on the path from 'u' to 'v'
    		edgesOnPath[e.index]=true;
    		edgesOnPath[e.getOpposite().index]=true;
    	}
    	
    	for(Halfedge e: (List<Halfedge>)mesh.halfedges) {
    		Vertex a=e.getVertex();
    		Vertex b=e.getOpposite().getVertex();
    		if(onPath[a.index]==true && onPath[b.index]==true && edgesOnPath[e.index]==false) {
    			System.out.println("edge ("+a.index+", "+b.index+") is a chord");
    			result=false;
    		}
    	}
    	
    	if(result==true)
    		System.out.println("ok the path is chord free ("+vertices.size()+" vertices)");
    	else
    		System.out.println("warning: the path is not chord free ("+vertices.size()+" vertices)");
    	return result;
    }
    
    /**
     * Remove an intersection from the path <p>
     * <p>
     * Warning: it works for only one intersection <p>
     * Warning: to check (it could be not safe)
     * <p>
     * Remark: the vertices are assumed to have an index, between 0..n-1<p>
     * 
     * @param showMsg  choose to print errors or not
     * @return  a new vertex path (intersection free)
     */		   
    public VertexPath removeAutointersection(Polyhedron_3 mesh, boolean showMsg){
    	if(mesh==null)
    		throw new Error("Error: path not defined");

    	int n=mesh.sizeOfVertices(); // number of vertices in the mesh
    	VertexPath result=new VertexPath(); // the new path (intersection free) to return
    	System.out.print("Removing auto-intersections...");
    	if(showMsg==true)
    		System.out.println();
    	
    	Vertex u, v; // first and last vertex
    	u=this.vertices.getFirst();
    	v=this.vertices.getLast();
    	
    	int[] multiplicity=new int[n]; // count vertex multiplicity (for all vertices)
    	Vertex multipleVertex=null; // the first multiple vertex
    	int lastOccurrence=-1;
    	int firstOccurrence=-1;
    	
    	int count=0;
    	for(Vertex x: this.vertices) {
    		multiplicity[x.index]++; // increment the counter (count the number of occurrences of 'x')
    		if(multiplicity[x.index]>1) {
    			if(multipleVertex==null) {
    				multipleVertex=x; // found the first multiple vertex
    				firstOccurrence=count; // store the rank of the first occurrence of the multiple vertex
    			}
    			if(x==multipleVertex)
    				lastOccurrence=count; // store the rank of the last occurrence of the (first) multiple vertex
    		}
    		count++;
    	}
    	
    	count=0;
    	for(Vertex x: this.vertices) {
    		if(count<=firstOccurrence) // add all elements up the first occurrence of the multiple vertex
    			result.add(x);
    		else if(count>lastOccurrence)
    			result.add(x); // add all elements after the last occurrence of the multiple vertex
    		count++;
    	}

    	if(result.getVertices().getFirst()!=u)
    		throw new Error("Error: removing auto-intersections failed (wrong first vertex)");
    	if(result.getVertices().getLast()!=v)
    		throw new Error("Error: removing auto-intersections falied (wrong last vertex)");
    	boolean checkIsSimple=this.isSimplePath(mesh);
    	if(checkIsSimple==true)
    		System.out.println("ok (the auto-intersection has been removed)");
    	else
    		System.out.println("warning: auto-intersection not removed");
    	return result;
    }

    /**
     * Perform the concatenation of two paths p1 and p2 sharing a vertex 'u'. The result is p1+p2 (where p1 is the current path).
     * <p>
     * Remark: the path2 is assumed to have as starting vertex 'u' (the ending vertex of p1)
     * <p>
     * Remark: the vertices are assumed to have an index, between 0..n-1
     * 
     * @param p2  second path to concatenate
     */		   
    public void concatenate(Polyhedron_3 mesh, VertexPath p2){
    	if(p2==null)
    		throw new Error("Error: vertex path not defined");
    	
    	Vertex v1=this.vertices.getLast();
    	Vertex u2=p2.getVertices().getFirst();
    	if(v1!=u2) {
    		System.out.println("Error: the paths to concatenate do not share a vertex");
    		return;
    	}
    	
    	int count=0;
    	for(Vertex v: p2.getVertices()) {
    		if(count!=0) // do not add the shared vertex (the source of path p2)
    			this.vertices.addLast(v);
    		count++;
    	}
    }

	/**
	 * Return the farthest vertex in the mesh from a given vertex path<p>
	 * 
	 * The computation is based on a random sampling of vertices (the result is approximate)
	 * 
	 * @param mesh  input triangulation (triangle mesh)
	 * @param r  size of the random sampling (a number in [0, 1])
	 * @return  the farthest vertex from the path
	 */
	public Vertex getFarthestVertexFromPath(Polyhedron_3<Point_3> mesh, double r) {
		if(r<0 || r>1 || mesh==null)
			throw new Error("Error: wrong input parameters ");
		
		double r1=Math.sqrt(mesh.sizeOfVertices())/mesh.sizeOfVertices();

		System.out.print("Computing farthest point from vertex path ...");
		ArrayList<Vertex<Point_3>> randomSample=new ArrayList<Vertex<Point_3>>(); // random sample of the vertices in the mesh
		for(Vertex<Point_3> v: mesh.vertices) {
			double randomNumber=generator.nextDouble();
			if(randomNumber<=r1)
				randomSample.add(v);
		}

		LinkedList<Vertex> randomSampleOnPath=new LinkedList<Vertex>(); // random sample of the vertices on the path
		for(Vertex<Point_3> v: this.vertices) { // 
			double randomNumber=generator.nextDouble();
			if(randomNumber<=r)
				randomSampleOnPath.add(v);
		}
		
		double maxDistance=0;
		Vertex farthest=mesh.vertices.get(0);
		for(Vertex v: randomSample) {
			double distance=MeshAlgorithms.getDistanceFromSet(mesh, v, randomSampleOnPath);
			if(distance>maxDistance) {
				maxDistance=distance;
				farthest=v;
			}
		}
		
		System.out.println("done");
		return farthest;
	}
    
    /**
     * Return the halfedges on the path (oriented
     * <p>
     * Remark: the vertices are assumed to have an index, between 0..n-1
     * 
     * @return  the list of halfedges on the path
     */		   
    public List<Halfedge> getEdgesOnPath(Polyhedron_3 mesh){
    	if(mesh==null || this.vertices==null)
    		return null;
    	
    	LinkedList<Halfedge> result=new LinkedList<Halfedge>();
    	
    	int count=0;
    	Vertex a=null, b=null;
    	for(Vertex w: this.vertices) { // visit all vertices in the path
    		if(count==0) { // no edge is known at the first iteration
    			a=null;
    			b=w;
    		}
    		else { // start adding edges from the second iteration
    			a=b;
    			b=w;
    			
    			result.add(getNextEdgeOnPath(a, b, null));
    		}
    		count++;
    	}
    	
    	return result;
    }

    /**
     * Given an edge e=(u, v), it returns the next edge e'=(v, w)
     * <p>
     * Warning: edge e'=(v, w) is assumed to exist <p>
     * Remark: the halfedge 'e' is assumed to have destination 'v'
     * 
     * @param v  the source vertex
     * @param w  the destination
     * @return  the halfedge u'=(v, w)
     */		   
    private static Halfedge getNextEdgeOnPath(Vertex v, Vertex w, Halfedge startEdge){
    	if(v==null || w==null)
    		return null;
    	
    	//if(startEdge==null) { // the starting edge is not defined
    		List<Halfedge> neighboringEdges=v.getOutgoingHalfedges(); // get all edges outgoing from vertex 'v'
    		for(Halfedge e: neighboringEdges) {
    			if(e.getVertex()==w)
    				return e;
    		}
    		return null;
    	//}
    	//else { // turn around vertex 'v', until edge (v, w) is reached	
    	//}
    }

}
