package sd.layout;


import java.util.ArrayList;

import Jcg.geometry.Point_3;
import Jcg.polyhedron.Polyhedron_3;
import jdg.graph.AdjacencyListGraph;
import jdg.graph.Node;
import sd.util.SphericalGeometry;
import sd.util.Stat;

/**
 * A spherical graph layout, based on the Alexa's algorithm
 * 
 * @author Luca Castelli Aleardi, Ecole Polytechnique
 * @version fev 2017
 */
public class Alexa extends SphericalLayout {
	// parameters of the alexa algorithm	
	public int iterationCount=0; // count the number of performed iterations
	
	public double [] weight;
	
	/** numeric precision for dealing with rounding errors */
	public double numericPrecision;

	/** for measuring runtime performances */
	private static double totalDuration;

	/**
	 *  @param g  input graph to draw
	 *  @param C  step length
	 *  @param stageEnergy threshold for the convergence of the iterative method
	 */
	public Alexa(AdjacencyListGraph g, Polyhedron_3<Point_3> mesh, double stageEnergy, double precision) {
		System.out.print("Initializing Alexa method...");
		if(g==null) {
			System.out.println("Input graph not defined");
			System.exit(0);
		}
		
		this.g=g;
		this.mesh=mesh;
		int N=g.sizeVertices();
		
		System.out.println("done ("+N+" nodes)");
		System.out.println(this.toString());
		
		//Compute the weight of each point of g
		this.weight=new double[this.g.sizeVertices()];
		for(int i=0;i<this.g.sizeVertices();i++){
			this.weight[i]=1.0/((double)this.g.degree(this.g.getNode(i)));
		}
		
		this.statistics=new Stat("Alexa",false);//Stat object that contains all the statistics we want to work on
		this.stageEnergy=stageEnergy;
		this.booleanContinu=true;
		this.Ek_1=0.;
		
		this.numericPrecision = precision;
		//this.verbosity=SphericalLayout.SHOW_ONLY_STATISTICS;
		this.verbosity=SphericalLayout.SHOW_ONLY_TIMINGS;
	}
	
	/**
	 * Perform one iteration of the Alexa algorithm.
	 */	
	public void computeOneIteration() {
		if(this.verbosity==SphericalLayout.SHOW_ALL)
			System.out.print("Alexa iteration: ");
		
		long startTime=System.nanoTime(), endTime; // for evaluating time performances
		
		double maxDisp=0.;
		Point_3[]vr1=new Point_3[this.g.sizeVertices()];
		for(Node u:this.g.vertices){
			Point_3 p=u.getPoint();
			double c=0.;
			for(Node v:u.neighbors){
				Point_3 q=v.getPoint();
				double d=Math.sqrt((p.x-q.x)*(p.x-q.x)+(p.y-q.y)*(p.y-q.y)+(p.z-q.z)*(p.z-q.z));
				if(d>c){
					c=d;
				}
			}
			/*if (c > this.tol){
				c=1.0/c;
			}
			else{
				c=0.;
			}*/
			c=1.0/c;
			Point_3 qr=new Point_3(0.,0.,0.);
			for (Node v:u.neighbors){
				Point_3 q=v.getPoint();
				double nuv=Math.sqrt((p.x-q.x)*(p.x-q.x)+(p.y-q.y)*(p.y-q.y)+(p.z-q.z)*(p.z-q.z));
				qr.x+=(c*this.weight[u.index]*(p.x-q.x)*nuv);
				qr.y+=(c*this.weight[u.index]*(p.y-q.y)*nuv);
				qr.z+=(c*this.weight[u.index]*(p.z-q.z)*nuv);
			}
			double e=Math.sqrt((p.x-qr.x)*(p.x-qr.x)+(p.y-qr.y)*(p.y-qr.y)+(p.z-qr.z)*(p.z-qr.z));
			
			//maxDisp=Math.max(qr.distanceFrom(new Point_3(0., 0., 0.)).doubleValue(), maxDisp);
			
			if (e > this.numericPrecision){
				vr1[u.index]=new Point_3((p.x-qr.x)/e,(p.y-qr.y)/e,(p.z-qr.z)/e);
			}
			else{
				vr1[u.index]=u.getPoint();
				System.err.println("fixed vertex");
			}
		}
		
		for(Node u:this.g.vertices){
			u.p=vr1[u.index];
		}
		
		// evaluate time performances
    	endTime=System.nanoTime();
        double duration=(double)(endTime-startTime)/1000000000.;
        this.totalDuration=this.totalDuration+duration;
        
        // show statistics
        if(this.verbosity==SphericalLayout.SHOW_ONLY_TIMINGS || this.verbosity==SphericalLayout.SHOW_ALL) {
        	System.out.print("iteration "+this.iterationCount+" done ("+duration+" seconds)");
        	System.out.println("[total duration: "+this.totalDuration+" seconds]");
        }
        if(this.verbosity==SphericalLayout.SHOW_ONLY_STATISTICS || this.verbosity==SphericalLayout.SHOW_ALL) {
        	int skipIteration=5; // show only one iteration every 'skipIteration'
        	if(this.iterationCount%skipIteration==0)
        		System.out.println(""+this.iterationCount+"\t"+this.getStats());
        	//System.out.println("maxDisp= "+maxDisp);
        }
        
        //if(maxDisp<this.residualError)
        //	System.out.println("convergence: true");
        
        // update and save statistics
        /*if(this.iterationCount==0){
        	this.statistics.time.add(duration);
        }
        else{
        	this.statistics.time.add(duration+this.statistics.time.get(this.iterationCount-1));
        }
        double E=computeTotalEnergy();
        System.out.println("\t Potential energy: "+this.currentEnergy);
        this.statistics.energy.add(E);*/
        
        //this.statistics.edgeLength.add(this.computeEdgeLengthAesthetic());
        //this.statistics.angleMean.add(this.computeAngleAesthetic());
        //this.statistics.ratioArea.add(this.ratioArea());
        //this.statistics.crossing.add(this.numberOfCrossing());
        
		this.iterationCount++; // increase counter (to count the number of performed iterations)
	}
	
	/**
	 * Compute and return the potential energy for the Tutte layout
	 */	
	public double computeTotalEnergy() {
		double energy=0.0;
		for(int i=0;i<this.g.sizeVertices();i++) {
			Node u=this.g.vertices.get(i);
			Point_3 p=u.getPoint();
			for(Node v: u.neighbors) {
				if(v!=null) {
					Point_3 q=v.getPoint();
					//double dist=SphericalGeometry.geodesicDistance(p,v.getPoint());
					double dist=Math.sqrt((p.x-q.x)*(p.x-q.x)+(p.y-q.y)*(p.y-q.y)+(p.z-q.z)*(p.z-q.z)); // use euclidean distance
					energy+=dist*dist;
				}
			}
		}
		energy=energy/2.0;
		
		this.currentEnergy=energy; // update the energy
		
		return energy;
	}
	
	/**
	 * Perform 'k' iteration of the Alexa algorithm.
	 * 
	 * @param iterations  number of iterations to perform
	 */	
	public void computeLayout(int iterations) {
		for(int i=0;i<iterations;i++)
			this.computeOneIteration();
	}
	
	/**
	 * Perform iterations until convergence <p>
	 * <p>
	 * Remark: as explained in Alexa 2000 (section 2.1 and 2.2), the algorithm
	 * continues while <p>
	 *  max|| p_i || <= tolerance <p>
	 *  
	 *  where 'p_i' is defined in Section 2.1 (p_i is the displacement of vertex v_i, before projecting on the sphere)
	 * 
	 * @param tolerance  numeric tolerance
	 */	
	public void computeLayout(double tolerance) {
		tolerance = this.stageEnergy;

		Point_3[] oldPos = this.storeVertexLocations();
		while (this.iterationCount == 0 || (this.iterationCount < this.maxIterations && getDistance(this.g.listOfPoints(), oldPos)>tolerance)){
			System.out.print("Performing iteration (Alexa): ");
			
			long startTime=System.nanoTime(), endTime; // for evaluating time performances
			
			Point_3[]vr1=new Point_3[this.g.sizeVertices()];
			for(Node u:this.g.vertices){
				oldPos[u.index] = u.p;
				Point_3 p=u.getPoint();
				double c=0.;
				for(Node v:u.neighbors){
					Point_3 q=v.getPoint();
					double d=Math.sqrt((p.x-q.x)*(p.x-q.x)+(p.y-q.y)*(p.y-q.y)+(p.z-q.z)*(p.z-q.z));
					if(d>c){
						c=d;
					}
				}
				/*if (c > this.tol){
					c=1.0/c;
				}
				else{
					c=0.;
				}*/
				c=1.0/c;
				Point_3 qr=new Point_3(0.,0.,0.);
				for (Node v:u.neighbors){
					Point_3 q=v.getPoint();
					double nuv=Math.sqrt((p.x-q.x)*(p.x-q.x)+(p.y-q.y)*(p.y-q.y)+(p.z-q.z)*(p.z-q.z));
					qr.x+=(c*this.weight[u.index]*(p.x-q.x)*nuv);
					qr.y+=(c*this.weight[u.index]*(p.y-q.y)*nuv);
					qr.z+=(c*this.weight[u.index]*(p.z-q.z)*nuv);
				}
				double e=Math.sqrt((p.x-qr.x)*(p.x-qr.x)+(p.y-qr.y)*(p.y-qr.y)+(p.z-qr.z)*(p.z-qr.z));
				
				if (e > this.numericPrecision){
					vr1[u.index]=new Point_3((p.x-qr.x)/e,(p.y-qr.y)/e,(p.z-qr.z)/e);
				}
				else{
					vr1[u.index]=u.getPoint();
					System.err.println("fixed vertex");
				}
			}
			for(Node u:this.g.vertices){
				u.p=vr1[u.index];
			}
			
			// evaluate time performances
	    	endTime=System.nanoTime();
	        double duration=(double)(endTime-startTime)/1000000000.;
	        System.out.println("iteration "+this.iterationCount+" done ("+duration+" seconds)");
	        
	        /**
	         * Update of the statistics
	         */
	        if(this.iterationCount==0){
	        	this.statistics.time.add(duration);
	        }
	        else{
	        	this.statistics.time.add(duration+this.statistics.time.get(this.iterationCount-1));
	        }
	        double E=computeTotalEnergy();
	        System.out.println("\t Potential energy: "+this.currentEnergy);
	        this.statistics.energy.add(E);
	        //this.statistics.edgeLength.add(this.computeEdgeLengthAesthetic());
	        //this.statistics.angleMean.add(this.computeAngleAesthetic());
	        //this.statistics.ratioArea.add(this.ratioArea());
	        //this.statistics.crossing.add(this.numberOfCrossing());
	        
			this.iterationCount++; // increase counter (to count the number of performed iterations)
		}	
	}
	
	public String toString() {
		String result="Alexa algorihm:\n";
		result=result+"current energy= "+this.currentEnergy;

		return result;
	}

	/**
	 * Get the name of the layout method
	 */
	public String getName() {
		return "Alexa";
	}
	
	public double getDistance(ArrayList<Point_3>graph1,Point_3[]graph2){
		double distance = 0.;
		int len = graph2.length;
		for (int i=0;i<len;i++){
			distance += (graph1.get(i).x-graph2[i].x)*(graph1.get(i).x-graph2[i].x)+(graph1.get(i).y-graph2[i].y)*(graph1.get(i).y-graph2[i].y)+(graph1.get(i).z-graph2[i].z)*(graph1.get(i).z-graph2[i].z);
		}
		return Math.sqrt(distance)/((double)len);
	}

}