package sd.layout;

import jdg.graph.AdjacencyListGraph;
import jdg.graph.Node;
import sd.util.SphericalGeometry;

import java.util.LinkedList;

import Jcg.geometry.Point_3;
import Jcg.geometry.Vector_3;

/**
 * A class implementing the force directed algorithm by Fruchterman and Reingold (1991)
 * 
 * @author Luca Castelli Aleardi, Ecole Polytechnique
 * @version fev 2017
 */
public class FR91Layout extends Layout3D implements SpringEmbedder {
	// parameters of the spring model (FR91)
	/** force model: it contains 3 parameters: 'k', 'C', 'p' */
	public SpringElectricalForce force; // force model

	/** area of the drawing region (2d or 3d) */
	public double area;
	
	// parameters of the "simulated annealing" process
	/** initial temperature */
	protected double temperature;
	/** minimal temperature (strictly positive) */
	protected double minTemperature;
	/** cooling constant: temperature decreases (e.g. linearly) at each iteration */
	protected double coolingConstant;
	/** decide whether to make use of simulated annealing*/
	protected boolean useCooling;

	/** step length: this parameters is used to control displacements, according to adaptive cooling system described in Hu2006 */
	protected double step; // by default step=1, in the original algorithm FR91

	/** decide whether to make use of weights for computing repulsive forces */
	protected boolean useVertexWeights;

	/** 
	 * Decide whether to move all vertices or only one vertex: <p>
	 * -) if TRUE: move all points all together, after computing all forces and displacements <p>
	 * -) if FALSE: move one vertex at time (computing forces and displacement only for that vertex
	 */
	public boolean moveAllVerticesAtOnce=true; // by default, move all vertices all together (as in the original FR91 layout algorithm)
	
	public int iterationCount=0; // count the number of performed iterations
	private int countRepulsive=0; // count the number of computed repulsive forces (to measure time performances)
	
	/** total runtime cost */
	protected double totalDuration;

	public FR91Layout(){
		
	}
	
	/**
	 * Initialize the parameters of the force-directed layout, as in the original FR91 algorithm
	 * 
	 *  @param g  input graph to draw
	 *  @param C  a real value defining the relative strength of repulsive forces
	 *  @param w  width of the drawing area
	 *  @param h  height of the drawing area
	 *  @param d  depth of the drawing area (d=0 if computations are in 2D)
	 */
	public FR91Layout(AdjacencyListGraph g, double w, double h, double d) {
		System.out.print("Initializing force-directed method: Fruchterman-Reingold 91...");
		if(g==null) {
			System.err.println("Warning: input graph not defined");
			System.exit(0);
		}
		this.g=g;
		int N=g.sizeVertices();
		
		this.w=w;
		this.h=h;
		this.d=d;
		
		double k;
		
		// set the parameters of the algorithm FR91
		if(d>0) {
			this.area=w*h*d; // 3d case
			k=2*Math.pow(area/N, 1./3.); 
		}
		else {
			this.area=w*h; // 2d case
			k=Math.sqrt(area/N); // standard choice adopted in FR91
		}
		
		this.step=1.0; // as in the original FR91 layout algorithm
		this.force=new SpringElectricalForce(k); // gravity constant is 0 (not defined in FR91)

		this.temperature=w/2; // the temperature is a fraction of the width of the drawing area
		this.minTemperature=temperature/50;
		this.coolingConstant=0.98;
		this.useCooling=true; // use cooling system by default
		
		this.useVertexWeights=false; // by default repulsive forces are computed in the unweighted case (as in FR91)
		
		System.out.println("done ("+N+" nodes)");
		System.out.println(this.toString());
	}

	/**
	 * Initialize the parameters of the force-directed layout (general version)
	 * 
	 *  @param g  input graph to draw
	 *  @param w  width of the drawing area
	 *  @param h  height of the drawing area
	 *  @param d  depth of the drawing area (d=0 if computations are in 2D)
	 *  @param step  initial step length (for the adaptive cooling system by Hu2006)
	 *  @param k  the natural spring length
	 *  @param C  a real value defining the relative strength of repulsive forces
	 *  @param minTemperature  minimal value of the temperature
	 *  @param coolingConstant  value of the cooling constant
	 */
	public FR91Layout(AdjacencyListGraph g, double w, double h, double d, double k, double step, double C, double minTemperature, double coolingConstant) {
		System.out.print("Initializing force-directed method: Fruchterman-Reingold 91...");
		if(g==null) {
			System.err.println("Warning: input graph not defined");
			System.exit(0);
		}
		this.g=g;
		int N=g.sizeVertices();
		
		this.w=w;
		this.h=h;
		this.d=d;
		
		// set the parameters of the algorithm FR91
		if(d>0)
			this.area=w*h*d;
		else
			this.area=w*h;
		
		this.step=step;
		this.force=new SpringElectricalForce(k, 0., C, 1); // gravity constant is 0 (not defined)

		this.temperature=w/2; // the temperature is a fraction of the width of the drawing area
		this.minTemperature=minTemperature;
		this.coolingConstant=coolingConstant;
		this.useCooling=true; // use cooling system by default
		
		System.out.println("done ("+N+" nodes)");
		System.out.println(this.toString());
	}
	
	/** Set the value of the parameter 'C', the relative strength of the repulsive and attractive forces (as used in Hu 2006) */
	public void setC(double c) {
		if(this.force!=null)
			this.force.setC(c);
	}

	/** Get the value of the parameter 'C', the relative strength of the repulsive and attractive forces (as used in Hu 2006) */
	public double getC() {
		if(this.force!=null)
			return this.force.C;
		else {
			System.err.println("Warning: the force model is not defined");
			return 0.;
		}
	}

	/** Set the value of the parameter 'K', the natural spring length */
	public void setK(double k) {
		if(this.force!=null)
			this.force.setK(k);
	}

	/** Get the value of the parameter 'K', the natural spring length */
	public double getK() {
		if(this.force!=null)
			return this.force.k;
		else {
			System.err.println("Warning: the force model is not defined");
			return 0.;
		}
	}

	/** Set the value of the step length (as described in Hu 2006) */
	public void setStep(double step) {
		this.step=step;
	}

	/** Get the value of the step length (as used in Hu 2006) */
	public double getStep() {
		return this.step;
	}

	/** Enable cooling process */	
	public void enableCooling() {
		this.useCooling=true;
	}

	/** Enable cooling process */	
	public void disableCooling() {
		this.useCooling=false;
	}
	
	/** Get the current value of the temperature */
	public double getTemperature() {
		return this.temperature;
	}

	/** Get the value of the minimal temperature */
	public double getMinimalTemperature() {
		return this.minTemperature;
	}

	/** Get the value of the cooling constant */
	public double getCoolingConstant() {
		return this.coolingConstant;
	}

	/** Set the initial value of the temperature */
	public void setInitialTemperature(double t) {
		this.temperature=t;
	}

	/** Set the initial value of the cooling constant */
	public void setCoolingConstant(double c) {
		this.coolingConstant=c;
	}

	/** Set the minimal value of the temperature */
	public void setMinimalTemperature(double min) {
		this.minTemperature=min;
	}
	
	/** decide to move all vertices all together, after computing all mutual forces (as in the original FR91 layout algorithm) */	
	public void moveAllVerticesAtOnce() {
		this.moveAllVerticesAtOnce=true;
	}

	/** decide to move only one vertex at a time, computing forces and displacement only for that vertex (as in Walshaw 2003) */	
	public void moveOnlyOneVertex() {
		this.moveAllVerticesAtOnce=false;
	}
	
	/** use vertex weights for computing repulsive forces (as in Walshaw 2003) */	
	public void useVertexWeights() {
		this.useVertexWeights=true;
	}

	/** do not use vertex weights for computing repulsive forces (as in FR91) */	
	public void noVertexWeights() {
		this.useVertexWeights=false;
	}

	/**
	 * Compute the (intensity of the) attractive force between two nodes at a given distance
	 * 
	 * @param distance  distance between two nodes
	 */	
	public double attractiveForce(double distance) {
		return this.force.attractiveForce(distance);
	}
	
	/**
	 * Compute the (intensity of the) repulsive force between two nodes at a given distance
	 * 
	 * @param distance  distance between two nodes
	 */	
	public double repulsiveForce(double distance) {
		countRepulsive++;
		return this.force.repulsiveForce(distance);
	}
	
	/**
	 * Return the weighted version of the repulsive force acting on vertex 'u'
	 * 
	 * @param distance  distance between two nodes
	 * @param weight  the weight of vertex 'u'
	 */	
	public double repulsiveForce(double distance, double weight) {
		return this.repulsiveForce(distance)*weight;
	}

	/**
	 * Perform one iteration of the Force-Directed algorithm FR91. <p>
	 * Positions of vertices are updated according to their mutual attractive and repulsive forces.<p>
	 */	
	public void computeOneIteration() {		
		//System.out.print("Performing iteration: "+this.iterationCount);
		/*if(this.verbosity==Layout3D.SHOW_ALL) {
		if(useCooling==true)
			System.out.println("(temperature: "+this.temperature+")");
		else
			System.out.println("(no simulated annealing)");
		}*/
				
		long startTime=System.nanoTime(), endTime; // for evaluating time performances
		
		if(this.moveAllVerticesAtOnce==true) { // all vertices are moves at the same time, after computing all forces
			
			// first step: for each vertex compute the displacements due to attractive and repulsive forces
			Vector_3[] attractiveDisp=this.computeAllAttractiveForces(); // displacement due to attractive forces between neighbors
			Vector_3[] repulsiveDisp=this.computeAllRepulsiveForces(); // displacement due to repulsive forces between all nodes

			// second step: compute the total displacements and move all nodes at the same time to their new locations
			for(int i=0;i<this.g.sizeVertices();i++) {
				Node u=this.g.vertices.get(i);
				if(u.degree()>0) { // only non isolated nodes are involved in the computation
					Point_3 p=u.getPoint(); // geometric position of node 'u'

					Vector_3 disp=attractiveDisp[i].sum(repulsiveDisp[i]); // total displacement for vertex 'u' 
					double norm=Math.sqrt(disp.squaredLength().doubleValue()); // length of the displacement of vertex 'u'

					// use temperature to limit the displacement
					double min=Math.min(norm, this.temperature);
					disp=disp.multiplyByScalar(this.step*min/norm); // use the variable step length (as in Hu2006)

					p.translateOf(disp);
				}
			}
			
		}
		else { // compute forces and displacements for one vertex at a time (as suggested in Walshaw 2003)
			
			for(int i=0;i<this.g.sizeVertices();i++) {
				Node u=this.g.vertices.get(i);
				if(u.degree()>0) { // only non isolated nodes are involved in the computation
					Point_3 p=u.getPoint(); // geometric position of node 'u'

					Vector_3 disp=this.computeAttractiveForce(u).sum(this.computeRepulsiveForce(u)); // total displacement for vertex 'u' 
					double norm=Math.sqrt(disp.squaredLength().doubleValue()); // length of the displacement of vertex 'u'

					// use temperature to limit the displacement
					double min=Math.min(norm, this.temperature);
					disp=disp.multiplyByScalar(this.step*min/norm); // use the variable step length (as in Hu2006)

					p.translateOf(disp); // move vertex 'u' to the new location
				}
			}

		}
		
		// evaluate time performances
    	endTime=System.nanoTime();
        double duration=(double)(endTime-startTime)/1000000000.;
        this.totalDuration=this.totalDuration+duration; // sum the timings over all iterations
        
        //if(this.verbosity==Layout3D.SHOW_ONLY_TIMINGS) {
        	System.out.print("FR91: iteration "+this.iterationCount+" done ("+duration+" seconds)");
        	System.out.println("[total duration: "+this.totalDuration+" seconds]");
        //}
        /*else if(this.verbosity==Layout3D.SHOW_ALL) {
        	System.out.print("it "+this.iterationCount+" ");
        	SphericalLayout.countTriangleCollisions(this.mesh, this.g);
        }*/

		if(useCooling==true) // check whether "simulated annealing" is used
			this.cooling(); // update temperature
		
		this.iterationCount++; // increase counter (to count the number of performed iterations)
	}
	
	/**
	 * Perform 'k' iteration of the Force-Directed algorithm.
	 * Positions of vertices are updated according to their mutual attractive and repulsive forces.
	 * 
	 * @param iterations  number of iterations to perform
	 */	
	public void computeLayout(int iterations) {
		for(int i=0;i<iterations;i++)
			this.computeOneIteration();
	}
	
	/**
	 * Perform layout computation until convergence
	 * 
	 * @param tolerance  numeric tolerance
	 */	
	public void computeLayout(double tolerance) {
		long startTime=System.nanoTime(), endTime; // for evaluating time performances

		boolean converged=false;
		
		while(converged==false && this.iterationCount<this.maxIterations) {
			converged=true;
			double maxDelta=0.;
			
			Point_3[] oldPos=this.storeVertexLocations(); // store the old locations of vertices (before performing the iteration)
			
			for(Node u: this.g.vertices) {
				if(u.degree()>0) { // only non isolated nodes are involved in the computation
					Point_3 p=u.getPoint(); // geometric position of node 'u'

					Vector_3 disp=this.computeAttractiveForce(u).sum(this.computeRepulsiveForce(u)); // total displacement for vertex 'u' 
					double norm=Math.sqrt(disp.squaredLength().doubleValue()); // length of the displacement of vertex 'u'

					// use temperature to limit the displacement
					double min=Math.min(norm, this.temperature);
					disp=disp.multiplyByScalar(min/norm);

					p.translateOf(disp); // move vertex 'u' to the new location
					
					double delta=p.distanceFrom(oldPos[u.index]).doubleValue();
					maxDelta=Math.max(maxDelta, delta);
					if(delta>tolerance*this.getK())
						converged=false;
				}
			}
			
			// evaluate time performances
	    	endTime=System.nanoTime();
	        double duration=(double)(endTime-startTime)/1000000000.;
			
			if(useCooling==true) // check whether "simulated annealing" is used
				this.cooling(); // update temperature
			
			this.iterationCount++; // increase counter (to count the number of performed iterations)

	        System.out.println("iteration "+this.iterationCount+", delta="+maxDelta+" done ("+duration+" seconds)");
		}
	}
	
	/**
	 * Compute the displacement of vertex 'u', due to repulsive forces (of all nodes)
	 * 
	 * @param u  the vertex to which repulsive forces are applied
	 * @return 'displacement' a 3d vector storing the displacement of vertex 'u'
	 */	
	public Vector_3 computeRepulsiveForce(Node u) {
		Vector_3 displacement=new Vector_3(0., 0., 0.);
		Vector_3 delta;
		Point_3 p=u.getPoint();
		
		if(this.useVertexWeights==false) { // standard implementation: repulsive forces are computed in the unweighted case
			for(Node v: this.g.vertices) {
				if(v!=null && u!=v && v.degree()>=0) {
					delta=new Vector_3(v.getPoint(),p);
					double norm=Math.sqrt(delta.squaredLength().doubleValue());
					displacement=displacement.sum(delta.multiplyByScalar(this.repulsiveForce(norm)));
				}
			}
		}
		else { // weighted version: repulsive forces are computed using vertex weights (as multiplicative factor)
			for(Node v: this.g.vertices) {
				if(v!=null && u!=v && v.degree()>=0) {
					delta=new Vector_3(v.getPoint(),p);
					double norm=Math.sqrt(delta.squaredLength().doubleValue());
					displacement=displacement.sum(delta.multiplyByScalar(this.repulsiveForce(norm, u.weight)));
				}
			}
		}


		return displacement;
	}
	
	/**
	 * Compute, for each vertex, the displacement due to repulsive forces (between all nodes)
	 * 
	 * @return a vector v[]: v[i] stores the geometric displacement of the i-th node
	 */	
	public Vector_3[] computeAllRepulsiveForces() {
		Vector_3[] repulsiveDisp=new Vector_3[this.g.sizeVertices()];
		for(int i=0;i<this.g.vertices.size();i++)
			repulsiveDisp[i]=new Vector_3(0., 0., 0.);
		
		long startTime=System.nanoTime(), endTime; // for evaluating time performances
		
		int countVertices=0;
		for(Node u: this.g.vertices) {
			repulsiveDisp[countVertices]=this.computeRepulsiveForce(u); // 'slow' version (forces are computed twice: for each vertex)
			countVertices++;
		}
		
		if(this.verbosity==Layout3D.SHOW_ONLY_TIMINGS) {
			endTime=System.nanoTime();
			double duration=(double)(endTime-startTime)/1000000000.;
			System.out.print("\ttimings: "+duration+" seconds (repulsive force)");
			System.out.println(" ["+this.countRepulsive+" forces computed]");
		}
        this.countRepulsive=0;
        
        return repulsiveDisp;
	}
	
	/**
	 * Compute for the entire system the energy corresponding to the attractive forces
	 */
	public double computeAttractiveEnergy(){
		double E=0.;
		for(Node u:this.g.vertices){
			Point_3 p=u.getPoint();
			for(Node v:u.neighbors){
				Point_3 q=v.getPoint();
				double dist=p.distanceFrom(q).doubleValue();
				E+=this.force.attractiveEnergy(dist);
			}
		}
		return E;
	}

	/**
	 * Compute for the entire system the energy corresponding to the repulsive forces
	 */
	public double computeRepulsiveEnergy(){
		double E=0.;
		for(Node u:this.g.vertices){
			Point_3 p=u.getPoint();
			for(Node v:this.g.vertices){
				if(!v.equals(u)){
					Point_3 q=v.getPoint();
					double dist=p.distanceFrom(q).doubleValue();
					E+=this.force.repulsiveEnergy(dist);
				}
			}
		}
		return E/2.;
	}

	/**
	 * Compute and return the potential energy for the force-directed layout
	 * (equal to the difference between the attractive energy and the repulsive energy)
	 */	
	public double computeTotalEnergy() {
		System.out.print("Computing potential energy of the layout (FR91): ");
		double E=this.computeAttractiveEnergy()-this.computeRepulsiveEnergy();
		return E;
	}

	/**
	 * Compute the displacement of vertex 'u', due to the attractive forces of its neighbors
	 * 
	 * @param u  the vertex to which attractive forces are applied
	 * @return 'disp' a 3d vector storing the displacement of vertex 'u'
	 */	
	public Vector_3 computeAttractiveForce(Node u) {
		Vector_3 displacement=new Vector_3(0., 0., 0.);
		Vector_3 delta;
		Point_3 p=u.getPoint();
		for(Node v: u.neighbors) {
			if(v!=null) {
				delta=new Vector_3(p, v.getPoint());
				double norm=Math.sqrt(delta.squaredLength().doubleValue());
				displacement=displacement.sum(delta.multiplyByScalar(this.attractiveForce(norm)));
			}
		}
		return displacement;
	}
	
	/**
	 * Compute, for each vertex, the displacement due to attractive forces (between neighboring nodes)
	 * 
	 * @return a vector v[]: v[i] stores the geometric displacement of the i-th node
	 */	
	public Vector_3[] computeAllAttractiveForces() {
		Vector_3[] attractiveDisp=new Vector_3[this.g.sizeVertices()];
		long startTime=System.nanoTime(), endTime; // for evaluating time performances
		
		int i=0;
		for(Node u: this.g.vertices) {
			attractiveDisp[i]=computeAttractiveForce(u);
			i++;
		}
		
		if(this.verbosity==Layout3D.SHOW_ONLY_TIMINGS) {
			endTime=System.nanoTime();
			double duration=(double)(endTime-startTime)/1000000000.;
			System.out.println("\ttimings: "+duration+" seconds (attractive force)");
		}
        
        return attractiveDisp;
	}
	
	/**
	 * Cooling system: the temperature decreases at each iteration (following the method adopted in Walshaw)
	 * 
	 * Remark: the temperature is assumed to remain strictly positive (>=minTemperature)
	 */	
	protected void cooling() {
		//this.temperature=Math.max(this.temperature-coolingConstant, minTemperature);
		this.temperature=Math.max(this.temperature*this.coolingConstant, minTemperature);
	}
	
	public String toString() {
		String result="(slow) force-directed algorihm: Fruchterman Reingold\n";
		result=result+"\t area= "+w+" x "+h+" x "+d+"\n";
		result=result+"\t k= "+this.force.k+"\n";
		result=result+"\t C= "+this.force.C+"\n";
		result=result+"\t temperature= "+this.temperature+"\n";
		result=result+"\t minimal temperature= "+this.minTemperature+"\n";
		result=result+"\t cooling constant= "+this.coolingConstant+"\n";
		result=result+"\t moving all vertices at once= "+this.moveAllVerticesAtOnce+"\n";
		
		return result;
	}
	
	/**
	 * Get the name of the layout method
	 */
	public String getName() {
		return "standard FR91 layout";
	}

}
