package sd.layout;

import Jcg.geometry.Point_3;
import Jcg.geometry.Vector_3;
import jdg.graph.AdjacencyListGraph;
import jdg.graph.Node;

import sd.util.Octree;

/**
 * A fast implementation of the force directed algorithm by Fruchterman and Reingold (1991):
 * repulsive forces are approximated using Octrees
 * 
 * @author Luca Castelli Aleardi, Ecole Polytechnique
 * @version fev 2018
 */
public class FastFR91Layout extends FR91Layout {

	private int countRepulsive=0; // count the number of computed repulsive forces (to measure time performances)
	
	/** total runtime cost */
	protected double totalDuration;

	/**
	 * Initialize the parameters of the force-directed layout, as in the original FR91 algorithm
	 * 
	 *  @param g  input graph to draw
	 *  @param C  a real value defining the relative strength of repulsive forces
	 *  @param w  width of the drawing area
	 *  @param h  height of the drawing area
	 *  @param d  depth of the drawing area (d=0 if computations are in 2D)
	 */
	public FastFR91Layout(AdjacencyListGraph g, double w, double h, double d) {
		System.out.print("Initializing fast force-directed method: Fruchterman-Reingold 91 (with Octrees)...");
		if(g==null) {
			System.err.println("Warning: input graph not defined");
			System.exit(0);
		}
		this.g=g;
		int N=g.sizeVertices();
		
		this.w=w;
		this.h=h;
		this.d=d;
		
		double k;
		
		// set the parameters of the algorithm FR91
		if(d>0) {
			this.area=w*h*d; // 3d case
			k=2*Math.pow(area/N, 1./3.); 
		}
		else {
			this.area=w*h; // 2d case
			k=Math.sqrt(area/N); // standard choice adopted in FR91
		}
		
		this.step=1.0; // as in the original FR91 layout algorithm
		this.force=new SpringElectricalForce(k); // gravity constant is 0 (not defined in FR91)

		this.temperature=w/2; // the temperature is a fraction of the width of the drawing area
		this.minTemperature=temperature/50;
		this.coolingConstant=0.98;
		this.useCooling=true; // use cooling system by default
		
		this.useVertexWeights=false; // by default repulsive forces are computed in the unweighted case (as in FR91)
		
		System.out.println("done ("+N+" nodes)");
		System.out.println(this.toString());
	}

	/**
	 * Initialize the parameters of the force-directed layout (general version)
	 * 
	 *  @param g  input graph to draw
	 *  @param w  width of the drawing area
	 *  @param h  height of the drawing area
	 *  @param d  depth of the drawing area (d=0 if computations are in 2D)
	 *  @param step  initial step length (for the adaptive cooling system by Hu2006)
	 *  @param k  the natural spring length
	 *  @param C  a real value defining the relative strength of repulsive forces
	 *  @param minTemperature  minimal value of the temperature
	 *  @param coolingConstant  value of the cooling constant
	 */
	public FastFR91Layout(AdjacencyListGraph g, double w, double h, double d, double k, double step, double C, double minTemperature, double coolingConstant) {
		System.out.print("Initializing fast force-directed method: Fruchterman-Reingold 91 (with Octrees)...");
		if(g==null) {
			System.err.println("Warning: input graph not defined");
			System.exit(0);
		}
		this.g=g;
		int N=g.sizeVertices();
		
		this.w=w;
		this.h=h;
		this.d=d;
		
		// set the parameters of the algorithm FR91
		if(d>0)
			this.area=w*h*d;
		else
			this.area=w*h;
		
		this.step=step;
		this.force=new SpringElectricalForce(k, 0., C, 1); // gravity constant is 0 (not defined)

		this.temperature=w/2; // the temperature is a fraction of the width of the drawing area
		this.minTemperature=minTemperature;
		this.coolingConstant=coolingConstant;
		this.useCooling=true; // use cooling system by default
		
		System.out.println("done ("+N+" nodes)");
		System.out.println(this.toString());
	}

	/**
	 * Compute the (intensity of the) attractive force between two nodes at a given distance
	 * 
	 * @param distance  distance between two nodes
	 */	
	public double attractiveForce(double distance) {
		return this.force.attractiveForce(distance);
	}
	
	/**
	 * Compute the (intensity of the) repulsive force between two nodes at a given distance
	 * 
	 * @param distance  distance between two nodes
	 */	
	public double repulsiveForce(double distance) {
		countRepulsive++;
		return this.force.repulsiveForce(distance);
	}

	/**
	 * Perform one iteration of the Force-Directed algorithm.
	 * Positions of vertices are updated according to their mutual attractive and repulsive forces.
	 */	
	public void computeOneIteration() {
		if(this.verbosity==Layout3D.SHOW_ALL) {
		if(useCooling==true)
			System.out.println("(temperature: "+this.temperature+")");
		else
			System.out.println("(no simulated annealing)");
		}

		System.out.print("Performing iteration (fast FR91): "+this.iterationCount);
		long startTime=System.nanoTime(), endTime; // for evaluating time performances
		
		Octree oc=new Octree(g);  //build octree corresponding to graph g

		// evaluate time performances
    	endTime=System.nanoTime();
        double duration=(double)(endTime-startTime)/1000000000.;
        System.out.println("octree initialized ("+duration+" seconds)");

		/*Vector_3[] tetaRepulsive = computeAllRepulsiveForces(oc);  // compute the displacements due to repulsive forces (for each vertex)
		Vector_3[] tetaAttractive = computeAllAttractiveForces();  // compute the displacements due to attractive forces (for each vertex)
		
		Vector_3 teta;
		Number oldNorm;
		int i =0;
		
		// second step: compute the total displacements and move all nodes to their new locations
		for (Node u:g.vertices){
			teta=tetaRepulsive[i].sum(tetaAttractive[i]);  // compute total displacement
			oldNorm=Math.sqrt((double) teta.squaredLength());  //compute norm of teta
			if (oldNorm.doubleValue()!=0){
				u.setPoint(u.p.sum(teta.multiplyByScalar(Math.min(temperature, oldNorm.doubleValue())/oldNorm.doubleValue())));  //modify coordinates of u in accordance with computed forces
			}
			i++;
		}*/
		
		//if(this.moveAllVerticesAtOnce==true) { // all vertices are moves at the same time, after computing all forces
			
			// first step: for each vertex compute the displacements due to attractive and repulsive forces
			Vector_3[] attractiveDisp=this.computeAllAttractiveForces(); // displacement due to attractive forces between neighbors
			Vector_3[] repulsiveDisp=this.computeAllRepulsiveForces(oc); // displacement due to repulsive forces between all nodes

			// second step: compute the total displacements and move all nodes at the same time to their new locations
			for(int i=0;i<this.g.sizeVertices();i++) {
				Node u=this.g.vertices.get(i);
				if(u.degree()>0) { // only non isolated nodes are involved in the computation
					Point_3 p=u.getPoint(); // geometric position of node 'u'

					Vector_3 disp=attractiveDisp[i].sum(repulsiveDisp[i]); // total displacement for vertex 'u' 
					double norm=Math.sqrt(disp.squaredLength().doubleValue()); // length of the displacement of vertex 'u'

					// use temperature to limit the displacement
					double min=Math.min(norm, this.temperature);
					disp=disp.multiplyByScalar(this.step*min/norm); // use the variable step length (as in Hu2006)

					p.translateOf(disp);
				}
			}
			
		//}
		/*else { // compute forces and displacements for one vertex at a time (as suggested in Walshaw 2003)
			
			for(i=0;i<this.g.sizeVertices();i++) {
				Node u=this.g.vertices.get(i);
				if(u.degree()>0) { // only non isolated nodes are involved in the computation
					Point_3 p=u.getPoint(); // geometric position of node 'u'

					Vector_3 disp=this.computeAttractiveForce(u).sum(this.computeRepulsiveForce(u)); // total displacement for vertex 'u' 
					double norm=Math.sqrt(disp.squaredLength().doubleValue()); // length of the displacement of vertex 'u'

					// use temperature to limit the displacement
					double min=Math.min(norm, this.temperature);
					disp=disp.multiplyByScalar(this.step*min/norm); // use the variable step length (as in Hu2006)

					p.translateOf(disp); // move vertex 'u' to the new location
				}
			}

		}*/
		
		// evaluate time performances
    	endTime=System.nanoTime();
        duration=(double)(endTime-startTime)/1000000000.;
        this.totalDuration=this.totalDuration+duration; // sum the timings over all iterations
        
        if(this.verbosity==Layout3D.SHOW_ONLY_TIMINGS || this.verbosity==Layout3D.SHOW_ALL) {
        	System.out.print("FR91: iteration "+this.iterationCount+" done ("+duration+" seconds)");
        	System.out.println("[total duration: "+this.totalDuration+" seconds]");
        }

		if(useCooling==true) // check whether "simulated annealing" is used
			this.cooling(); // update temperature
		
		this.iterationCount++; // increase counter (to count the number of performed iterations)
	}
	
	
	/**
	 * Compute, for each vertex, the displacement due to repulsive forces (between all nodes)
	 * 
	 * @return a vector v[]: v[i] stores the geometric displacement of the i-th node
	 */	
	
	private Vector_3 computeRepulsiveForce(Node v,Octree oc) {
		Vector_3 displacement = new Vector_3(0,0,0);  //initialize the vector of displacement
		Vector_3 delta;  //vector from Node v to another Node of the graph
		double norm;  //norm of delta
		
		//if the barycenter of the Octree is sufficiently "far" from v
		if(((oc.max.x-oc.min.x) < oc.getTheta() * v.p.distanceFrom(oc.bary).doubleValue())){
			//if v is in the Octree : we compute the barycenter of the Octree without taking v into account and compute the displacement of v due to this barycenter
			if(!Octree.isInCube(v.p,oc.min,oc.max)){
				delta = new Vector_3(oc.bary, v.p);
				//delta = delta.divisionByScalar(oc.nb);
				norm= Math.sqrt(delta.squaredLength().doubleValue());
				//displacement=displacement.sum(delta.divisionByScalar(norm).multiplyByScalar(repulsiveForce(norm)));
				displacement=displacement.sum(delta.multiplyByScalar(this.repulsiveForce(norm))); // FR91 by Luca
			} 
			//if v is not in the Octree we simply compute the displacement of v due to the points in the Octree (through its barycenter)
			else{
				double n=(double) oc.nb;
				Point_3 newBary=Point_3.linearCombination(new Point_3[]{oc.bary, v.p},new Number[]{n/(n-1.),-1./n});
				delta = new Vector_3(newBary, v.p);
				//delta = delta.divisionByScalar(oc.nb);
				norm= Math.sqrt(delta.squaredLength().doubleValue());
				//displacement=displacement.sum(delta.divisionByScalar(norm).multiplyByScalar(repulsiveForce(norm)));
				displacement=displacement.sum(delta.multiplyByScalar(this.repulsiveForce(norm))); // FR91 by Luca
			}	
		} 
		//if the barycenter is not "far enough" of v
		else {
			//if the Octree is a leaf : we compute separately each repulsive force due to each node in the Octree 
			if(oc.isLeaf()){
				for(Node u:oc.l){
					if(u.p!=v.p){
						delta = new Vector_3(u.p, v.p);
						norm= Math.sqrt(delta.squaredLength().doubleValue());
						//displacement=displacement.sum(delta.divisionByScalar(norm).multiplyByScalar(repulsiveForce(norm)));
						displacement=displacement.sum(delta.multiplyByScalar(this.repulsiveForce(norm))); // FR91 by Luca
					}	
				}
			} 
			//if the Octree is not a leaf, we compute each repulsive force due to the Octrees in children
			else {
				for(int i =0;i<8;i++){
					if(oc.children[i]!=null){
						displacement=displacement.sum(computeRepulsiveForce(v,oc.children[i]));
					}
				}
			}
		}
		
		return displacement;
	}

	/**
	 * Compute, for each vertex, the displacement due to repulsive forces (between all nodes)
	 * 
	 * @return a vector v[]: v[i] stores the geometric displacement of the i-th node
	 */	
	private Vector_3[] computeAllRepulsiveForces(Octree oc) {
		Vector_3[] teta = new Vector_3[g.sizeVertices()];
		int i=0;
		for (Node v : g.vertices){  						//oc.l : List of the vertices in o
			teta[i]=computeRepulsiveForce(v,oc);
			i++;
		}
		return teta;
	}
	
	public String toString() {
		String result="(fast) force-directed algorihm: Fruchterman Reingold\n";
		result=result+"\t area= "+w+" x "+h+" x "+d+"\n";
		result=result+"\t k= "+this.force.k+"\n";
		result=result+"\t C= "+this.force.C+"\n";
		result=result+"\t temperature= "+this.temperature+"\n";
		result=result+"\t minimal temperature= "+this.minTemperature+"\n";
		result=result+"\t cooling constant= "+this.coolingConstant+"\n";
		result=result+"\t moving all vertices at once= "+this.moveAllVerticesAtOnce+"\n";
		
		return result;
	}
	
	/**
	 * Get the name of the layout method
	 */
	public String getName() {
		return "Fast FR91 layout (with Octrees)";
	}

}
