package sd.layout;


import java.util.ArrayList;

import Jcg.geometry.Point_3;
import Jcg.geometry.Vector_3;
import Jcg.polyhedron.Polyhedron_3;
import jdg.graph.AdjacencyListGraph;
import jdg.graph.Node;
import sd.util.SphericalGeometry;
import sd.util.Stat;
/**
 * A spherical graph layout, based on the iterative relaxation algorithm
 * 
 * @author Luca Castelli Aleardi, Ecole Polytechnique
 * @version fev 2017
 */
public class IterativeRelaxation extends SphericalLayout {
	// parameters of the iterative relaxation algorithm	
	public int iterationCount=0; // count the number of performed iterations
	
	public double[] weight;
	public double[] alpha;
	
	public double residual=0.;
	
	/** numeric precision for dealing with rounding errors */
	public double numericPrecision;
	
	/** for measuring runtime performances */
	private static double totalDuration;
	
	private int lastIteration=-1; // store the number of iterations before convergence
	
	/**
	 *  @param g  input graph to draw
	 *  @param C  step length
	 */
	public IterativeRelaxation(AdjacencyListGraph g, Polyhedron_3<Point_3> mesh, double stageEnergy, double numericPrecision) {
		System.out.print("Initializing iterative relaxation method...");
		if(g==null) {
			System.out.println("Input graph not defined");
			System.exit(0);
		}
		
		this.g=g;
		this.mesh=mesh;
		int N=g.sizeVertices();
				
		System.out.println("done ("+N+" nodes)");
		//System.out.println("k="+k+" - temperature="+temperature);
		System.out.println(this.toString());
		
		//Compute the weight of each point of g
		this.weight=new double[this.g.sizeVertices()];
		this.alpha=new double[this.g.sizeVertices()];
		for(int i=0;i<this.g.sizeVertices();i++){
			this.weight[i]=1.0/((double)this.g.degree(this.g.getNode(i)));
		}
		
		this.statistics=new Stat("IterativeRelaxation",false);//Stat object that contains all the statistics we want to work on
		this.stageEnergy=stageEnergy;
		
		this.numericPrecision = numericPrecision;
		
		this.verbosity=SphericalLayout.SHOW_ONLY_TIMINGS;
		//this.verbosity=SphericalLayout.SHOW_ONLY_STATISTICS;
	}
	
	/**
	 * Perform a fixed number of iterations of the iterative relaxation algorithm.
	 * 
	 * @param iterations  number of iterations to perform
	 */	
	public void computeLayout(int iterations) {
		for(int i=0;i<iterations;i++)
			this.computeOneIteration();
	}
	
	/**
	 * Perform iterations until convergence <p>
	 * <p>
	 * Remark: as explained in Saba/Gotsman/Sheffer 2005 (section 2), the algorithm
	 * continues while <p>
	 *  ||x(k) - x(k-1) || >= tolerance <p>
	 *  
	 *  where 'k' is the number of the iterations.
	 * 
	 * @param tolerance  numeric tolerance
	 */	
	public void computeLayout(double tolerance) {
		// TO DO Gapsard the algorithm continues while ||x(k) - x(k-1) || >= tolerance
		// remark: you can use to store the old locations of vertices (before performing the iteration)
		// Point_3[] oldPos=this.storeVertexLocations();
		tolerance = this.stageEnergy;
		
		Point_3[] oldPos = this.storeVertexLocations();
		while (this.iterationCount == 0 || (this.iterationCount < this.maxIterations && getDistance(this.g.listOfPoints(), oldPos)>tolerance)){
			System.out.print("Performing iteration: "+this.iterationCount);
			long startTime=System.nanoTime(), endTime; // for evaluating time performances
			
			Point_3[]vr1=new Point_3[this.g.sizeVertices()];
			for(Node u:this.g.vertices){// for each node u of g, replace the node u by the barycenter of its neighbors
				oldPos[u.index] = u.p;
				
				Point_3 p=new Point_3(0.,0.,0.);
				for(Node v:u.neighbors){
					p.x+=v.getPoint().x;
					p.y+=v.getPoint().y;
					p.z+=v.getPoint().z;
				}
				p.x*=this.weight[u.index];
				p.y*=this.weight[u.index];
				p.z*=this.weight[u.index];
				
				double norm=Math.sqrt(p.x*p.x+p.y*p.y+p.z*p.z);
				this.alpha[u.index]=norm;
				if (norm > this.numericPrecision){
					vr1[u.index]=new Point_3(p.x/norm,p.y/norm,p.z/norm);	
				}
				else{
					vr1[u.index]=u.getPoint();
				}
			}
			for(int i=0;i<this.g.sizeVertices();i++){
				this.g.getNode(i).p=vr1[i];
			}
			//}
			// evaluate time performances
	    	endTime=System.nanoTime();
	        double duration=(double)(endTime-startTime)/1000000000.;
	        System.out.println("iteration "+this.iterationCount+" done ("+duration+" seconds)");
	        
	        System.out.println(this.g.sizeVertices());
	        /**
	         * Update of the statistics
	         */
	        if(this.iterationCount==0){
	        	this.statistics.time.add(duration);
	        }
	        else{
	        	this.statistics.time.add(duration+this.statistics.time.get(this.iterationCount-1));
	        }
	        double E=computeTotalEnergy();
	        this.statistics.energy.add(E);
	        //this.statistics.edgeLength.add(this.computeEdgeLengthAesthetic());
	        //this.statistics.angleMean.add(this.computeAngleAesthetic());
	        //this.statistics.ratioArea.add(this.ratioArea());
	        //this.statistics.crossing.add(this.numberOfCrossing());
	        
			this.iterationCount++; // increase counter (to count the number of performed iterations)
		}
		
	}

	/**
	 * Perform one iteration of the algorithm.
	 */	
	public void computeOneIteration() {
		if(this.verbosity==SphericalLayout.SHOW_ALL)
			System.out.print("Gauss-Seidel iteration: ");
		
		long startTime=System.nanoTime(), endTime; // for evaluating time performances
		
		Point_3[]vr1=new Point_3[this.g.sizeVertices()];
		for(Node u:this.g.vertices){// for each node u of g, replace the node u by the barycenter of its neighbors
			Point_3 p=new Point_3(0.,0.,0.);
			for(Node v:u.neighbors){
				p.x+=v.getPoint().x;
				p.y+=v.getPoint().y;
				p.z+=v.getPoint().z;
			}
			p.x*=this.weight[u.index];
			p.y*=this.weight[u.index];
			p.z*=this.weight[u.index];
			
			double norm=Math.sqrt(p.x*p.x+p.y*p.y+p.z*p.z);
			if (norm > this.numericPrecision){
				vr1[u.index]=new Point_3(p.x/norm,p.y/norm,p.z/norm);	
			}
			else{
				System.err.print("Warning: norm zero, "+norm);
				vr1[u.index]=u.getPoint();
			}
		}
		
		double maxDisp=0.;
		for(Node u:this.g.vertices){
			maxDisp=Math.max(u.getPoint().distanceFrom(vr1[u.index]).doubleValue(), maxDisp);
			u.p=vr1[u.index];
		}
		
		double residualNew=this.getResidualL2();
		double DeltaResidual=residualNew-residual;// difference of L2 residuals: Res(k)-Res(k-1)
		residual=residualNew;
		

		//}
		// evaluate time performances
    	endTime=System.nanoTime();
        double duration=(double)(endTime-startTime)/1000000000.;        
        this.totalDuration=this.totalDuration+duration;
        
        if(this.verbosity==SphericalLayout.SHOW_ONLY_TIMINGS) {
        	System.out.print("iteration "+this.iterationCount+" done ("+duration+" seconds)");
        	System.out.println("[total duration: "+this.totalDuration+" seconds]");
        }
        else if(this.verbosity==SphericalLayout.SHOW_ONLY_STATISTICS) {
        	int skipIteration=5; // show only one iteration every 'skipIteration'
        	if(this.iterationCount%skipIteration==0)
        		System.out.println(""+this.iterationCount+"\t"+this.getStats());
        }
        else if(this.verbosity==SphericalLayout.SHOW_ALL) {
        	//System.out.println(""+this.getStats());
        	System.out.println("maxDisp= "+maxDisp+", Res(k)-Res(k-1)= "+DeltaResidual);
        	if(maxDisp<this.residualError) {
        		if(this.lastIteration<0)
        			lastIteration=this.iterationCount;
        		System.out.println("convergence: true, it"+this.lastIteration);
        	}
        }

        // Update of the statistics
        /*if(this.iterationCount==0){
        	this.statistics.time.add(duration);
        }
        else{
        	this.statistics.time.add(duration+this.statistics.time.get(this.iterationCount-1));
        }
        double E=computeTotalEnergy();
        this.statistics.energy.add(E);*/
        
        //this.statistics.edgeLength.add(this.computeEdgeLengthAesthetic());
        //this.statistics.angleMean.add(this.computeAngleAesthetic());
        //this.statistics.ratioArea.add(this.ratioArea());
        //this.statistics.crossing.add(this.numberOfCrossing());
        
		this.iterationCount++; // increase counter (to count the number of performed iterations)
	}
	
	/**
	 * Compute and return the potential energy for the Tutte layout
	 */	
	public double computeTotalEnergy() {
//		System.out.print("Computing potential energy of the layout: ");
		
		double energy=0.0;
		for(int i=0;i<this.g.sizeVertices();i++) {
			Node u=this.g.vertices.get(i);
			Point_3 p=u.getPoint();
			for(Node v: u.neighbors) {
				if(v!=null) {
					Point_3 q=v.getPoint();
					//double dist=SphericalGeometry.geodesicDistance(p,v.getPoint());
					double dist=Math.sqrt((p.x-q.x)*(p.x-q.x)+(p.y-q.y)*(p.y-q.y)+(p.z-q.z)*(p.z-q.z));
					energy+=dist*dist;
				}
			}
		}
		energy=energy/2.0;
		
		this.currentEnergy=energy;
		
		return energy;
	}
	
	public String toString() {
		String result="iterative relaxation (Projected Gauss-Seidel) \n";
		result=result+"current energy= "+this.currentEnergy+"\n";
		
		return result;
	}

	/**
	 * Get the name of the layout method
	 */
	public String getName() {
		return "Gauss-Seidel iterative relaxation";
	}
	
	public double getDistance(ArrayList<Point_3>graph1,Point_3[]graph2){
		double distance = 0.;
		int len = graph2.length;
		for (int i=0;i<len;i++){
			distance += (graph1.get(i).x-graph2[i].x)*(graph1.get(i).x-graph2[i].x)+(graph1.get(i).y-graph2[i].y)*(graph1.get(i).y-graph2[i].y)+(graph1.get(i).z-graph2[i].z)*(graph1.get(i).z-graph2[i].z);
		}
		return Math.sqrt(distance)/((double)len);
	}
	
	/**
	 * Compute the L2 residual error (L2 distance from the solution)
	 */
	public double getResidualL2(){
		double n=this.g.sizeVertices();
		double s=0;
		for(Node v: this.g.vertices){
			Point_3 p=new Point_3(0.,0.,0.);
			Point_3 q=new Point_3(0.,0.,0.);
			for(Node u: v.neighbors){
				p.x+=u.p.x;
				p.y+=u.p.y;
				p.z+=u.p.z;
			}
			/*q.x=v.getPoint().x*this.alpha[v.index]-this.weight[v.index]*p.x;
			q.y=v.getPoint().y*this.alpha[v.index]-this.weight[v.index]*p.y;
			q.z=v.getPoint().z*this.alpha[v.index]-this.weight[v.index]*p.z;*/

			q.x=v.getPoint().x-this.weight[v.index]*p.x;
			q.y=v.getPoint().y-this.weight[v.index]*p.y;
			q.z=v.getPoint().z-this.weight[v.index]*p.z;
			
			s+=q.x*q.x+q.y*q.y+q.z*q.z;
		}
		return (1.0/n)*Math.sqrt(s); 
	}	


}
