package sd.layout;

import java.util.Random;

import jdg.graph.AdjacencyListGraph;
import jdg.graph.Node;
import sd.util.SphericalGeometry;
import Jcg.geometry.Point_3;
import Jcg.polyhedron.*;

/**
 * Network visualization: abstract implementation of the force-directed paradigm (for graph drawing in 3D)
 * 
 * @author Luca Castelli Aleardi, Ecole Polytechnique
 * @version april 2017
 */
public abstract class Layout3D {
	public static final int NO_STATISTICS=0; // do not show statistics
	public static final int SHOW_ONLY_TIMINGS=1; // show runtime performances
	public static final int SHOW_ALL=2; // show all details
	
	/** choose to show or not all statistics during the layout computation */
	public int verbosity=2;

	/** input graph to draw */
	public AdjacencyListGraph g;
	
	/** input 3d mesh to draw */
	public Polyhedron_3 mesh;
	
	public double w, h, d; // dimensions of the drawing area
	
	/** maximum number of iterations to perform in any case */
	public final int maxIterations=2000;
    
	public static int seed=10;
	/** Random generator */	
	public static Random generator = new Random(seed); // initialize random generator
	
	/**
	 * Initialize vertex locations at random in 3D (in a box of given size WxHxD, centered at the origin)
	 */	
	public static void setRandomPoints3D(AdjacencyListGraph g, double w, double h, double d) {
		Point_3 p; 
		double w1=w/2., h1=h/2., d1=d/2.;
		for(Node u: g.vertices){
			double n1=w1-2*w1*generator.nextDouble();
			double n2=h1-2*h1*generator.nextDouble();
			double n3=d1-2*d1*generator.nextDouble();
		    p = new Point_3 (n1, n2, n3); // 3D point
		    u.setPoint(p);
		}		
	}

	/**
	 * Store a copy of the current locations of all vertices
	 */	
	public Point_3[] storeVertexLocations() {
		Point_3[] current=new Point_3[g.sizeVertices()]; 
		for(Node u: g.vertices){
			Point_3 p=u.getPoint();
		    current[u.index] = new Point_3 (p.x, p.y, p.z); // store a copy of the coordinates of vertex 'u'
		}
		return current;
	}
	
	/**
	 * Compute the average of the edge lengths (using euclidean distance in R^3)
	 */
	public static double computeEdgeLengthAverage(AdjacencyListGraph graph){
		double avg=0.0;
		for(int i=0;i<graph.sizeVertices();i++){
			for(Node u:graph.getNode(i).neighbors){
				Point_3 p=u.getPoint();
				avg+=p.distanceFrom(graph.getNode(i).getPoint()).doubleValue();
			}
		}
		avg=avg/(2.0*graph.sizeVertices());
		return avg;
	}

	/**
	 * Perform one iteration of the layout algorithm.
	 */	
	public abstract void computeOneIteration();
	
	/**
	 * Perform 'k' iteration of the layout algorithm.
	 * 
	 * @param iterations  number of iterations to perform
	 */	
	public abstract void computeLayout(int iterations);

	/**
	 * Perform layout computation until convergence
	 * 
	 * @param tolerance  numeric tolerance
	 */	
	public abstract void computeLayout(double tolerance);
	
	/**
	 * Compute the average of the edge lengths (using standard euclidean distance)
	 */
	public double computeEdgeLengthAverage(){
		double avg=0.0;
		for(Node u: this.g.vertices){
			Point_3 p=u.getPoint();
			for(Node v:u.neighbors){
				Point_3 q=v.getPoint();
				avg+=q.distanceFrom(p).doubleValue();
			}
		}
		avg=avg/(2.0*this.g.sizeEdges()); // bug: fixed by Luca
		return avg;
	}
	
	/**
	 * Compute the maximum length of the edges
	 */
	public double computeEdgeLengthMax(){
		double max=0.0;
		for(int i=0;i<this.g.sizeVertices();i++){
			for(Node u:this.g.getNode(i).neighbors){
				Point_3 p=u.getPoint();
				double dist=this.g.getNode(i).getPoint().distanceFrom(p).doubleValue();
				if(dist>max){
					max=dist;
				}
			}
		}
		return max;
	}

	/**
	 * Compute the edge lengths aesthetic metric
	 */	
	public double computeEdgeLengthAesthetic() {
		double d=0.0;
		double m=this.g.sizeEdges();
		double avg=computeEdgeLengthAverage();
		double max=computeEdgeLengthMax();
		System.out.println("max="+max+", avg="+avg);
		
		int edgeCounter=0;
		for(Node u: this.g.vertices){
			Point_3 p=u.getPoint();
			for(Node v: u.neighbors) { // visit all neighbors of vertex 'u'
				
				if(u.index<v.index) {
					Point_3 q=v.getPoint();
					double distance=p.distanceFrom(q).doubleValue();

					double term=Math.abs((distance-avg)/(Math.max(avg, max-avg)));
					System.out.println("term: "+term+", d="+d+" - distance(p, q)="+distance);
					d=d+term;
					edgeCounter++;
				}
			}
		}
		
		d=d/m;
		System.out.println("max="+max+", avg="+avg);
		System.out.println("|E|="+m+", e="+d);
		System.out.println("edge counter="+edgeCounter);
		return (1-d);
	}

	/**
	 * Get the name of the layout method
	 */
	public String getName() {
		return "euclidean layout (generic)";
	}

}
