package sd.layout;

import java.util.LinkedList;

import Jcg.geometry.Point_3;
import Jcg.geometry.Vector_3;
import Jcg.polyhedron.Polyhedron_3;

import jdg.graph.AdjacencyListGraph;
import jdg.graph.Node;

import sd.util.SphericalGeometry;
import sd.util.Stat;

import Jama.*;

/**
 * A spherical graph layout, based on the original force directed algorithm by Fruchterman and Reingold (1991) <p>
 * <p>
 * Features: <p>
 * -) no use of Octrees: computation of repulsive forces is slow <p>
 * -) simulated annealing is supported <p>
 * 
 * @author Luca Castelli Aleardi, Ecole Polytechnique
 * @version april 2017
 */
public class SphericalFR91 extends SphericalLayout implements SpringEmbedder {
	// parameters of the algorithm by Fruchterman and Reingold
	/** force model, defining attractive and repulsive forces */
	public SpringElectricalForce force; // force model
	
	/** computation of repulsive forces: it can be <p>
	 * -) exact (slow, quadratic time) <p>
	 * -) approximate (fast, with Octrees)
	 * */
	public ForceComputation forceComputation; // method for the computation of repulsive forces

	/** area of the drawing region (2d or 3d) */
	public double area;

	// parameters of the "simulated annealing" process
	/** initial temperature */
	protected double temperature;
	/** minimal temperature (strictly positive) */
	protected double minTemperature;
	/** cooling constant: temperature decreases (e.g. linearly) at each iteration */
	protected double coolingConstant;
	/** decide whether to make use of simulated annealing*/
	protected boolean useCooling;
	
	/** step length: this parameters is used to control displacements, according to adaptive cooling system described in Hu2006 */
	protected double step; // by default step=1, in the original algorithm FR91

	/** decide whether to make use of weights for computing repulsive forces (used in multi-level layouts) */
	protected boolean useVertexWeights;

	/** 
	 * Decide whether to move all vertices or only one vertex: <p>
	 * -) if TRUE: move all points all together, after computing all forces and displacements <p>
	 * -) if FALSE: move one vertex at time (computing forces and displacement only for that vertex
	 */
	protected boolean moveAllVerticesAtOnce=false; // by default, move all vertices all together (as in the original FR91 layout algorithm)
	
	public int iterationCount=0; // count the number of performed iterations
	
	public double tol;
	
	/** total runtime cost */
	protected double totalDuration;

	/**
	 * Initialize the parameters of the force-directed layout
	 * 
	 *  @param g  input graph to draw
	 *  @param C  step length
	 */
	public SphericalFR91(AdjacencyListGraph g, Polyhedron_3<Point_3> mesh, double stageEnergy, double tol) {
		System.out.print("Initializing force-directed method: Spherical Fruchterman-Reingold 91...");
		if(g==null) {
			System.out.println("Input graph not defined");
			System.exit(0);
		}
		this.g=g;
		this.mesh=mesh;
		int N=g.sizeVertices();
		
		// set the parameters of the spherical FR91
		this.area=4*Math.PI;
		double k=Math.sqrt(area/N)/2.; // as in the original FR91
		this.force=new SphericalSpringForce(k); // u (gravity not defined for the spherical case)
		this.force.setC(1.);
		this.force.setP(1);
		
		// setting parameters for the computation of repulsive forces: exact (slow) computation
		this.forceComputation=new ExactForceComputation(this.force);
		
		// setting parameters for simulated annealing
		this.temperature=0.1;
		this.minTemperature=0.05;
		this.coolingConstant=0.95;
		this.useCooling=true; // use cooling system by default
		this.step=1;
		
		this.statistics=new Stat("FR91", true);//Stat object that contains all the statistics we want to work on
		
		System.out.println("done ("+N+" nodes)");
		//System.out.println("k="+k+" - temperature="+temperature);
		System.out.println(this.toString());
		
		this.stageEnergy=stageEnergy;
		this.booleanContinu=true;
		this.Ek_1=0.;
		this.tol=tol;
	}
	
	/** Set the value of the parameter 'C', the relative strength of the repulsive and attractive forces (as used in Hu 2006) */
	public void setC(double c) {
		if(this.force!=null)
			this.force.setC(c);
	}
	
	/** Get the value of the parameter 'C', the relative strength of the repulsive and attractive forces (as used in Hu 2006) */
	public double getC() {
		if(this.force!=null)
			return this.force.C;
		else {
			System.err.println("Warning: the force model is not defined");
			return 0.;
		}
	}
	
	/** Set the value of the parameter 'K', the natural spring length */
	public void setK(double k) {
		if(this.force!=null)
			this.force.setK(k);
	}

	/** Get the value of the parameter 'K', the natural spring length */
	public double getK() {
		if(this.force!=null)
			return this.force.k;
		else {
			System.err.println("Warning: the force model is not defined");
			return 0.;
		}
	}
	
	/** Set the value of the step length (as described in Hu 2006) */
	public void setStep(double step) {
		this.step=step;
	}
	
	/** Get the value of the step length (as used in Hu 2006) */
	public double getStep() {
		return this.step;
	}
	
	/** Enable cooling process */	
	public void enableCooling() {
		this.useCooling=true;
	}

	/** Enable cooling process */	
	public void disableCooling() {
		this.useCooling=false;
	}
	
	/** Get the current value of the temperature */
	public double getTemperature() {
		return this.temperature;
	}

	/** Get the value of the minimal temperature */
	public double getMinimalTemperature() {
		return this.minTemperature;
	}

	/** Get the value of the cooling constant */
	public double getCoolingConstant() {
		return this.coolingConstant;
	}

	/** Set the initial value of the temperature */
	public void setInitialTemperature(double t) {
		this.temperature=t;
	}

	/** Set the initial value of the cooling constant */
	public void setCoolingConstant(double c) {
		this.coolingConstant=c;
	}

	/** Set the minimal value of the temperature */
	public void setMinimalTemperature(double min) {
		this.minTemperature=min;
	}
	
	/** decide to move all vertices all together, after computing all mutual forces (as in the original FR91 layout algorithm) */	
	public void moveAllVerticesAtOnce() {
		this.moveAllVerticesAtOnce=true;
	}

	/** decide to move only one vertex at a time, computing forces and displacement only for that vertex (as in Walshaw 2003) */	
	public void moveOnlyOneVertex() {
		this.moveAllVerticesAtOnce=false;
	}
	
	/** use vertex weights for computing repulsive forces (as in Walshaw 2003) */	
	public void useVertexWeights() {
		this.useVertexWeights=true;
	}

	/** do not use vertex weights for computing repulsive forces (as in FR91) */	
	public void noVertexWeights() {
		this.useVertexWeights=false;
	}

	/**
	 * Perform one iteration of the Force-Directed algorithm.
	 * Positions of vertices are updated according to their mutual attractive and repulsive forces.
	 * 
	 * @param iterations  number of iterations to perform
	 */	
	public void computeOneIteration() {	
		if(this.booleanContinu==true || this.iterationCount<10){
			/**
			 * Update statistics
			 */
			/*boolean boolGeodesic=this.isGeodedic(0.01);
			if(this.iterationCount==0 && boolGeodesic){
				this.statistics.IterationCountNullCrossing=0;
				this.statistics.TimeNullCrossing=0.;
			}
			if(this.iterationCount>0 && boolGeodesic){
				this.statistics.IterationCountNullCrossing=this.iterationCount;
				this.statistics.TimeNullCrossing=this.statistics.time.get(this.iterationCount-1);
			}*/
			System.out.print(this.getName()+": performing iteration "+this.iterationCount);
			if(useCooling==true)
				System.out.println("(temperature: "+this.temperature+")");
			else
				System.out.println("(no simulated annealing)");

			long startTime=System.nanoTime(), endTime; // for evaluating time performances

			//first case: we compute all the repulsive forces (no use of the octree)

			// first step: for each vertex compute the displacements due to attractive and repulsive forces
			Vector_3[] attractiveDisp=this.computeAllAttractiveForces(); // displacement due to attractive forces between neighbors
			Vector_3[] repulsiveDisp=this.computeAllRepulsiveForces(); // displacement due to repulsive forces between all nodes

			for(Node u:this.g.vertices){
				int index=u.index;
				Point_3 p=u.getPoint();
				if(u.degree()>0) { // only non isolated nodes are involved in the computation
					Vector_3 disp=attractiveDisp[index].sum(repulsiveDisp[index]); //total displacement 
					//System.out.println("u"+u.index+": fa="+attractiveDisp[index].squaredLength()+" - fr="+repulsiveDisp[index].squaredLength());
					double normdisp=Math.sqrt(disp.squaredLength().doubleValue());
					if(normdisp>this.tol){
						Vector_3 v1=new Vector_3(new Point_3(0.,0.,0.),p);
						Vector_3 v2=disp.multiplyByScalar(1.0/normdisp);
						Vector_3 vectorrotation=v1.crossProduct(v2); //vector/axis for the rotation to find the new point q
						double min=Math.min(normdisp, this.temperature);
						disp=disp.multiplyByScalar(this.step*min/normdisp); // use the variable step length (as in Hu2006)
						double norm=Math.sqrt(disp.squaredLength().doubleValue());
						double angle;
						if(v1.innerProduct(v2.crossProduct(vectorrotation)).doubleValue()>=0){ //to find the direction the rotation has to be done
							angle=norm;
						}
						else{
							angle=-norm;
						}
						Matrix matrotation=SphericalGeometry.rotationMatrix(vectorrotation,angle);
						double [][] vp= new double[][]{{p.x},{p.y},{p.z}};
						Matrix res= matrotation.times(new Matrix(vp));//apply the rotation to p
						Point_3 q=new Point_3(res.get(0, 0),res.get(1,0),res.get(2,0));
						//double normq=  A FAIRE
						u.p=q;
					}
					else{
						System.err.println("\t fixed vertex: v"+u.index);
						u.p=p;
					}

				}
			}

			// evaluate time performances
			endTime=System.nanoTime();
			double duration=(double)(endTime-startTime)/1000000000.;
			this.totalDuration=this.totalDuration+duration;
			System.out.print("Spherical FR91: iteration "+this.iterationCount+" done ("+duration+" seconds)");
			System.out.println("[total duration: "+this.totalDuration+" seconds]");

	        /**
	         * Update of the statistics
	         */
	        /*if(this.iterationCount==0){
	        	this.statistics.time.add(duration);
	        }
	        else{
	        	this.statistics.time.add(duration+this.statistics.time.get(this.iterationCount-1));
	        }*/
	        //double E=computeTotalEnergy();
	        //this.statistics.energy.add(E);
	        //this.statistics.edgeLength.add(this.computeEdgeLengthAesthetic());
	        //this.statistics.angleMean.add(this.computeAngleAesthetic());
	        //this.statistics.ratioArea.add(this.ratioArea());
	        //this.statistics.crossing.add(this.numberOfCrossing());
	
	        
	        /*if(this.iterationCount==0){
	        	this.Ek_1=E;
	        }
	        else{
	        	if(this.iterationCount>=10 && this.Ek_1-E<this.stageEnergy){
	        		this.booleanContinu=false;
	        	}
	        	else{
	        		this.Ek_1=E;
	        	}
	        }*/
	        
			if(useCooling==true) // check whether "simulated annealing" is used
				this.cooling(); // update temperature
			
			this.iterationCount++; // increase counter (to count the number of performed iterations)
			
			//System.out.println(computeTotalEnergy());
		}
		
	}
	
	/**
	 * Perform 'k' iteration of the Force-Directed algorithm.
	 * Positions of vertices are updated according to their
	 * mutual attractive and repulsive forces.
	 * 
	 * @param iterations  number of iterations to perform
	 */	
	public void computeLayout(int iterations) {
		for(int i=0;i<iterations;i++)
			this.computeOneIteration();

	}
	
	/**
	 * Perform iterations until convergence <p>
	 * <p>
	 * Remark: as explained in Saba/Gotsman/Sheffer 2005 (section 2), the algorithm
	 * continues while <p>
	 *  ||x(k) - x(k-1) || >= tolerance <p>
	 *  
	 *  where 'k' is the number of the iterations.
	 * 
	 * @param tolerance  numeric tolerance
	 */	
	public void computeLayout(double tolerance) {
		// TO DO LUCA: the algorithm continues while ||x(k) - x(k-1) || >= tolerance
	}
		
	/**
	 * Compute, for each vertex, the displacement due to repulsive forces (between all nodes)
	 * 
	 * @return a vector v[]: v[i] stores the geometric displacement of the i-th node
	 */	
	public Vector_3[] computeAllRepulsiveForces() {
		Vector_3[] repulsiveDisp=new Vector_3[this.g.sizeVertices()];
		for(int i=0;i<this.g.vertices.size();i++)
			repulsiveDisp[i]=new Vector_3(0., 0., 0.);
		
		long startTime=System.nanoTime(), endTime; // for evaluating time performances
		
		double toleranceNorm=0.0001;
		int countVertices=0;
		for(Node u: this.g.vertices) {
			//repulsiveForce(u, repulsiveDisp); // faster version (forces are computed only once)
			repulsiveDisp[countVertices]=this.forceComputation.computeRepulsiveForce(g, u, toleranceNorm);
			//System.out.println("\t f("+u.index+")="+repulsiveDisp[countVertices]);
			countVertices++;
		}
    	endTime=System.nanoTime();
        double duration=(double)(endTime-startTime)/1000000000.;
        System.out.print("\ttimings: "+duration+" seconds (repulsive force)");
        System.out.println(" ["+this.force.countRepulsive+" forces computed]");
        this.force.countRepulsive=0;
        
        /**
         * Statistics
         */
        this.statistics.timeRepulsive.add(duration);
        return repulsiveDisp;
	}
  
	
	/**
	 * Compute the displacement of vertex 'u', due to the attractive forces of its neighbors
	 * 
	 * @param u  the vertex to which attractive forces are applied
	 * @return 'disp' a 3d vector storing the displacement of vertex 'u'
	 */	
	public Vector_3 computeAttractiveForce(Node u) {
		Vector_3 ex=new Vector_3(1.,0.,0.);
		Vector_3 ey=new Vector_3(0.,1.,0.);
		Vector_3 ez=new Vector_3(0.,0.,1.);
		Vector_3 displacement=new Vector_3(0., 0., 0.);
		Vector_3 delta=new Vector_3(0.,0.,0.);
		Point_3 p=u.getPoint();
		
		Vector_3[]tangentialVectors=SphericalGeometry.getTangentialVector(p,0.01);
		Vector_3 v1=tangentialVectors[0];
		Vector_3 v2=tangentialVectors[1];
		double distance=0.;
		for (Node v: u.neighbors){
			if(v!=null){
				distance=SphericalGeometry.geodesicDistance(p,v.getPoint());
				delta=new Vector_3(p,v.getPoint());
				delta=new Vector_3(v1.innerProduct(delta).doubleValue()*v1.innerProduct(ex).doubleValue()+v2.innerProduct(delta).doubleValue()*v2.innerProduct(ex).doubleValue(),delta.innerProduct(v1).doubleValue()*v1.innerProduct(ey).doubleValue()+delta.innerProduct(v2).doubleValue()*ey.innerProduct(v2).doubleValue(),delta.innerProduct(v1).doubleValue()*v1.innerProduct(ez).doubleValue()+delta.innerProduct(v2).doubleValue()*v2.innerProduct(ez).doubleValue());
				double norm=Math.sqrt(delta.squaredLength().doubleValue());
				if(norm>this.tol){
					delta=delta.multiplyByScalar(1.0/norm);
				}
				else{
					delta=new Vector_3(0.,0.,0.);
				}
				displacement=displacement.sum(delta.multiplyByScalar(this.force.attractiveForce(distance)));
			}
		}
		return displacement;
	}
	
	/**
	 * Compute, for each vertex, the displacement due to attractive forces (between neighboring nodes)
	 * 
	 * @return a vector v[]: v[i] stores the geometric displacement of the i-th node
	 */	
	public Vector_3[] computeAllAttractiveForces() {
		Vector_3[] attractiveDisp=new Vector_3[this.g.sizeVertices()];
		long startTime=System.nanoTime(), endTime; // for evaluating time performances
		
		int i=0;
		for(Node u: this.g.vertices) {
			attractiveDisp[i]=computeAttractiveForce(u);
			i++;
		}
    	endTime=System.nanoTime();
        double duration=(double)(endTime-startTime)/1000000000.;
        System.out.println("\ttimings: "+duration+" seconds (attractive force)");
        
        /**
         * Statistics
         */
        this.statistics.timeAttractive.add(duration);
        
        return attractiveDisp;
	}
	
	/**
	 * Compute for the entire system the energy corresponding to the attractive forces
	 */
	public double computeAttractiveEnergy(){
		double E=0.;
		for(Node u:this.g.vertices){
			Point_3 p=u.getPoint();
			for(Node v:u.neighbors){
				Point_3 q=v.getPoint();
				double dist=SphericalGeometry.geodesicDistance(p, q);
				E+=this.force.attractiveEnergy(dist);
			}
		}
		return E;
	}
	
	/**
	 * Compute for the entire system the energy corresponding to the repulsive forces
	 */
	public double computeRepulsiveEnergy(){
		double E=0.;
		if(this.booleanContinu==true){
			for(Node u:this.g.vertices){
				Point_3 p=u.getPoint();
				for(Node v:this.g.vertices){
					if(!v.equals(u)){
						Point_3 q=v.getPoint();
						double dist=SphericalGeometry.geodesicDistance(p, q);
						E+=this.force.repulsiveEnergy(dist);
					}
				}
			}
			return E;
		}
		else{
			LinkedList<Point_3> l =new LinkedList<Point_3>();
			for (int i=0;i<this.g.sizeVertices();i++){
				l.add(this.g.getNode(i).getPoint());
			}

			for(Node u:this.g.vertices){
				E=E+this.computeRepulsiveEnergy();
			}
			return E;
		}
	}
	
	/**
	 * Compute and return the potential energy for the force-directed layout
	 * (equal to the difference between the attractive energy and the repulsive energy)
	 */	
	public double computeTotalEnergy() {
		System.out.print("Computing potential energy of the layout: ");
		double E=this.computeAttractiveEnergy()-this.computeRepulsiveEnergy();
		this.currentEnergy=E;
		return E;
	}
	
	/**
	 * Cooling system: the temperature decreases at each iteration (following the method adopted in Walshaw)
	 * 
	 * Remark: the temperature is assumed to remain strictly positive (>=minTemperature)
	 */	
	protected void cooling() {
		//this.temperature=Math.max(this.temperature-coolingConstant, minTemperature);
		this.temperature=Math.max(this.temperature*this.coolingConstant, minTemperature); // variant described in Walshaw 2003
	}
	
	public String toString() {
		String result="Spring embedder: standard FR91 (exact force calculation)";
		result=result+"\tk= "+this.force.k+", ";
		result=result+"C= "+this.force.C+"\n";
		result=result+"\t initial temperature= "+this.temperature+"\n";
		result=result+"\t minimal temperature= "+this.minTemperature+"\n";
		result=result+"\t cooling constant= "+this.coolingConstant+"\n";
		result=result+"current energy= "+this.currentEnergy;
		
		return result;
	}

	/**
	 * Get the name of the layout method
	 */
	public String getName() {
		return "spherical FR91";
	}

}
