package sd.layout;

import java.util.List;
import java.util.Random;

import Jcg.geometry.GeometricOperations_3;
import Jcg.geometry.Point_3;
import Jcg.geometry.Segment_3;
import Jcg.geometry.Triangle_3;
import Jcg.geometry.Vector_3;
import Jcg.polyhedron.Face;
import Jcg.polyhedron.Halfedge;
import Jcg.polyhedron.Polyhedron_3;
import Jcg.polyhedron.Vertex;
import jdg.graph.AdjacencyListGraph;
import jdg.graph.Node;
import sd.util.SphericalGeometry;
import sd.util.Stat;

/**
 * Abstract implementation of the force-directed paradigm, for graph drawing on the sphere
 * 
 * @author Luca Castelli Aleardi, Ecole Polytechnique
 * @version 2017
 */
public abstract class SphericalLayout {
	public static final int NO_STATISTICS=0; // do not show statistics
	public static final int SHOW_ONLY_TIMINGS=1; // show runtime performances
	public static final int SHOW_ALL=2; // show all details
	public static final int SHOW_ONLY_STATISTICS=3; // show statistics (no runtime performances)
	
	/** choose to show or not all statistics during the layout computation */
	public int verbosity=SHOW_ONLY_TIMINGS;
	
	public AdjacencyListGraph g; // input graph to draw
	public Polyhedron_3<Point_3> mesh; // input mesh to draw

	public double stageEnergy;
	
	public boolean booleanContinu;
	public double Ek_1;
	
	/** current potential energy of the layout: computed by the last iteration */
	public double currentEnergy=0;
	
	/** maximum number of iterations to perform in any case */
	public final int maxIterations=10000; 

	/** residual error (for convergence) */
	public final double residualError=0.0001; 

	//boolean[] isFixed;

	public static int seed=10;
	/** Random generator */	
	public static Random generator = new Random(seed); // initialize random generator
	
	public Stat statistics;
	
	/**
	 * Initialize vertex locations at random on a 3D sphere (in a 3D BOX of given size WxHxD, centered at the origin)
	 */	
	public static void setRandomPointsOnSphere(AdjacencyListGraph g) {
		Point_3 p; 
		double w1=1./2., h1=1./2., d1=1./2.;
		for(Node u: g.vertices){
			double n1=w1-2*w1*generator.nextDouble();
			double n2=h1-2*h1*generator.nextDouble();
			double n3=d1-2*d1*generator.nextDouble();
		    p = new Point_3 (n1, n2, n3);
		    u.setPoint(projectOnSphere(p));
		}		
	}

	/**
	 * Project the vertices of a 3d mesh on the unit sphere <p>
	 * <p>
	 * Remark: the original 3d mesh remains unchanged
	 */	
	public static void projectMeshOnSphere(AdjacencyListGraph g, Polyhedron_3<Point_3> inputMesh) {
		System.out.print("Projecting the mesh on the sphere");
		Point_3 bar=new Point_3();
		Point_3[] points=new Point_3[inputMesh.sizeOfVertices()];
		for(Vertex<Point_3> v: inputMesh.vertices)
			points[v.index]=v.getPoint();
		
		bar.barycenter(points);
		Point_3 origin=new Point_3(0., 0., 0.);
		Vector_3 barVector=new Vector_3(bar, origin);
		
		Point_3 p;
		for(Node u: g.vertices){
			p = inputMesh.vertices.get(u.index).getPoint();
		    u.setPoint(projectOnSphere(p.sum(barVector)));
		}		
		System.out.println("done");
	}

	/**
	 * Project the vertices of a graph on the unit sphere <p>
	 */	
	public static void projectMeshOnSphere(AdjacencyListGraph inputGraph) {
		System.out.print("Projecting the graph on the sphere");
		Point_3 bar=new Point_3();
		Point_3[] points=new Point_3[inputGraph.sizeVertices()];
		for(Node v: inputGraph.vertices)
			points[v.index]=v.getPoint();
		
		// first compute the barycenter of all vertices
		bar.barycenter(points);
		Point_3 origin=new Point_3(0., 0., 0.);
		Vector_3 barVector=new Vector_3(bar, origin);
		
		Point_3 p;
		for(Node u: inputGraph.vertices){
			p = inputGraph.vertices.get(u.index).getPoint();
		    u.setPoint(projectOnSphere(p.sum(barVector)));
		}
		System.out.println("done");
	}

	/**
	 * Perform one iteration of the Force-Directed algorithm.
	 * Positions of vertices are updated according to their
	 * mutual attractive and repulsive forces.
	 */	
	public abstract void computeOneIteration();

	/**
	 * Perform 'k' iteration of the Force-Directed algorithm.
	 * Positions of vertices are updated according to their
	 * mutual attractive and repulsive forces.
	 * 
	 * @param iterations  number of iterations to perform
	 */	
	public abstract void computeLayout(int iterations);

	/**
	 * Perform layout computation until convergence
	 * 
	 * @param tolerance  numeric tolerance for convergence
	 */	
	public abstract void computeLayout(double tolerance);
	
	/**
	 * Compute and return the potential energy for the layout
	 */	
	public abstract double computeTotalEnergy();


	/**
	 * Check whether the layout is a "geodesic embedding" (no edge crossing)
	 */	
	public boolean isGeodedic(double tolerance) {
		boolean b;
		double area=0.0;
		double angle0=0.0;
		double angle1=0.0;
		double angle2=0.0;
		Point_3 p0=new Point_3();
		Point_3 p1=new Point_3();
		Point_3 p2=new Point_3();
		
		int nV=this.g.sizeVertices();
		for(int i=0;i<this.mesh.sizeOfFacets();i++){ //for each face of the mesh, compute and sum the area
			int[]indices=getVerticesInFace(i);
			if(indices[0]>=nV || indices[1]>=nV || indices[2]>=nV)
				return false;
			//System.out.println("ind: "+indices[0]);
			
			p0=this.g.getNode(indices[0]).getPoint();
			p1=this.g.getNode(indices[1]).getPoint();
			p2=this.g.getNode(indices[2]).getPoint();
			//System.out.println("p: "+p0+" "+p1+" "+p2);
			//System.out.println("norm: "+p0.distanceFrom(new Point_3(0.,0.,0.))+" "+p1.distanceFrom(new Point_3(0.,0.,0.))+" "+p2.distanceFrom(new Point_3(0.,0.,0.)));
			angle0=SphericalGeometry.sphericalAngle(p0, p1, p2);			
			angle1=SphericalGeometry.sphericalAngle(p1, p2, p0);
			angle2=SphericalGeometry.sphericalAngle(p2, p0, p1);
			//System.out.println("angle: "+angle0+" "+angle1+" "+angle2);
			//System.out.println("angle0: "+angle0);
			//System.out.println("angle1: "+angle1);
			//System.out.println("angle2: "+angle2);
			area+=angle0+angle1+angle2-Math.PI;
			//System.out.println("area: "+area);
		}
		//System.out.println("area: "+area);
		b=(Math.abs(4.0*Math.PI-area)<tolerance);
		
		this.computeAreaAverage();
		this.computeAreaAesthetic();
		
		return b;
	}
	
	/**
	 * Compute the minimum area of all triangles
	 */	
	public double minArea() {
		double area=0.0;
		double areaMin=4.0*Math.PI;
		double angle0=0.0;
		double angle1=0.0;
		double angle2=0.0;
		Point_3 p0=new Point_3();
		Point_3 p1=new Point_3();
		Point_3 p2=new Point_3();
		
		for(int i=0;i<this.mesh.sizeOfFacets();i++){ //for each face of the mesh, compute and sum the area
			int[]indices=getVerticesInFace(i);
			p0=this.g.getNode(indices[0]).getPoint();
			p1=this.g.getNode(indices[1]).getPoint();
			p2=this.g.getNode(indices[2]).getPoint();
			angle0=SphericalGeometry.sphericalAngle(p0, p1, p2);
			angle1=SphericalGeometry.sphericalAngle(p1, p2, p0);
			angle2=SphericalGeometry.sphericalAngle(p2, p0, p1);
			area=angle0+angle1+angle2-Math.PI;
			if(area<areaMin){
				areaMin=area;
			}
		}

		return areaMin;
	}
	
	/**
	 * Compute the average of the edge lengths (according to geodesic distance on the unit sphere)
	 */
	public double computeAreaAverage(){
		double avg=0.0;
		
		int countFaces=0;
		for(Face<Point_3> f: this.mesh.facets){
			double area=this.computeArea(f.index);
			avg=avg+area;
			countFaces++;
		}
		avg=avg/countFaces;
		
		//System.out.println("Average area= "+avg);
		return avg;
	}
	
	/**
	 * Compute the maximum area of all triangles
	 */	
	public double computeMaxArea() {
		double area=0.0;
		double areaMax=0.;
		double angle0=0.0;
		double angle1=0.0;
		double angle2=0.0;
		Point_3 p0=new Point_3();
		Point_3 p1=new Point_3();
		Point_3 p2=new Point_3();
		
		for(int i=0;i<this.mesh.sizeOfFacets();i++){ //for each face of the mesh, compute and sum the area
			int[]indices=getVerticesInFace(i);
			p0=this.g.getNode(indices[0]).getPoint();
			p1=this.g.getNode(indices[1]).getPoint();
			p2=this.g.getNode(indices[2]).getPoint();
			angle0=SphericalGeometry.sphericalAngle(p0, p1, p2);
			angle1=SphericalGeometry.sphericalAngle(p1, p2, p0);
			angle2=SphericalGeometry.sphericalAngle(p2, p0, p1);
			area=angle0+angle1+angle2-Math.PI;
			if(area>areaMax){
				areaMax=area;
			}
		}

		return areaMax;
	}

	/**
	 * Compute the area of a triangle face (having number 'i')
	 */	
	public double computeArea(int i) {
		double area=0.0;
		double angle0=0.0;
		double angle1=0.0;
		double angle2=0.0;
		Point_3 p0, p1, p2;
		
			int[]indices=getVerticesInFace(i);
			p0=this.g.getNode(indices[0]).getPoint();
			p1=this.g.getNode(indices[1]).getPoint();
			p2=this.g.getNode(indices[2]).getPoint();
			angle0=SphericalGeometry.sphericalAngle(p0, p1, p2);
			angle1=SphericalGeometry.sphericalAngle(p1, p2, p0);
			angle2=SphericalGeometry.sphericalAngle(p2, p0, p1);
			area=angle0+angle1+angle2-Math.PI;

		return area;
	}

	/**
	 * Compute areaMin/areaMax
	 */
	public double ratioArea(){
		return minArea()/this.computeMaxArea();
	}

	/**
	 * Compute the minimum angle aesthetic metric
	 */	
	public double computeMinimumAngleAesthetic() {
		double d=0.0;
		double n=(double)this.g.sizeVertices();
		double thetai=0.0;
		double thetaimin=0.0;
		int degree=0;
		double min=2*Math.PI;
		
		for(Node u:this.g.vertices){
			degree=u.degree();
			thetai=2*Math.PI/(double)degree;
			double[]angles=new double [degree];
			angles=getAnglesAtVertex(this.mesh.vertices.get(u.index));
			for(int i=0;i<degree;i++){
				if(angles[i]<min){
					min=angles[i];
				}
			}
			thetaimin=min;
			d+=(1.0/n)*Math.abs((thetai-thetaimin)/thetai);
			min=2*Math.PI;
		}
		return d;
	}

	/**
	 * Compute the maximum angle
	 */	
	public double computeMaxAngle() {
		double d=0.0;
		double n=(double)this.g.sizeVertices();

		int degree=0;
		double max=0.;
		
		for(Node u:this.g.vertices){
			degree=u.degree();
			double[] angles=new double [degree];
			angles=getAnglesAtVertex(this.mesh.vertices.get(u.index));
			for(int i=0;i<degree;i++){
				if(angles[i]>max){
					max=angles[i];
				}
			}
		}
		return max;
	}

	/**
	 * Compute the maximum angle
	 */	
	public double computeAngleAverage() {
		double n=(double)this.g.sizeVertices();

		int degree=0;
		double angleSum=0.;
		
		int countAngles=0;
		for(Node u:this.g.vertices){
			degree=u.degree();
			double[] angles=new double [degree];
			angles=this.getAnglesAtVertex(this.mesh.vertices.get(u.index));
			//System.out.println("Angle v"+u.index);
			for(int i=0;i<degree;i++){
				angleSum=angleSum+angles[i];
				//System.out.println("\tangle= "+angles[i]);
				countAngles++;
			}
		}
		return angleSum/countAngles;
	}

	/**
	 * Compute the minimum angle aesthetic metric
	 */	
	public double computeAngleAesthetic() {
		double angle=0.0;
		double avg=computeAngleAverage();
		double max=computeMaxAngle();
		//System.out.println("max angle="+max+", avg angle="+avg);
		
		return 0;
	}

	/**
	 * Compute and return the spherical angles around a vertex
	 */	
	public double[] getAnglesAtVertex(Vertex<Point_3> v) {
		int d=this.mesh.vertexDegree(v); // degree of the vertex
		double[] angles=new double[d];
		
		Halfedge<Point_3> e=v.getHalfedge(); // halfedge incident to v, and oriented toward v
		
		Point_3 p0=new Point_3();
		Point_3 p1=new Point_3();
		Point_3 p2=new Point_3();
		
		int nV=this.g.sizeVertices();
		for(int i=0;i<d;i++){
			if(e.getVertex().index>=nV || e.getNext().getVertex().index>=nV || e.getNext().getNext().getVertex().index>=nV)
				return angles;

			p0=this.g.getNode(e.getVertex().index).getPoint();
			p1=this.g.getNode(e.getNext().getVertex().index).getPoint();
			p2=this.g.getNode(e.getNext().getNext().getVertex().index).getPoint();
			angles[i]=SphericalGeometry.sphericalAngle(p0, p1, p2);
			e=e.getNext().getOpposite();
		}
		
		return angles;
	}
	
	/**
	 * Compute the average of the edge lengths (according to geodesic distance on the unit sphere)
	 */
	public double computeEdgeLengthAverage(){
		double avg=0.0;
		
		int countEdges=0;
		for(Node u: this.g.vertices){
			Point_3 p=u.getPoint();
			for(Node v: u.neighbors) { // visit all neighbors of vertex 'u'
				avg=avg+SphericalGeometry.geodesicDistance(v.getPoint(), p);
				countEdges++;
			}
		}
		avg=avg/countEdges;
		return avg;
	}
	
	/**
	 * Compute the maximum length of the edges
	 */
	public double computeEdgeLengthMax(){
		double max=0.0;
		for(int i=0;i<this.g.sizeVertices();i++){
			for(Node u:this.g.getNode(i).neighbors){
				Point_3 p=u.getPoint();
				if(SphericalGeometry.geodesicDistance(this.g.getNode(i).getPoint(), p)>max){
					max=SphericalGeometry.geodesicDistance(this.g.getNode(i).getPoint(),p);
				}
			}
		}
		return max;
	}

	/**
	 * Compute the edge lengths aesthetic metric
	 */	
	public double computeEdgeLengthAesthetic() {
		double d=0.0;
		double avg=computeEdgeLengthAverage();
		double max=computeEdgeLengthMax();
		//System.out.println("max="+max+", avg="+avg);
		
		int edgeCounter=0;
		for(Node u: this.g.vertices){
			Point_3 p=u.getPoint();
			for(Node v: u.neighbors) { // visit all neighbors of vertex 'u'
				
				if(u.index<v.index) {
					Point_3 q=v.getPoint();
					//double distance=p.distanceFrom(q).doubleValue();
					double distance=SphericalGeometry.geodesicDistance(p, q);
					//System.out.println("distance "+distance);

					double term=Math.abs((distance-avg)/(Math.max(avg, max-avg)));
					//System.out.println("term: "+term+", d="+d+" - distance(p, q)="+distance);
					d=d+term;
					edgeCounter++;
				}
			}
		}
		
		//System.out.println("average, max "+avg+", "+max+", m="+m);
		
		d=d/edgeCounter;
		return (1-d);
	}
	
	/**
	 * Compute the edge lengths aesthetic metric
	 */	
	public double computeAreaAesthetic() {
		double d=0.0;
		double avg=computeAreaAverage();
		double max=this.computeMaxArea();
		//System.out.println("max="+max+", avg="+avg);
		
		int faceCounter=0;
		for(Face f: this.mesh.facets){
			double area=this.computeArea(f.index);
			double term=Math.abs((area-avg)/(Math.max(avg, max-avg)));
			d=d+term;
			faceCounter++;
		}
		
		//System.out.println("average, max "+avg+", "+max+", m="+m);
		
		d=d/faceCounter;
		return (1-d);
	}

	/**
	 * Method that return a boolean: true
	 * @param tolerance
	 * @return
	 */
	public boolean checkPointOnSphere(double tolerance){
		for(Node u: this.g.vertices){
			double norm=Math.sqrt(u.getPoint().squareDistance(new Point_3(0.,0.,0.)).doubleValue());
			if(Math.abs(1-norm)>tolerance){
				return false;
			}		
		}
		return true;
	}
	
	/**
	 * Return the number of crossing between edges
	 */
	public double numberOfCrossing(){
		double n=0.;
		
		for(Node u: this.g.vertices){
			for(Node v: u.neighbors){
				for (Node w: this.g.vertices){
					for(Node x:w.neighbors){
						//if(!u.equals(w) && !u.equals(x) && !v.equals(w) && !v.equals(x)){
							if(SphericalGeometry.geodesicIntersection(u.getPoint(), v.getPoint(), w.getPoint(), x.getPoint())){
								System.out.println("("+u.index+", "+v.index+") - ("+w.index+", "+x.index+")");
								System.out.println(u.getPoint().distanceFrom(w.getPoint()));
								n++;
							}
						//}
					}
				}
			}
		}
		return n/4.0;
	}

	/**
	 * Compute the node separation aesthetic metric
	 */	
	public double computeNodeSeparationAesthetic() {
		throw new Error("To be completed");
	}
	
	/**
	 * Compute the linear interpolation of two points.
	 * 
	 * @return the 3D point given by: ta+(1-t)b
	 */
	public int[] getVerticesInFace(int face) {
		if(face>=this.mesh.sizeOfFacets())
			throw new Error("Error: wrong face index "+face);
		
		int v1=this.mesh.facets.get(face).getEdge().getVertex().index;
		int v2=this.mesh.facets.get(face).getEdge().getNext().getVertex().index;
		int v3=this.mesh.facets.get(face).getEdge().getNext().getNext().getVertex().index;
		
		return new int[]{v1, v2, v3};
	}
	
	/**
	 * Compute the linear interpolation of two points.
	 * 
	 * @return the 3D point given by: ta+(1-t)b
	 */
	public static Point_3 linearInterpolation(Point_3 a, Point_3 b, double t) {
		Point_3 result;
		result=Point_3.linearCombination(new Point_3[] {a, b}, new Double[] {t, 1-t});
		return result;
	}

	/**
	 * Project a 3D point on the unit sphere.<p>
	 * It assumes the vector P-O has positive norm.
	 * 
	 * @param p the point to project
	 * @return a 3D point of norm 1
	 */
	public static Point_3 projectOnSphere(Point_3 p) {
		Point_3 origin=new Point_3(0., 0., 0.);
		
		if(p.squareDistance(origin).doubleValue()==0.0)
			p.translateOf(new Vector_3(Math.random()/1000., Math.random()/1000., Math.random()/1000.)); // perform random perturbation
		Vector_3 v=new Vector_3(origin, p);
		double norm=Math.sqrt(v.innerProduct(v).doubleValue());
		//System.out.println(" norm: "+norm);
		Vector_3 normalized=v.divisionByScalar(norm);
		return origin.sum(normalized);
	}

	/**
	 * Compute the total displacement of nodes (with respect to the node positions in the previous layout)
	 * 
	 * @param previousPositions  an array storing the 3D positions for all nodes in the previous layout
	 * @return the total node displacement (the sum of all displacement, for all nodes)
	 */	
	public double computeTotalDisplacement(Point_3[] previousPositions) {
		if(previousPositions==null)
			throw new Error("Error: previous positions not defined");
		
		double disp=0.; // displacement
		for(int i=0;i<this.g.sizeVertices();i++) { // compute the displacement for all nodes
			Point_3 p1=previousPositions[i]; // position of the node in the previous layout
			Point_3 p2=this.g.vertices.get(i).getPoint();
			disp=disp+p1.distanceFrom(p2).doubleValue(); // (euclidean) displacement of the i-th node // LUCA: to be adapted to geodesic distance
		}
		
		return disp;
	}
	
	/**
	 * Get the 'd' edge segments [(v0, v1), (v1, v2), ..., (v_{d-1}, v_0)] around a vertex 'u' <p>
	 * <p>
	 * Remark: v0, v1, v2, ... v_{d-1} are the neighbors of vertex 'u'
	 * 
	 * @param u  input node in the graph
	 * @return the set of edge segments [(v0, v1), (v1, v2), ...] of the star of edges around 'u' in the mesh
	 */
	public Segment_3[] getStarEdges(Node u) {
		int nV=this.g.sizeVertices(); // number of vertices in the graph
		if(u.index>=nV)
			return null; // the south and north poles are not considered
		
		Segment_3[] result=null; // store the edges around 'u'
		
		// first case: vertex 'u' is not neighbor of the south or north poles
		if(u.degree()==this.mesh.vertexDegree(this.mesh.vertices.get(u.index))) {
			int d=u.degree();
			result=new Segment_3[d];
			
			Vertex vertexU=this.mesh.vertices.get(u.index); // vertex in the mesh corresponding to 'u'
			
			int count=0;
			Halfedge firstEdge=vertexU.getHalfedge(); // first edge incident to 'u' (oriented toward 'u')
			Halfedge<Point_3> pEdge=firstEdge.getNext().getOpposite();
			while(pEdge!=firstEdge) { // visit all edges around 'u' in cw direction
				Point_3 p=pEdge.getOpposite().getVertex().getPoint();
				Point_3 q=pEdge.getNext().getVertex().getPoint();
				result[count]=new Segment_3(p, q);
				
				pEdge=pEdge.getNext().getOpposite(); // next neighbor around 'u' in cw direction
				count++;
			}
			Point_3 p=pEdge.getOpposite().getVertex().getPoint();
			Point_3 q=pEdge.getNext().getVertex().getPoint();
			result[count]=new Segment_3(p, q);
			
			if(count!=d-1)
				throw new Error("Error: wrong number of edges for vertex v"+u.index +" (vertex not incident to the south/north face)");
		}
		//second case: vertex 'u' is on the south or north face in the mesh: it has one more neighbor in the mesh (the south or north pole)
		else {
			return null;
		}
		
		return result;
	}
	
	/**
	 * Store a copy of the current locations of all vertices
	 */	
	public Point_3[] storeVertexLocations() {
		Point_3[] current=new Point_3[g.sizeVertices()]; 
		for(Node u: g.vertices){
			Point_3 p=u.getPoint();
		    current[u.index] = new Point_3 (p.x, p.y, p.z); // store a copy of the coordinates of vertex 'u'
		}
		return current;
	}
	
	/**
	 * Compute layout statistics
	 */
	public String getStats() {
		String result="";
		double energy=this.computeTotalEnergy();
		double metricEdgeLength=this.computeEdgeLengthAesthetic();
		double metricArea=this.computeAreaAesthetic();
		boolean isGeodesic=this.isGeodedic(0.0001);
		
		//double ratio=this.ratioArea();
		//double max=this.computeMaxArea();
		//double angle=this.computeAngleAesthetic();
		
		int precision=3;
		energy=this.round(energy, precision);
		metricEdgeLength=this.round(metricEdgeLength, precision);
		metricArea=this.round(metricArea, precision);

		//ratio=this.round(ratio, precision);
		//max=this.round(max, precision);
		//angle=this.round(angle, precision);
		
		//result=result+"energy"+"\tarea"+"\tedge"+"\tangle\n";
		result=result+energy+"\t"+metricEdgeLength+"\t"+metricArea+"\t"+isGeodesic;
		
		return result;
	}
	
	/**
	 * Given a 3D triangle mesh, compute the number of collisions (intersections) between non adjacent triangles
	 * 
	 * Remark: adjacent triangles (sharing at least one vertex) are assumed to be not intersecting
	 * 
	 * @return the number of collision between all pair of triangles
	 */
	public static int countTriangleCollisions(Polyhedron_3<Point_3> polyhedron, AdjacencyListGraph graph) {
		//System.out.print("Computing triangle collisions...");
		int result=0;
		int intersectingTriangles=0;
		int nF=polyhedron.sizeOfFacets();
		
		for(Face<Point_3> f: polyhedron.facets) {
			int index1=f.index;
			
			boolean intersecting=false;
			for(int i=index1+1; i<nF; i++) {
				Face g=polyhedron.facets.get(i);
				
		    	Point_3 a1=graph.vertices.get(f.getEdge().getVertex().index).getPoint();
		    	Point_3 b1=graph.vertices.get(f.getEdge().getNext().getVertex().index).getPoint();
		    	Point_3 c1=graph.vertices.get(f.getEdge().getNext().getNext().getVertex().index).getPoint();
		    	Point_3 a2=graph.vertices.get(g.getEdge().getVertex().index).getPoint();
		    	Point_3 b2=graph.vertices.get(g.getEdge().getNext().getVertex().index).getPoint();
		    	Point_3 c2=graph.vertices.get(g.getEdge().getNext().getNext().getVertex().index).getPoint();
		    	
		    	// if the two triangles share a vertex, then we do not count the collision
		    	if(a1.equals(a2)==false &&
		    	   a1.equals(b2)==false &&
		    	   a1.equals(c2)==false &&
		    	   b1.equals(a2)==false &&
		    	   b1.equals(b2)==false &&
		    	   b1.equals(c2)==false &&
		    	   c1.equals(a2)==false &&
		    	   c1.equals(b2)==false &&
		    	   c1.equals(c2)==false) {
		    		
		    		Triangle_3 t1=new Triangle_3(a1, b1, c1);
		    		Triangle_3 t2=new Triangle_3(a2, b2, c2);
		    		
		    		//System.out.println("Testing collisions: "+index1+" - "+i);
		    		
		    		if(GeometricOperations_3.doIntersect(t1, t2)==true) {
		    			intersecting=true;
		    			result++;
		    		}
		    	}
			}
			if(intersecting==true)
				intersectingTriangles++;
		}
		
		//System.out.println(" "+result+" "+intersectingTriangles);
		return intersectingTriangles;
	}

	/**
	 * Get the name of the layout method
	 */
	public String getName() {
		return "spherical layout (generic)";
	}
	
	/** Return the value after truncation */
	public static double round(double x, int prec) {
		double precision=Math.pow(10, prec);
		return ((int)(x*precision)/((double)precision));
	}

}
