package sd.layout;

/**
 * Definition of 'spherical forces' acting on nodes on the unit sphere. It extends the original spring model FR91 to the spherical case.
 * 
 * @author Luca Castelli Aleardi, Ecole Polytechnique
 * @version august 2017
 */
public class SphericalSpringForce extends SpringElectricalForce {
	
	/** The repulsive force is shift by a constant value, in order to get '0' for vertices at antipodal locations */
	private double vShift;

	/**
	 * Define the parameters of the original 'spring electrical' model of forces introduced in FR91<p>
	 * -) C=1 <p>
	 * -) p=1 <p>
	 * -) no gravity force <p>
	 * 
	 * @param k  natural spring length
	 */	
	public SphericalSpringForce(double k) {
		super(k);
		this.updateVerticalShift();
	}

	/**
	 * Define the parameters of the 'general spring electrical' model of forces adopted in Walshaw03, extended with a 'gravity' term <p>
	 * <p>
	 * 
	 * @param k  natural spring length
	 * @param gravity  the gravity constant
	 * @param C  relative strength of the repulsive and attractive forces (introduced in Walshaw2003 and Hu2006)
	 * @param p  this parameter regulates long-range repulsive force (as described in Hu2006)
	 */	
	public SphericalSpringForce(double k, double gravity, double C, int p) {
		super(k, gravity, C, p);
		this.updateVerticalShift();
	}
	
	/** Set the relative strength of the repulsive forces (as described in Walshaw2003 and Hu2006) **/
	public void setC(double C) {
		this.C=C;
		this.updateVerticalShift();
	}
	
	/** 
	 * Set the natural spring length <p>
	 * <p>
	 * Remark: by default in the original model FR91 the parameter k is set to <p>
	 *    k = C * sqrt(area/N) <p>
	 * where N is the number of vertices, and 'area' is the area of the layout
	 **/
	public void setK(double k) {
		this.k=k;
		this.updateVerticalShift();
	}

	/**
	 * Compute the (intensity of the) attractive force between two nodes at a given distance
	 * 
	 * @param distance  geodesic distance between two nodes on the unit sphere
	 */	
	public double attractiveForce(double distance) {
		return (distance*distance)/k;
	}
	
	/**
	 * Compute the (intensity of the) repulsive force between two nodes at a given distance
	 * 
	 * @param distance  geodesic distance between two nodes on the unit sphere
	 */	
	public double repulsiveForce(double distance) {
		this.countRepulsive++;
		
		if(this.p==1)
			return (C*(k*k)/distance)-this.vShift; // slight modification of repulsive forces
		else {
			System.err.println("Warning: this method is not supported in the spherical case for p>1");
			return C*(Math.pow(k, 1+this.p))/(Math.pow(distance, this.p)); // more general model, described in HU2006
		}
	}
	
	/**
	 * Compute the energy corresponding to the repulsive force between two nodes at a given distance
	 * 
	 * @param distance  distance between two nodes
	 */	
	public double repulsiveEnergy(double distance) {
		if(p==1)
			return C*((k*k)/2.)*(Math.log(distance*distance));
		else if (p>1)
			return C*(Math.pow(k, 1+this.p))/((1-p)*(Math.pow(distance, this.p-1))); // for the more general model, described in HU2006
		else
			throw new Error("Error: the energy is not defined for the value p="+p);
	}

	/**
	 * Compute the energy corresponding to the attractive force between two nodes at a given distance
	 * 
	 * @param distance  geodesic distance between two nodes on the unit sphere
	 */	
	public double attractiveEnergy(double distance) {
		return (distance*distance*distance)/(3*k);
	}

	/**
	 * Compute the vertical shift 'S': the repulsive force, in the spherical case, is shifted by a constant: f(d)-S <p>
	 * <p>
	 * Remark: shifting by a term 'S' allows to get a '0' repulsive force for antipodal vertices (for which the mutual distance is Math.PI)
	 */	
	private void updateVerticalShift() {
		this.vShift=this.C*this.k*this.k/Math.PI;
	}

	/**
	 * Return the name of the force model
	 */	
	public String toString() {
		return "spherical spring force (for spherical FR91)";
	}

}
