package sd.layout;

/**
 * A class implementing the 'spring electrical' force model described by Fruchterman and Reingold (1991)
 * 
 * @author Luca Castelli Aleardi, Ecole Polytechnique
 * @version april 2017
 */
public class SpringElectricalForce extends ForceModel {

	/** natural spring length */
	public double k;

	/** this parameter regulates the relative strength of the repulsive and attractive forces and was introduced in Walshaw2003 and Hu2006 */
	public double C;

	/** this parameter regulates long-range repulsive force (as described in Hu2006): <p>
	 *  the larger the parameter p, the weaker the long-range repulsive force **/
	protected int p;

	/** gravity constant */
	protected double gravity;

	/**
	 * Define the parameters of the original 'spring electrical' model of forces introduced in FR91<p>
	 * -) C=1 <p>
	 * -) p=1 <p>
	 * -) no gravity force <p>
	 * 
	 * @param k  natural spring length
	 */	
	public SpringElectricalForce(double k) {
		this.k=k;
		this.C=1; // this corresponds to the value of the original FR91 algorithm
		this.p=1; // this corresponds to the value of the original FR91 algorithm
		this.gravity=0;
	}

	/**
	 * Define the parameters of the 'spring electrical' model of forces of FR91, extended with a 'gravity' term <p>
	 * <p>
	 * Remark: by default default this model corresponds to the original spring electrical model FR91 <p>
	 * -) C=1 <p>
	 * -) p=1 <p>
	 * 
	 * @param k  natural spring length
	 * @param gravity  the gravity constant
	 */	
	public SpringElectricalForce(double k, double gravity) {
		this.k=k;
		this.C=1; // this corresponds to the value of the original FR91 algorithm
		this.p=1; // this corresponds to the value of the original FR91 algorithm
		this.gravity=gravity;
	}

	/**
	 * Define the parameters of the 'general spring electrical' model of forces adopted in Walshaw03, extended with a 'gravity' term <p>
	 * <p>
	 * 
	 * @param k  natural spring length
	 * @param gravity  the gravity constant
	 * @param C  relative strength of the repulsive and attractive forces (introduced in Walshaw2003 and Hu2006)
	 * @param p  this parameter regulates long-range repulsive force (as described in Hu2006)
	 */	
	public SpringElectricalForce(double k, double gravity, double C, int p) {
		this.k=k;
		this.C=C;
		this.p=p;
		this.gravity=gravity;
	}

	/** Set the relative strength of the repulsive forces (as described in Walshaw2003 and Hu2006) **/
	public void setC(double C) {
		this.C=C;
	}

	/** 
	 * Return the relative strength of the repulsive (parameter described in Walshaw2003 and Hu2006)
	 **/
	public double getC() {
		return this.C;
	}

	/** 
	 * This parameter leads a more general spring model, as described in Hu2006 <p>
	 * <p>
	 * For p=1, the repulsive forces correspond to the original spring electrical model (FR91)
	 **/
	public void setP(int p) {
		this.p=p;
	}
	
	/** 
	 * Set the natural spring length <p>
	 * <p>
	 * Remark: by default in the original model FR91 the parameter k is set to <p>
	 *    k = C * sqrt(area/N) <p>
	 * where N is the number of vertices, and 'area' is the area of the layout
	 **/
	public void setK(double k) {
		this.k=k;
	}
	
	/** 
	 * Return the natural spring length <p>
	 **/
	public double getK() {
		return this.k;
	}

	/**
	 * Compute the (intensity of the) attractive force between two nodes at a given distance
	 * 
	 * @param distance  distance between two nodes
	 */	
	public double attractiveForce(double distance) {
		return (distance*distance)/k;
	}
	
	/**
	 * Compute the energy corresponding to the attractive force between two nodes at a given distance
	 * 
	 * @param distance  distance between two nodes
	 */	
	public double attractiveEnergy(double distance) {
		return (distance*distance*distance)/(3*k);
	}

	/**
	 * Compute the (intensity of the) repulsive force between two nodes at a given distance
	 * 
	 * @param distance  distance between two nodes
	 */	
	public double repulsiveForce(double distance) {
		this.countRepulsive++;
		
		if(this.p==1) // original spring electrical model (FR91), for C=1
			return C*(k*k)/distance;
		else
			return C*(Math.pow(k, 1+this.p))/(Math.pow(distance, this.p)); // more general model, described in HU2006
	}
	
	/**
	 * Compute the energy corresponding to the repulsive force between two nodes at a given distance
	 * 
	 * @param distance  distance between two nodes
	 */	
	public double repulsiveEnergy(double distance) {
		if(p==1)
			return C*((k*k)/2.)*(Math.log(distance*distance));
		else if (p>1)
			return C*(Math.pow(k, 1+this.p))/((1-p)*(Math.pow(distance, this.p-1))); // for the more general model, described in HU2006
		else
			throw new Error("Error: the energy is not defined for the value p="+p);
	}

	/**
	 * Compute the (intensity of the) gravity force on a node
	 * 
	 * @param distance  distance a node and the gravity center
	 */	
	public double gravityForce(double distance) {
		return k*this.gravity*distance;
	}

	/**
	 * Return the name of the force model
	 */	
	public String toString() {
		return "Spring Electrical force (FR91)";
	}

}
