package sd.layout2D;

import Jcg.geometry.Point_2;
import jdg.graph.AdjacencyListGraph;
import jdg.graph.Node;

/**
 * Iterative implementation of the planar (2D) Tutte barycentric method
 * 
 * @author Luca Castelli Aleardi, Ecole Polytechnique
 * @version 2017
 */
public class IterativeTutteLayout2D extends TutteLayout2D {
	/** difference between new and old position (x coordinate)*/
	double[] deltaX;

	/** difference between new and old position (y coordinate)*/
	double[] deltaY;

	/**
	 * Initialize the drawing
	 */	
	public IterativeTutteLayout2D(AdjacencyListGraph g, Point_2[] points, boolean[] isOnBoundary, boolean[] isInside) {
		super(g, points, isOnBoundary, isInside); // class the constructor of the ancestor class
		
		int n=this.g.sizeVertices();
		this.deltaX=new double[n];
		this.deltaY=new double[n];
	}
	
	/**
	 * Compute the Tutte drawing of a planar graph, by iteratively
	 * computing barycenters of neighboring vertices <p>
	 * <p>
	 * The algorithm stops after a given number of iterations
	 */	
	public void computeLayout(int iterations) {
		System.out.print("Computing Tutte layout 2D...");
		
    	long startTime=System.nanoTime(), endTime; // for evaluating time performances

		// computes planar coordinates iteratively
		for(int i=0;i<iterations;i++) {
			this.computeOneIteration();
		}
		
    	// evaluate time performances
    	endTime=System.nanoTime();
    	double duration=(double)(endTime-startTime)/1000000000.;
    	System.out.println("done ("+duration+" seconds)");
	}

	/**
	 * Compute the Tutte drawing of a planar graph, by iteratively
	 * computing barycenters of neighboring vertices <p>
	 * <p>
	 * The algorithm stops when a given convergence condition is reached
	 */	
	public void computeLayout(double tolerance) {
		//boolean convergence=false;
		
		// computes planar coordinates iteratively
		//while(convergence==false) {
		//	this.computeOneIteration();
		//}
		throw new Error("To be completed");
	}

	/**
	 * Perform one iteration of the iterative Tutte planar layout. <p>
	 * Positions of vertices are updated according to the position of the barycenter of their neighbors
	 */	
	public void computeOneIteration() {
			//System.out.println("iteration "+i);
		int n=g.sizeVertices(); // total number of vertices in the graph
		double[] barX=new double[n];
		double[] barY=new double[n];
		
		for(Node u:this.g.vertices){
			int i=u.index; // index of the i-th vertex
			if(this.isInside[i]==true) { // only inner nodes (inside the cycle) are involved in the computation
				int degree=this.g.degree(u); // degree of the i-th vertex
				
				// compute the 2D coordinates of the barycenter
				for(Node v: u.neighborsList()) {// visit all neighbors of the i-th vertex
					int j=v.index; // index of the neighbor
					barX[i]=barX[i]+this.points[j].getX().doubleValue();
					barY[i]=barY[i]+this.points[j].getY().doubleValue();
				}
				
				barX[i]=barX[i]/((double)degree);
				barY[i]=barY[i]/((double)degree);
				
				//System.out.println("v"+i+": "+barX[i]+", "+barY[i]+", degree "+degree);
			}
		}

		// set the new coordinates of all inner vertices
		for(Node u:this.g.vertices){
			int i=u.index; // index of the i-th vertex
			if(this.isInside[i]==true) { // only inner nodes which are not fixed are involved in the computation
				this.deltaX[i]=this.points[i].x.doubleValue()-barX[i];
				this.deltaY[i]=this.points[i].y.doubleValue()-barY[i];
				
				// move each point to the its new position (barycenter of its neighbors)
				this.points[i].setX(barX[i]);
				this.points[i].setY(barY[i]);
			}
		}
	}
	
}
