package sd.layout2D;

import java.util.LinkedList;
import java.util.List;

import Jcg.polyhedron.Halfedge;
import Jcg.polyhedron.Polyhedron_3;
import Jcg.polyhedron.Vertex;

/**
 * Planar grid drawing based on the FPP algorithm: for a rectangular frame
 * 
 * @author Luca Castelli Aleardi, Ecole Polytechnique
 * @version 2017
 */
public class RectangularFPPLayout {
	
	/** 2D integer coordinates of the vertices of the graph */
	public int[] coordX, coordY;

	/** say whether a vertex belongs to the region (enclosed by the bottom path and the cut-border: the boundary vertices are included) */
	public boolean[] isInnerVertex; // LCA: only for debugging and drawing in 2D

    /** store, for each edge, the corresponding stretch in the x-direction (for the shift algorithm) */
    public int[] xStretch; 

    /** store, for each edge, the corresponding stretch for the region 'below' (on the other side) */
    public int[] xStretchOtherSide; 

	/** store the vertices according to their (standard) canonical ordering */
	public Vertex[] canonicalOrdering;

	/** it stores for each vertex the left and right edges on the cut-border (vertices are numbered according to their canonical ordering) */
	private Halfedge[] left, right;
  
	/** it stores the delta stretch, for the 'x' and 'y' direction (for each edge) */
	private int[] deltaX, deltaY;
	
	/** bottom path */
	public LinkedList<Halfedge> path;
	
	/** number of 'vertex conquests' */
	public int removedVertices;

	/** width of the final layout */
	public int width;

    /** say whether a vertex is lying on one of the three disjoint paths p0, p1 or p2 */
    //public boolean[] isOnPath;
	
	public static double totalDuration=0; // for evaluating time performances
    
	/** adjust the delta X of vertices on the disjoint paths: to obtain more balanced spaces vertices */
	public static int adjustDeltaX=0;
    /**
     * Initialize the FPP algorithm
     * 
     * @param removedVertices  number of vertex conquests (removals)
     */
    public RectangularFPPLayout(Polyhedron_3 mesh, LinkedList<Halfedge> path, int[] reverseCanonicalOrdering, 
							int[] xStretch, int[] xStretchOtherSide,
							Halfedge[] left, Halfedge[] right, int removedVertices, int[] deltaX, int[] deltaY) {
    	
    	System.out.print("Initializing FPP layout [n="+removedVertices+"]...");
    	
    	int nV=mesh.sizeOfVertices();
    	this.coordX=new int[nV];
    	this.coordY=new int[nV];
    	this.isInnerVertex=new boolean[nV];
    	
    	for(int i=0;i<nV;i++) {
    		this.coordY[i]=-10; // this value is chosen to check whether the final layout is correct
    	}
    	
    	this.xStretch=xStretch;
    	this.xStretchOtherSide=xStretchOtherSide;
    	this.path=path;
    	this.removedVertices=removedVertices;
    	
    	this.canonicalOrdering=new Vertex[removedVertices];
    	this.left=new Halfedge[removedVertices];
    	this.right=new Halfedge[removedVertices];
    	
    	for(Vertex u: (List<Vertex>)mesh.vertices) {
    		if(reverseCanonicalOrdering[u.index]!=-1) {
    			int index=removedVertices-reverseCanonicalOrdering[u.index]-1; // get the canonical labeling, from the reverse canonical labeling
    			this.canonicalOrdering[index]=u;
    			this.left[index]=left[u.index];
    			this.right[index]=right[u.index];
    		}
    	}
    	
    	this.deltaX=deltaX;
    	this.deltaY=deltaY;
    	
    	System.out.println("done");
    }
    
    /** 
     * Return the height of the final rectangular layout defined by: layoutBottom+River+LayoutTop <p>
     * <p>
     * The final height is given by: 2*max(w1/0, w2/0), where w1 and w2 are the 'widths' of the bottom layout and the top layout <p>
     * Recall the the height of the bottom (resp. top) sub-layout is: w/2
     *  
     * @return max(w1, w2)
     */
    public static int getLayoutHeight(RectangularFPPLayout layoutBottom, RectangularFPPLayout layoutTop) {
    	if(layoutBottom.width<=0 || layoutTop.width<=0)
    		throw new Error("Error: the width of the layout(s) is not defined");
    	
    	// the '+1' term is required to deal with 'pathological' cases: the edges on the river should not superpose to edges in the bottom and top layout
    	int result=Math.max(layoutBottom.width, layoutTop.width)+1; 
    	//System.out.println("Height of the rectangular layout: "+result);
    	
    	return result;
    }
    
    /** 
     * Return the normalized x-coordinate of the i-th vertex 
     * 
     * @param index  (original) index of a vertex (in the mesh) 
     * @return the x-coordinate of the i-th vertex: it must range in [0..1]
     */
    public double getX(int index) {
    	if(this.width<=0)
    		throw new Error("Error: the width of the layout is not defined");
    	
    	double result=((double)this.coordX[index])/this.width;
    	
    	if(result<0. || result>1.) // LCA: for debugging
    		throw new Error("Error: wrong x-coordinates for vertex v"+index+", x="+result);
    	
    	return result;
    }

    /** 
     * Return the normalized y-coordinate of the i-th vertex 
     * 
     * @param index  (original) index of a vertex (in the mesh)
     * @param w  width of the final layout, defined by the union of the bottom and top layouts
     * @param h  the final 'height' of the rectangular layout of a region (G0, G1 or G2)
     * @param isBottomRegion  it says whether the vertex belongs to the bottom sub-region (below the river)
     *  
     * @return the y-coordinate of the i-th vertex: it must range in [0..1]
     */
    public double getY(int index, int w, int h, boolean isBottomRegion) {
    	if(h<=0)
    		throw new Error("Error: the height of the layout is not well defined: "+h);
    	if(w<=0)
    		throw new Error("Error: the width of the layout is not defined: "+w);
    	//if(height%2!=0)
    	//	throw new Error("Error: the height of the layout is not defined or wrong: "+height);
    	
    	double result;
    	if(this.width==width) { // the width of this layout does coincide with the width of the final rectangular layout
    		if(isBottomRegion==true)
    			result=((double)this.coordY[index])/h;
    		else
    			result=((double)this.coordY[index]+(h))/h;
    	}
    	else { // the final width is larger of the width of the current layout: scaling must be applied
    		double scaleFactor=((double)this.width)/w; // compute the scale factor: it should be >=1.0
    		if(scaleFactor<1.)
    			throw new Error("Error: wrong scale factor: "+scaleFactor);
    		
    		if(isBottomRegion==true)
    			result=((double)this.coordY[index]*scaleFactor)/h;
    		else
    			result=((double)this.coordY[index]*scaleFactor+(h))/h;    		
    	}
    	
    	if(result<0. || result>1.) // LCA: for debugging
    		throw new Error("Error: wrong y-coordinates for vertex v"+index+", x="+result);
    	
    	return result;
    }

    /**
     * Compute the x and y delta stretch for each edge
     * 
     */
    public void computeDeltaStretch() {
    	int n=this.canonicalOrdering.length; // number of vertices to process
    	System.out.print("Computing delta stretches...");
    	long startTime=System.nanoTime(), endTime; // for evaluating time performances
    	
    	int k=path.size()-2; // number of edges on the bottom path (the first and last halfedges are discarded)
    	int counter=0;
    	
    	int stretchMin=(this.adjustDeltaX/5)*2; // it must be a multiple of 2
    	int adaptiveXStretch;
    	
    	for(Halfedge e: path) {
    		if(counter>0 && counter<=k) {
    			// the stretch of edges on the bottom path is set to '2'+2 times the stretch in the graph 'below'
    			// LCA: chose the good 'stretch' to obtain 'balanced' x-coordinates for the vertices on the paths
    			
    			adaptiveXStretch=2*Math.abs(k/2-counter)+stretchMin;
    			//adaptiveXStretch=this.adjustDeltaX;
    			
    			this.deltaX[e.index]=2+2*this.xStretchOtherSide[e.getOpposite().index]+adaptiveXStretch; 
    			this.deltaX[e.getOpposite().index]=2+2*this.xStretchOtherSide[e.getOpposite().index]+adaptiveXStretch; 
    			//System.out.println("e"+counter+", xStretch e: ("+e.getOpposite().getVertex().index+", "+e.getVertex().index+") ="+this.xStretch[e.index]);
    		}
    		counter++;
    	}
    	
    	Halfedge leftEdge, rightEdge;
    	int maxX=0, maxY=0;
    	for(int i=0;i<n;i++) {
    		int[] delta=this.getDeltaBelow(i); // total sum of stretches of the edges on the cut-border below the i-th vertex
    		
    		//System.out.print("\tDelta stretch v"+this.canonicalOrdering[i].index+": ");
    		leftEdge=this.left[i];
    		rightEdge=this.right[i];
    		
    		if((delta[0]+delta[1])%2!=0) // LCA: just for debugging
    			throw new Error("Warning: wrong x stretch");
    		
    		this.deltaX[leftEdge.index]=(delta[0]+delta[1])/2;
    		this.deltaX[rightEdge.index]=delta[0]-this.deltaX[leftEdge.index];
    		
    		this.deltaY[leftEdge.index]=this.deltaX[leftEdge.index];
    		this.deltaY[rightEdge.index]=-this.deltaX[rightEdge.index];
    		
    		maxX=Math.max(maxX, this.deltaX[leftEdge.index]); // LCA: just for stats
    		maxY=Math.max(maxY, this.deltaY[leftEdge.index]); // LCA: just for stats
    		
    		//System.out.print("left stretch: ["+this.deltaX[leftEdge.index]+", "+this.deltaY[leftEdge.index]+"]");
    		//System.out.println(", right stretch: ["+this.deltaX[rightEdge.index]+", "+this.deltaY[rightEdge.index]+"]");
    	}
    	
    	endTime=System.nanoTime();
        double duration=(double)(endTime-startTime)/1000000000.;
        totalDuration=totalDuration+duration;

    	System.out.print("done (n="+n+") ");
    	System.out.println(" ("+duration+" seconds)");
    }

    /**
     * Compute the 2D coordinates for all vertices in the region
     * 
     */
    public void computeVertexCoordinates() {
    	int n=this.canonicalOrdering.length; // number of vertices to process
    	System.out.print("Computing vertex coordinates...");
    	long startTime=System.nanoTime(), endTime; // for evaluating time performances
    	
    	Halfedge leftEdge;
    	for(int i=0;i<n;i++) {
    		leftEdge=this.left[i];
    		
    		int x=this.coordX[leftEdge.getVertex().index]; // x-coordinate of the left neighbor of 'v_i' on the cut-border
    		int y=this.coordY[leftEdge.getVertex().index]; // y-coordinate of the left neighbor
    		
    		this.coordX[this.canonicalOrdering[i].index]=x+this.deltaX[leftEdge.index]+2*this.xStretch[leftEdge.index];
    		this.coordY[this.canonicalOrdering[i].index]=y+this.deltaY[leftEdge.index];
    		
    		/*System.out.print("v"+this.canonicalOrdering[i].index+": ("+this.coordX[this.canonicalOrdering[i].index]+", "+this.coordY[this.canonicalOrdering[i].index]+") ");
    		System.out.print(", left neighbor: "+this.left[i].getVertex().index);
    		System.out.print(", left stretch: ("+this.deltaX[i]+", "+this.deltaY[i]+") ");
    		System.out.println(" [v"+this.left[i].getVertex().index+", v"+this.right[i].getOpposite().getVertex().index+"]");*/
    	}
    	
    	endTime=System.nanoTime();
        double duration=(double)(endTime-startTime)/1000000000.;
        totalDuration=totalDuration+duration;

    	System.out.println("done (n= "+n+") ");
    }

    /**
     * Rotate the layout of 'pi': this is used from drawing the upper sub-graph (above a river) <p>
     * 
     */
    public void rotateLayout() {
    	int n=this.canonicalOrdering.length; // number of vertices to process
    	int k=path.size()-2; // number of edges on the path: the path is assumed to be oriented from the source to the destination
    	
    	if(this.width<=0)
    		throw new Error("Error: the width of the layout is wrong: "+width);
    	
    	// store the indices: the order is standard (since the path is oriented from the source to the destination)
    	int counter=0;
    	for(Halfedge e: path) {
    		if(counter<=k)
    			this.coordX[e.getVertex().index]=width-this.coordX[e.getVertex().index]; // change the x-coorinates of the vertices on the path
    		counter++;
    	}

    	// set all 'inner vertices' (not lying on the path)
    	for(int i=0;i<n;i++) {   		
    		this.coordX[this.canonicalOrdering[i].index]=width-this.coordX[this.canonicalOrdering[i].index];
    		this.coordY[this.canonicalOrdering[i].index]=-this.coordY[this.canonicalOrdering[i].index];
    		
    		/*System.out.print("v"+this.canonicalOrdering[i].index+": ("+this.coordX[this.canonicalOrdering[i].index]+", "+this.coordY[this.canonicalOrdering[i].index]+") ");
    		System.out.print(", left neighbor: "+this.left[i].getVertex().index);
    		System.out.print(", left stretch: ("+this.deltaX[i]+", "+this.deltaY[i]+") ");
    		System.out.println(" [v"+this.left[i].getVertex().index+", v"+this.right[i].getOpposite().getVertex().index+"]");*/
    	}
    	
    }

    /**
     * Get the x-delta and y-delta of all edges 'below' the i-th vertex
     * 
     * @param i  the i-th vertex (its canonical labeling), for which the x and y stretches have to be computed
     */
    private int[] getDeltaBelow(int i) {
    	//System.out.print(" --- delta stretch below v"+this.canonicalOrdering[i].index);
    	Halfedge leftEdge=this.left[i];
    	Halfedge rightEdge=this.right[i];
    	
    	this.isInnerVertex[rightEdge.getVertex().index]=true;
    	
    	int[] result=new int[2];
    	
    	if(rightEdge.getNext()==leftEdge) { // the 'i-th' vertex is incident to a triangle on the cut-border
    		result[0]=this.deltaX[rightEdge.getPrev().getOpposite().index]+4;
    		result[1]=this.deltaY[rightEdge.getPrev().getOpposite().index];
    	}
    	else {
    		int x=0, y=0;
    		
    		Halfedge pEdge=rightEdge;
    		while(pEdge!=leftEdge.getOpposite()) { // turn around the i-th vertex, in cw order
    			x=x+this.deltaX[pEdge.getPrev().getOpposite().index];
    			y=y+this.deltaY[pEdge.getPrev().getOpposite().index];
    			
    			pEdge=pEdge.getNext().getOpposite();
    		}
    		
    		result=new int[] {x+4, y};
    	}
    	
    	//System.out.print(": ["+result[0]+", "+result[1]+"] --- ");
    	return result;
    }

    /**
     * Set the coordinates for all vertices on the bottom path <p>
     * <p>
     * Remark: the bottom path is assumed to be oriented from the source to the destination
     * 
     * @param path  the list of halfedges defining the path
     */
    public void setBottomPath() {
    	int k=path.size()-2; // number of edges on the bottom path (the first and last halfedges are discarded)
    	System.out.print("Setting coordinates on the bottom path, k="+k);
    	
    	// store the indices in reverse order
    	Halfedge[] halfedges=new Halfedge[k];
    	int counter=0;
    	for(Halfedge e: path) {
    		if(counter>0 && counter<=k)
    			halfedges[k-counter]=e; // store halfedges in reverse order (at the exception of the ones containing the source and destination vertices)
    		counter++;
    	}
    	
    	// set x and y coordinates for all vertices on the bottom path
    	//System.out.print(" [");
    	// set the x and y coordinates of the left-most bottom vertex
    	Halfedge e0=halfedges[0]; // left-most halfedge on the (bottom) path: the one preceding the last edge on the path
    	this.coordX[e0.getVertex().index]=0;
    	this.coordY[e0.getVertex().index]=0;
    	
    	for(Halfedge e: halfedges) {
    		// get the indices of the extremities of the edge (v, u) (oriented toward u)
    		int uIndex=e.getVertex().index;
    		int vIndex=e.getOpposite().getVertex().index;
    		//System.out.print("v"+vIndex);
    		
    		int sizeHalfedge=+2*this.xStretch[e.index]+this.deltaX[e.index];
    		//System.out.print(sizeHalfedge+", ");
    		this.coordX[vIndex]=this.coordX[uIndex]+sizeHalfedge;
    		this.coordY[vIndex]=0;
    		
    		// only for debugging and drawing in 2D
    		this.isInnerVertex[uIndex]=true;
    		this.isInnerVertex[vIndex]=true;
    		
    		//System.out.println(": ("+this.coordX[vIndex]+", "+this.coordY[vIndex]+") ");
    	}
    	//System.out.print("] ");
    	
    	this.width=this.coordX[this.path.getFirst().getVertex().index]; // set the width of the layout (coordinate of the right-most bottom vertex)
    	
    	System.out.println("...done");
    }

    /**
     * Set the coordinates and stretches for all vertices on the bottom path <p>
     * <p>
     * Remark: the bottom path is assumed to be oriented from the source to the destination
     * 
     * @param path  the list of halfedges defining the path
     */
    public void setTopPath() {
    	int k=path.size()-2; // number of edges on the bottom path (the first and last halfedges are discarded)
    	System.out.print("Setting coordinates on the top path, k="+k);
    	
    	// store the indices: the order is standard (since the path is oriented from the source to the destination)
    	Halfedge[] halfedges=new Halfedge[k];
    	int counter=0;
    	for(Halfedge e: path) {
    		if(counter>0 && counter<=k)
    			halfedges[counter-1]=e.getOpposite(); // store halfedges (at the exception of the ones containing the source and destination vertices)
    		counter++;
    	}
    	
    	//System.out.print(" [");
    	// set x and y coordinates
    	for(Halfedge e: halfedges) {
    		// get the indices of the extremities of the edge (v, u) (oriented toward u)
    		int uIndex=e.getVertex().index;
    		int vIndex=e.getOpposite().getVertex().index;
    		//System.out.print("v"+vIndex);
    		
    		int sizeHalfedge=+2*this.xStretch[e.index]+this.deltaX[e.index];
    		//System.out.print(sizeHalfedge+", ");
    		this.coordX[vIndex]=this.coordX[uIndex]+sizeHalfedge;
    		this.coordY[vIndex]=0;
    		
    		// only for debugging and drawing in 2D
    		this.isInnerVertex[uIndex]=true;
    		this.isInnerVertex[vIndex]=true;
    		
    		//System.out.println(": ("+this.coordX[vIndex]+", "+this.coordY[vIndex]+") ");
    	}
    	//System.out.print("] ");

    	// set the x and y coordinates of the right-most top vertex
    	this.coordY[this.path.getFirst().getVertex().index]=0;
    	
    	// set the width of the layout (coordinate of the left-most top vertex (before rotation)
    	this.width=this.coordX[this.path.getLast().getOpposite().getVertex().index]; 

    	System.out.println("...done");
    }

    // -------- auxiliary methods for debuggin ---------
    
    public void print() {
    	int k=path.size()-2; // number of edges on the path (the first and last halfedges are discarded)
    	System.out.print("Path informations");
    	System.out.print(" - r="+this.removedVertices);
    	System.out.println(" - k="+k);
    	
    	// store the indices in reverse order
    	Halfedge[] halfedges=new Halfedge[k];
    	int counter=0;
    	for(Halfedge e: path) {
    		if(counter>0 && counter<=k)
    			halfedges[k-counter]=e; // store halfedges in reverse order (at the exception of the ones containing the source and destination vertices)
    		counter++;
    	}
    	
    	System.out.print("\t\t stretch  [");
    	String print="";
    	// set x and y coordinates
    	for(Halfedge e: halfedges) {
    		print=print+this.xStretch[e.index]+", ";
    	}
    	System.out.println(print+"] ");
    	
    	System.out.print("\t\t opposite [");
    	print="";
    	// set x and y coordinates
    	for(Halfedge e: halfedges) {
    		print=print+this.xStretchOtherSide[e.index]+", ";
    	}
    	System.out.println(print+"] ");

    	System.out.print("\t\t coordinates [");
    	print="";
    	// set x and y coordinates
    	for(Halfedge e: halfedges) {
    		print=print+"("+this.coordX[e.getVertex().index]+", "+this.coordY[e.getVertex().index]+")";
    	}
    	print=print+"("+this.coordX[this.path.getFirst().getVertex().index]+", "+this.coordY[this.path.getFirst().getVertex().index]+")";
    	System.out.println(print+"] ");
    	
    	System.out.println("\t\t width="+this.width);

    }

}
