package sd.layout2D;

import java.util.Arrays;
import java.util.Collection;

import Jama.Matrix;
import cern.colt.matrix.tdouble.DoubleMatrix2D;
import cern.colt.matrix.tdouble.impl.SparseDoubleMatrix2D;
import jdg.graph.AdjacencyListGraph;
import jdg.graph.Node;
import linalg.MatrixUtilities;
import linalg.JamaMatrices.JamaMatrix;
import linalg.MTJ.*;
import linalg.PColt.*;

import no.uib.cipr.matrix.sparse.CompRowMatrix;

public class TutteLaplacianMatrix {
	
	/** input graph to draw (with 'n' vertices) */
	public AdjacencyListGraph g;
	
	/** says whether a vertex in an inner vertex and thus must be drawn (boolean array of size 'n') */
	public boolean[] isInside;

	/** says whether a vertex belongs to the exterior boundary */
	public boolean[] isOnBoundary;
	
	/** number of inner vertices of the graph */
	public int nInnerVertices;

	/** number of the boundary vertices of the graph (on the outer cycle) */
	public int nBoundaryVertices;
	
	/** vertex numbering in the sub-graph consisting of inner vertices (array of size 'n') */
	public int[] vertexOrder;

	public TutteLaplacianMatrix(AdjacencyListGraph g, boolean[] isInside, boolean[] isOnBoundary, int nInnerVertices, int nBoundaryVertices, int[] vertexOrder) {
		this.g=g;
		this.isInside=isInside;
		this.isOnBoundary=isOnBoundary;
		
		this.nInnerVertices=nInnerVertices;
		this.nBoundaryVertices=nBoundaryVertices;
		this.vertexOrder=vertexOrder;
	}
	
	/**
	 * Computes and returns the Tutte laplacian matrix of the graph (Jama matrix)
	 */	
	public JamaMatrix computeJama() {
		System.out.print("Creating Laplacian (dense) Tutte laplacian matrix of size"+this.nInnerVertices+" (using Jama library)...");
		double[][] A=this.getDenseAdjacencyMatrix();
		double[][] D=this.getDenseDiagonalMatrix();
		Matrix adjacency=new Matrix(A);
		Matrix diagonal=new Matrix(D);
		
		Matrix Q=diagonal.minus(adjacency);
		
		System.out.println("done");
		return new JamaMatrix(Q.transpose());
	}

	/**
	 * Create the Laplacian matrix of a graph (with n vertices)
	 * (not efficient implementation: it takes nxn time)
	 * 
	 */	
	public MTJSparseMatrix computeMTJ() {
		CompRowMatrix A; // the result
		int n=this.nInnerVertices; // matrix size
		System.out.print("Creating Laplacian (sparse) matrix from a graph of size "+n+" (using MTJ library)...");
		
    	long startTime=System.nanoTime(), endTime; // for evaluating time performances
		
		int[][] m=new int[n][]; // indices of non-zero entries for each row
		
		for(Node u: this.g.vertices) {
			if(this.isInside[u.index]==true) { // consider only inner vertices
				int indexU=this.vertexOrder[u.index]; // index of vertex 'u' in the sub-graph: an integer in [0..n-k)
				
				int degreeU=u.degree();
				int[] neighbors=new int[degreeU]; // a non-zero entry for each neighbor + the diagonal entry
				
				int countInnerNeighbors=0;
				for(Node v: u.neighbors) { // visit all neighbors of vertex 'u'
					if(this.isInside[v.index]==true) { // check whether 'v' is an inner vertex
						int indexV=this.vertexOrder[v.index]; // index of vertex 'v' in the sub-graph: an integer in [0..n-k)
						neighbors[countInnerNeighbors]=indexV;
						countInnerNeighbors++;
					}
				}
				
				//MatrixUtilities.print(neighbors);
				
				m[indexU]=new int[countInnerNeighbors+1]; // store the non diagonal entries + one diagonal entry
				for(int i=0;i<countInnerNeighbors;i++) {
					m[indexU][i]=neighbors[i]; // set non diagonal entries corresponding to inner vertices (neighbors of vertex 'u')
				}
				m[indexU][countInnerNeighbors]=indexU; // add the diagonal entry
				
				Arrays.sort(m[indexU]); // entries must correspond to the increasing order of vertices
				//MatrixUtilities.print(m[indexU]);				
			}
		}
		
		A=new CompRowMatrix(n, n, m); // create a sparse matrix of size nxn

		for(Node u: this.g.vertices) {
			if(this.isInside[u.index]==true) { // consider only inner vertices
				int indexU=this.vertexOrder[u.index]; // index of vertex 'u' in the sub-graph: an integer in [0..n-k)
				
				// set the non-diagonal entries
				for(Node v: u.neighbors) { // visit all neighbors of vertex 'u'
					if(this.isInside[v.index]==true) { // check whether 'v' is an inner vertex
						int indexV=this.vertexOrder[v.index]; // index of vertex 'v' in the sub-graph: an integer in [0..n-k)
						A.set(indexU, indexV, -1.);
					}
				}
				
				int degreeU=u.degree();
				A.set(indexU, indexU, degreeU); // set the diagonal entries
			}
		}

    	endTime=System.nanoTime();
    	double duration=(double)(endTime-startTime)/1000000000.;
    	System.out.println("done ("+duration+" seconds)");
    	
		return new MTJSparseMatrix(A);
	}

	/**
	 * Create the Tutte laplacian matrix of a graph (PColt sparse matrix) <p>
	 */	
	public PColtMatrix computePColt() {
		DoubleMatrix2D L; // (sparse) matrix implementation based on Parallel Colt library
		int n=this.nInnerVertices; // matrix size
		System.out.print("Creating Laplacian matrix from a graph of size "+n+" (using Parallel Colt library)...");
		
    	long startTime=System.nanoTime(), endTime; // for evaluating time performances

		L=new SparseDoubleMatrix2D(n, n); // create a sparse matrix of size nxn

		for(Node u: this.g.vertices) {
			if(this.isInside[u.index]==true) { // consider only inner vertices
				int indexU=this.vertexOrder[u.index]; // index of vertex 'u' in the sub-graph: an integer in [0..n-k)
				L.setQuick(indexU, indexU, u.degree()); // set diagonal entries
				
				for(Node v: u.neighbors) { // visit all neighbors of vertex 'u'
					if(this.isInside[v.index]==true) { // check whether 'v' is an inner vertex
						int indexV=this.vertexOrder[v.index]; // index of vertex 'v' in the sub-graph: an integer in [0..n-k)
						L.setQuick(indexU, indexV, -1.); // store a '-1', when there is an edge between 'u' and 'v'
					}
				}
			}
		}
		
    	endTime=System.nanoTime();
    	double duration=(double)(endTime-startTime)/1000000000.;
    	System.out.println("done ("+duration+" seconds)");
    	
		return new PColtMatrix(L);
	}
	
	/**
	 * Compute and return the diagonal matrix corresponding to the (n-k) inner vertices
	 * 
	 * @return the (n-k)x(n-k) diagonal matrix storing the degrees of the inner vertices in the original (entire) graph G
	 */	
	public double[][] getDenseDiagonalMatrix() {
		double[][] m=new double[this.nInnerVertices][this.nInnerVertices];
		
		for(Node u: this.g.vertices) {
			if(this.isInside[u.index]==true) { // consider only inner vertices
				int subIndex=this.vertexOrder[u.index]; // index of vertex 'u' in the sub-graph: an integer in [0..n-k)
				m[subIndex][subIndex]=u.degree();
			}
		}
		
	   	return m;
	}

	/**
	 * Compute and return the adjacency matrix corresponding to the (n-k) inner vertices
	 * 
	 * @return the (n-k)x(n-k) adjacency matrix corresponding to the inner vertices
	 */	
	public double[][] getDenseAdjacencyMatrix() {
		double[][] m=new double[this.nInnerVertices][this.nInnerVertices];
		
		for(Node u: this.g.vertices) {
			if(this.isInside[u.index]==true) { // consider only inner vertices
				int indexU=this.vertexOrder[u.index]; // index of vertex 'u' in the sub-graph: an integer in [0..n-k)
			
				for(Node v: u.neighbors) { // visit all neighbors of vertex 'u'
					if(this.isInside[v.index]==true) { // check whether 'v' is an inner vertex
						int indexV=this.vertexOrder[v.index]; // index of vertex 'v' in the sub-graph: an integer in [0..n-k)
						m[indexU][indexV]=1.; // store a '1', when there is an edge between 'u' and 'v'
					}
				}
			}
		}
		
	   	return m;
	}


	/**
	 * Create the Laplacian matrix of a graph (with n vertices)
	 * (not efficient implementation: it takes nxn time)
	 * 
	 */	
	/*public MTJSparseMatrix testMTJ() {
		CompRowMatrix A; // the result
		int n=this.nInnerVertices; // matrix size
		System.out.print("Creating Laplacian (sparse) matrix from a graph of size "+n+" (using MTJ library)...");
		
    	long startTime=System.nanoTime(), endTime; // for evaluating time performances
		
		int[][] m=new int[n][]; // indices of non-zero entries for each row
		
		for(int i=0;i<n;i++) {
				
			int nonZeroEntries=?;
				m[i]=new int[nonZeroEntries]; // store the non diagonal entries + one diagonal entry
				for() {
					m[i][?]=?; // set non diagonal entries corresponding to inner vertices (neighbors of vertex 'u')
				}
			}
		}
		
		A=new CompRowMatrix(n, n, m); // create a sparse matrix of size nxn

		for() {
				A.set(i, j, valeur); // set the diagonal entries
		}

    	
		return new MTJSparseMatrix(A);
	}*/

}
