package sd.layout2D;

import java.util.Random;

import Jcg.geometry.Point_2;
import jdg.graph.AdjacencyListGraph;

/**
 * Planar layout (2D) based on the Tutte barycentric method
 * 
 * @author Luca Castelli Aleardi, Ecole Polytechnique
 * @version 2017
 */
public abstract class TutteLayout2D {
	/** input graph to draw (with 'n' vertices) */
	public AdjacencyListGraph g;
	
	/** says whether a vertex in an inner vertex and thus must be drawn (boolean array of size 'n') */
	public boolean[] isInside;

	/** says whether a vertex belongs to the exterior boundary */
	public boolean[] isOnBoundary;

	/** 2D positions of the vertices of the graph */
	public Point_2[] points;
	
	/** number of inner vertices of the graph */
	public int nInnerVertices=0;

	/** number of the boundary vertices of the graph (on the outer cycle) */
	public int nBoundaryVertices=0;

	public static int seed=10;
	/** Random generator */	
	static Random generator = new Random(seed); // initialize random generator
	
	/**
	 * Initialize the drawing
	 * 
	 * @param points  initial locations of the vertices (2D coordinates)
	 * @param isOnBoundary  an array of booleans: it says whether a vertex is on the outer boundary or not
	 * @param isInside  an array of booleans: it says whether a vertex is 'inside' the boundary (and thus it must be drawn)
	 */	
	public TutteLayout2D(AdjacencyListGraph g, Point_2[] points, boolean[] isOnBoundary, boolean[] isInside) {
		this.g=g;
		this.points=points;
		this.isOnBoundary=isOnBoundary;
		this.isInside=isInside;
		
		int n=this.g.sizeVertices();
		
		if(this.points==null)
			throw new Error("Vertex locations not defined");
		
		for(int i=0;i<points.length;i++) {
			if(points[i]==null && isOnBoundary[i]==true)
				throw new Error("Boundary vertex location not defined for vertex v"+i);
			else if(points[i]==null && isInside[i]==true)
				throw new Error("Inner vertex location not defined for vertex v"+i);
		}
		
		for(int i=0;i<n;i++) {
			if(isInside[i]==true)
				this.nInnerVertices++;
		}

		for(int i=0;i<n;i++) {
			if(isOnBoundary[i]==true)
				this.nBoundaryVertices++;
		}

	}
	
	/**
	 * Compute the Tutte drawing of a planar graph, by iteratively
	 * computing barycenters of neighboring vertices <p>
	 * <p>
	 * The algorithm stops after a given number of iterations
	 */	
	public abstract void computeLayout(int iterations);

	/**
	 * Compute the Tutte drawing of a planar graph: inner vertices are placed at the barycenter of its neighbors
	 * The algorithm stops when a given convergence condition is reached.
	 * 
	 * @param tolerance the numeric tolerance (defining the halt condition)
	 */	
	public abstract void computeLayout(double tolerance);
	
	/**
	 * Return the vertices of a regular polygon (equally spaced on a circle of radius r)
	 * 
	 * @param r the radius of the circle
	 * @param n the number of points on the outer cycle
	 * @return Point_2[] an array of 2D points, storing the vertices of a regular polygon
	 */	
	public static Point_2[] regularPolygonVertices(int n, double r) {
		Point_2[] vertices=new Point_2[n];
		double x,y;
		
		for(int i=0;i<n;i++) {
			x=r*Math.cos((2.*Math.PI/n)*i);
			y=r*Math.sin((2.*Math.PI/n)*i);
			vertices[i]=new Point_2(x,y);
		}
		return vertices;
	}

}
