package sd.layout2D;

import Jama.Matrix;
import Jcg.geometry.Point_2;
import jdg.graph.AdjacencyListGraph;
import jdg.graph.Node;
import linalg.JamaMatrices.JamaLinearSolver;
import linalg.JamaMatrices.JamaMatrix;
import linalg.MTJ.MTJIterativeLinearSolver;
import linalg.MTJ.MTJSparseMatrix;
import linalg.PColt.PColtLinearSolver;
import linalg.PColt.PColtMatrix;
import sd.util.MatrixUtilities;

/**
 * Implementation of the planar (2D) Tutte barycentric method, based on the resolution of 2 linear systems
 * 
 * @author Luca Castelli Aleardi, Ecole Polytechnique
 * @version 2017
 */
public class TutteLayout2DLinearSystem extends TutteLayout2D {

	/** vertex numbering in the sub-graph consisting of inner vertices (array of size 'n') */
	public int[] vertexOrder;
	
	/** choice of the solver for linear systems */
	public String linearSolver;

	/**
	 * Initialize the drawing
	 */	
	public TutteLayout2DLinearSystem(AdjacencyListGraph g, Point_2[] points, boolean[] isOnBoundary, boolean[] isInside, String linearSolver) {
		super(g, points, isOnBoundary, isInside); // class the constructor of the ancestor class
		
		int n=this.g.sizeVertices();
		
		vertexOrder=new int[n];
		
		int counter=0; // counter for inner vertices
		for(int i=0;i<n;i++) {
			if(this.isInside[i]==true) {
				vertexOrder[i]=counter; // store the permutation between vertex indices: from G to G1 (the sub-graph consisting of inner vertices)
				counter++;
			}
		}
		
		this.linearSolver=linearSolver;
	}

	/**
	 * Compute the Tutte drawing of a planar graph, by iteratively
	 * computing barycenters of neighboring vertices <p>
	 * <p>
	 * The algorithm stops after a given number of iterations
	 */	
	public void computeLayout(int iterations) {
		throw new Error("Not supported");
	}

	/**
	 * Compute the Tutte drawing of a planar graph, by iteratively
	 * computing barycenters of neighboring vertices <p>
	 * <p>
	 * The algorithm stops when a given convergence condition is reached
	 */	
	public void computeLayout(double tolerance) {
		if(this.nInnerVertices==0) // nothing to compute: there are no inner vertices (all vertices belong to the boundary cycle)
			return;
		
		double[] rightTermX=this.getRightTerm()[0];
		double[] rightTermY=this.getRightTerm()[1];
		
		double[][] solution=null;
		
		if(this.linearSolver.equals("Jama"))
			solution=this.solveLinearSystemJama(rightTermX, rightTermY); // an array containing the two vectors of size 'n' (for the x and y coordinates)
		else if(this.linearSolver.equals("PColt"))
			solution=this.solveLinearSystemPColt(rightTermX, rightTermY, tolerance); // an array containing the two vectors of size 'n' (for the x and y coordinates)
		else if(this.linearSolver.equals("MTJ"))
			solution=this.solveLinearSystemMTJ(rightTermX, rightTermY, tolerance); // an array containing the two vectors of size 'n' (for the x and y coordinates)
		else
			throw new Error("Error: solver for linear systems not supported");
				
		for(Node u: this.g.vertices) {
			if(this.isInside[u.index]==true) { // set coordinated only for inner vertices
				int index=this.vertexOrder[u.index]; // index of vertex 'u' in the sub-graph: an integer in [0..n-k)
				
				this.points[u.index]=new Point_2(solution[0][index], solution[1][index]);
			}
		}
		
	}

	/**
	 * Solve the linear system (A+D)x=b using Jama library <p>
	 * <p>
	 * Remark: this method is not suitable for a large linear system, since Jama matrices are dense and the solver is not fast
	 * 
	 * @param A  the adjacency matrix (for the sub-graph consisting of inner vertices)
	 * @param D  the diagonal matrix (for the original graph)
	 * @return the solution of the linear system (A+D)x=b
	 */	
	public double[][] solveLinearSystemJama(double[] x, double y[]) {
    	TutteLaplacianMatrix laplacian=new TutteLaplacianMatrix(this.g, this.isInside, this.isOnBoundary, 
    															this.nInnerVertices, this.nBoundaryVertices, this.vertexOrder);
    	
    	long startTime=System.nanoTime(), endTime; // for evaluating time performances
    	JamaMatrix L=laplacian.computeJama();
    	JamaLinearSolver solver=new JamaLinearSolver(L);

    	System.out.print("Solving linear system with Jama...");
   	
    	double[] solutionX=solver.solve(x); // solve the linear system Lx=b
    	double[] solutionY=solver.solve(y); // solve the linear system Ly=b
		
    	// evaluate time performances
    	endTime=System.nanoTime();
    	double duration=(double)(endTime-startTime)/1000000000.;
    	System.out.println("done ("+duration+" seconds)");
    	
		return new double[][] {solutionX, solutionY};
	}

	/**
	 * Solve the linear system (A+D)x=b using Jama library <p>
	 * <p>
	 * Remark: this method is not suitable for a large linear system, since Jama matrices are dense and the solver is not fast
	 * 
	 * @param x  right term (x-coordinates)
	 * @param y  right term (y coordinates)
	 * @return an array with two rows (for x and y coordinates), storing the solution (coordinates of the 'n' inner vertices)
	 */	
	public double[][] solveLinearSystemPColt(double[] x, double y[], double tolerance) {
    	TutteLaplacianMatrix laplacian=new TutteLaplacianMatrix(this.g, this.isInside, this.isOnBoundary, 
    															this.nInnerVertices, this.nBoundaryVertices, this.vertexOrder);
    	
    	long startTime=System.nanoTime(), endTime; // for evaluating time performances
    	PColtMatrix L=laplacian.computePColt();
    	PColtLinearSolver solver=new PColtLinearSolver(L, tolerance);
    	
		System.out.print("Solving linear system with PColt...");
		
    	double[] solutionX=solver.solve(x); // solve the linear system Lx=b
    	double[] solutionY=solver.solve(y); // solve the linear system Ly=b
		
    	// evaluate time performances
    	endTime=System.nanoTime();
    	double duration=(double)(endTime-startTime)/1000000000.;
    	System.out.println("done ("+duration+" seconds) [tolerance= "+tolerance+"]");
    	
		return new double[][] {solutionX, solutionY};
	}

	/**
	 * Solve the linear system (A+D)x=b using MTJ library <p>
	 * <p>
	 * Remark: this method is not suitable for a large linear system, since Jama matrices are dense and the solver is not fast
	 * 
	 * @param x  right term (x-coordinates)
	 * @param y  right term (y coordinates)
	 * @return an array with two rows (for x and y coordinates), storing the solution (coordinates of the 'n' inner vertices)
	 */	
	public double[][] solveLinearSystemMTJ(double[] x, double y[], double tolerance) {
    	TutteLaplacianMatrix laplacian=new TutteLaplacianMatrix(this.g, this.isInside, this.isOnBoundary, 
    															this.nInnerVertices, this.nBoundaryVertices, this.vertexOrder);
    	
    	long startTime=System.nanoTime(), endTime; // for evaluating time performances
    	MTJSparseMatrix L=laplacian.computeMTJ();
    	MTJIterativeLinearSolver solver=new MTJIterativeLinearSolver(L, tolerance);
    	
		System.out.print("Solving linear system with MTJ...");
		
    	double[] solutionX=solver.solve(x); // solve the linear system Lx=b
    	double[] solutionY=solver.solve(y); // solve the linear system Ly=b
		
    	// evaluate time performances
    	endTime=System.nanoTime();
    	double duration=(double)(endTime-startTime)/1000000000.;
    	System.out.println("done ("+duration+" seconds) [tolerance= "+tolerance+"] ");
    	
		return new double[][] {solutionX, solutionY};
	}

	/**
	 * Compute and return the vector right term given by the locations of the boundary vertices (on the outer cycle)
	 * 
	 * @return the (n-k)x(n-k) adjacency matrix corresponding to the inner vertices
	 */	
	public double[][] getRightTerm() {
		double[][] result=new double[2][this.nInnerVertices];
		for(Node u: this.g.vertices) {
			if(this.isInside[u.index]==true) { // consider only inner vertices
				int indexU=this.vertexOrder[u.index]; // index of vertex 'u' in the sub-graph: an integer in [0..n-k)
			
				double x=0., y=0.;
				for(Node v: u.neighbors) { // visit all neighbors of vertex 'u'
					if(this.isOnBoundary[v.index]==true) { // check whether 'v' is an inner vertex
						int indexV=this.vertexOrder[v.index]; // index of vertex 'v' in the sub-graph: an integer in [0..n-k)
						x=x+this.points[v.index].getX().doubleValue();
						y=y+this.points[v.index].getY().doubleValue();
					}
				}
				
				result[0][indexU]=x;
				result[1][indexU]=y;
			}
		}
		System.out.println();
		
		return result;
	}

}
