package sd.util;

import java.awt.Color;
import java.util.ArrayList;
import java.util.LinkedList;
import java.util.List;
import java.util.Random;

import Jcg.geometry.Point_;
import Jcg.geometry.Point_3;
import Jcg.polyhedron.Face;
import Jcg.polyhedron.Halfedge;
import Jcg.polyhedron.Polyhedron_3;
import Jcg.polyhedron.Vertex;
import jdg.graph.AdjacencyListGraph;
import jdg.graph.Node;
import sd.graphalgorithms.DisjointPaths;
import sd.graphalgorithms.VertexPath;
import tc.TC;

/**
 * A class providing implementation of several graph algorithms on triangle meshes<p>
 * 
 * @author Luca Castelli Aleardi, Ecole Polytechnique
 * @version april 2017
 */
public class MeshAlgorithms {
	
	/** Random generator */	
	static Random generator=new Random(1);

    /**
     * Compute a (BFS) shortest path between two vertices v and w
     * It performs a BFS visit of the mesh starting from vertex v
     * <p>
     * Remark: the vertices are assumed to have an index, between 0..n-1
     * 
     * @param v  the starting vertex
     * @param w  the destination
     * @return the list of vertices on the path between v and w
     */		   
    public static VertexPath computeShortestPath(Polyhedron_3 mesh, Vertex v, Vertex w){
    	if(v==null || w==null)
    		return null;
    	
    	VertexPath path=new VertexPath(); // the shortest path between v and w (to return)
    	int n=mesh.sizeOfVertices(); // number of vertices in the mesh
    	
    	int[] distance=new int[n]; // it stores, for each vertex the distance from the source
    	Vertex[] previous=new Vertex[n];
    	LinkedList<Vertex> queue=new LinkedList<Vertex>(); // stack containing the node to visit
    	boolean reached=false;
    	
    	for(int i=0;i<n;i++)
    		distance[i]=-1; // unvisited vertices have distance -1 from the source vertex 'v'
    	
    	distance[v.index]=0;
    	previous[v.index]=null; // the first vertex has no previous vertex
    	queue.add(v);
    	while(queue.isEmpty()==false && reached==false) {    			
    		Vertex u=queue.poll(); // get and removes the node in the head of the stack
    		//System.out.println("\n current node: "+u.index);

    		if(u==w) {
    			reached=true;
    		}
    		
    		else{
    			List<Halfedge> neighbors=u.getOutgoingHalfedges();
    			for(Halfedge e: neighbors) { // visit incident (outgoing) halfedges around 'u'
    				Vertex neighbor=e.getVertex();
    				if(distance[neighbor.index]==-1) {
    					distance[neighbor.index]=distance[u.index]+1; // increment the distance by 1
    					previous[neighbor.index]=u;
    					queue.add(neighbor);
    				}
    			}
    		}
    	}
    	
    	// reconstruct the shortest path, from distances
    	Vertex u=w; // start from last vertex 'w'
    	while(previous[u.index]!=null) { // traverse the path in reverse order
    		path.addFirst(u);
    		u=previous[u.index];
    	}
    	path.addFirst(u); // add the first vertex 'v'
    	
    	return path;
    }

    /**
     * Compute, for all vertices, the distance from a given set of 'seed vertices'
     * It performs a BFS visit of the mesh starting from the 'seed vertices'
     * <p>
     * Remark: the vertices are assumed to have an index, between 0..n-1
     * 
     * @param mesh  the input triangle mesh
     * @param set  the set of 'seed vertices'
     * @return  an array of integers storing, for each vertex, the distance from the 'seed vertices'
     */		   
    public static int[] getDistancesFromVertexSet(Polyhedron_3 mesh, List<Vertex> set){
    	if(set==null)
    		return null;
    	
    	int n=mesh.sizeOfVertices(); // number of vertices in the mesh
    	
    	int[] distance=new int[n]; // it stores, for each vertex the distance from the 'seed vertices'
    	LinkedList<Vertex> queue=new LinkedList<Vertex>(); // stack containing the node to visit
    	
    	for(int i=0;i<n;i++)
    		distance[i]=-1; // unvisited vertices have distance -1 from the 'seed vertices'

    	for(Vertex v: set) {
    		distance[v.index]=0; // set the distance to '0', for all seed vertices
    		queue.add(v); // add each seed vertex to the queue
    	}
    	
    	while(queue.isEmpty()==false) {    			
    		Vertex u=queue.poll(); // get and removes the node in the head of the stack
    		//System.out.println("\n current node: "+u.index);
    		
    			List<Halfedge> neighbors=u.getOutgoingHalfedges();
    			for(Halfedge e: neighbors) { // visit incident (outgoing) halfedges around 'u'
    				Vertex neighbor=e.getVertex();
    				if(distance[neighbor.index]==-1) {
    					distance[neighbor.index]=distance[u.index]+1; // increment the distance by 1
    					queue.add(neighbor);
    				}
    			}

    	}
    	
    	return distance;
    }

    /**
     * Compute, fro each vertex, the (graph) distance to vertex v
     * It performs a BFS visit of the entire mesh starting from vertex v
     * <p>
     * Remark: the vertices are assumed to have an index, between 0..n-1
     * 
     * @param v  the starting vertex
     * @return an array containing the (integer) distances from vertex v
     */		   
    public static int[] computeAllDistancesFromVertex(Polyhedron_3 mesh, Vertex v){
    	if(v==null)
    		return null;
    	
    	int n=mesh.sizeOfVertices(); // number of vertices in the mesh
    	
    	int[] distance=new int[n]; // it stores, for each vertex the distance from the source
    	LinkedList<Vertex> queue=new LinkedList<Vertex>(); // stack containing the node to visit
    	
    	for(int i=0;i<n;i++)
    		distance[i]=-1; // unvisited vertices have distance -1 from the source vertex 'v'
    	
    	distance[v.index]=0;
    	queue.add(v);
    	while(queue.isEmpty()==false) {    			
    		Vertex u=queue.poll(); // get and removes the node in the head of the stack
    		//System.out.println("\n current node: "+u.index);

    		List<Halfedge> neighbors=u.getOutgoingHalfedges();
    		for(Halfedge e: neighbors) { // visit incident (outgoing) halfedges around 'u'
    			Vertex neighbor=e.getVertex();
    			if(distance[neighbor.index]==-1) {
    				distance[neighbor.index]=distance[u.index]+1; // increment the distance by 1
    				queue.add(neighbor);
    			}
    		}
    	}
    	
    	return distance;
    }

    /**
     * Given a separating cycle 'C' (in a planar triangle mesh), it compute all vertices inside the region delimited by 'C' <p>
     * <p>
     * The algorithm perform a BFS visit of the mesh, starting from a vertex "at the left" of the cycle<p>
     * <p>
     * Remarks: <p>
     * 'inner vertices' (inside the cycle) are those who are accessible "from the left" starting from the cycle 'C'<p>
     * The input mesh is assumed to be a genus 0 triangle mesh (without boundary)
     * <p>
     * Remark: the vertices are assumed to have an index, between 0..n-1
     * 
     * @param cycle  a collection of edges (halfedges, having the same orientation), providing a separating cycle
     * @return an array of booleans: an entry is true iff the corresponding vertex is inside 'C'
     */		   
    public static boolean[] getInnerVertices(Polyhedron_3 mesh, LinkedList<Halfedge> cycle){
    	if(cycle==null)
    		throw new Error("Error: the cycle is not defined");
    	
    	System.out.print("Computing sub-graph...");
    	long startTime=System.nanoTime(), endTime; // for evaluating time performances

    	int n=mesh.sizeOfVertices(); // number of vertices in the mesh
    	
    	boolean[] inside=new boolean[n]; // the result to return
    	boolean[] onCycle=new boolean[n]; // mark the vertices on the cycle C
    	boolean[] visited=new boolean[n]; // say whether a vertex is visited during the traversal
    	LinkedList<Vertex> queue=new LinkedList<Vertex>(); // stack containing the node to visit
    	
    	for(Halfedge e: cycle)
    		onCycle[e.getVertex().index]=true; // vertices on the cycle are marked
    	
    	// find a first vertex to start the traversal (the vertex must be inside the region)
    	Vertex v=null;
    	Halfedge ePrevious=cycle.getLast();
    	

    	for(Halfedge e: cycle) {
    		Halfedge pEdge=e; // visit the edges inside the cycle, turning around the source vertex of 'e' in ccw order
    		while(pEdge!=ePrevious.getOpposite()) { // add all inner vertices which are at distance 1 from the cycle
    			Vertex u=pEdge.getPrev().getPrev().getVertex();
    			if(onCycle[u.index]==false) {
    				v=u; // a first inner vertex is found
    				queue.add(v);
    				visited[v.index]=true;
    			}
    			
    			pEdge=pEdge.getPrev().getOpposite();
    		}
    		ePrevious=e;
    	}
    	
    	while(queue.isEmpty()==false) {
    		Vertex u=queue.poll(); // get and removes the node in the head of the stack
    		inside[u.index]=true;
    		//System.out.println("\n current node: "+u.index);

    		List<Halfedge> neighbors=u.getOutgoingHalfedges();
    		for(Halfedge e: neighbors) { // visit incident (outgoing) halfedges around 'u'
    			Vertex neighbor=e.getVertex();
    			if(visited[neighbor.index]==false && onCycle[neighbor.index]==false) {
    				visited[neighbor.index]=true; // mark the vertex as visited
    				queue.add(neighbor);
    				inside[u.index]=true;
    			}
    		}
    	}
    	
    	int count=0; // count the number of inner vertices
    	for(int i=0;i<n;i++)
    		if(inside[i]==true)
    			count++;
    	
    	// evaluate time performances
    	endTime=System.nanoTime();
    	double duration=(double)(endTime-startTime)/1000000000.;
    	System.out.print("done ("+duration+" seconds) - ");

    	System.out.println("["+count+" vertices, and a boundary of size "+cycle.size()+"]");
    	return inside;
    }

    /**
     * Given a separating cycle 'C' (in a planar triangle mesh), it compute all edges inside the region delimited by 'C' <p>
     * <p>
     * The algorithm perform a BFS visit of the mesh, starting from a vertex "at the left" of the cycle<p>
     * <p>
     * Remarks: <p>
     * 'inner edges' (inside the cycle) are those who are accessible "from the left" starting from the cycle 'C'<p>
     * The input mesh is assumed to be a genus 0 triangle mesh (without boundary)
     * <p>
     * Remark: the vertices are assumed to have an index, between 0..n-1
     * 
     * @param cycle  a collection of edges (halfedges, having the same orientation), providing a separating cycle
     * @return an array of booleans: an entry is true iff the corresponding halfedge is inside 'C'
     */		   
    public static boolean[] getInnerEdges(Polyhedron_3 mesh, LinkedList<Halfedge> cycle, boolean[] innerVertices){
    	if(cycle==null)
    		throw new Error("Error: the cycle is not defined");
    	
    	System.out.print("Computing sub-graph...");
    	
    	int n=mesh.sizeOfVertices(); // number of vertices in the mesh
    	int nE=mesh.sizeOfHalfedges(); 
    	
    	boolean[] result=new boolean[nE]; // the result to return
    	
    	// first part: mark all edges incident to inner vertices
    	for(Vertex v: (List<Vertex>)mesh.vertices) {
    		if(innerVertices[v.index]==true) {
    			for(Halfedge e: (List<Halfedge>)v.getOutgoingHalfedges()) {
    				result[e.index]=true;
    				result[e.getOpposite().index]=true;
    			}
    		}
    	}
    	
    	
    	// second part: mark all 'inner' chords (chords which are at the left of the cycle
    	
    	// mark vertices on the separating cycle
    	boolean[] onCycle=new boolean[n]; // mark the vertices on the cycle C
    	for(Halfedge e: cycle)
    		onCycle[e.getVertex().index]=true; // vertices on the cycle are marked
    	
    	Halfedge ePrevious=cycle.getLast();
    	for(Halfedge e: cycle) {
    		// visit the edges inside the cycle (at its left), turning around the source vertex of 'e' in ccw order
    		Halfedge pEdge=e.getPrev(); 
    		while(pEdge!=ePrevious) { // add all inner vertices which are at distance 1 from the cycle
    			// mark the halfedge and its opposite as inner edge
    			result[pEdge.index]=true;
    			result[pEdge.getOpposite().index]=true;
    			
    			pEdge=pEdge.getOpposite().getPrev();
    		}
    		
    		ePrevious=e;
    	}
    	
    	int count=0; // count the number of inner vertices
    	for(int i=0;i<nE;i++)
    		if(result[i]==true)
    			count++;
    	
    	System.out.println("done ("+(count/2)+" edges, and a boundary of size "+cycle.size()+")");
    	return result;
    }

    /**
     * Return the farthest vertex from v
     * <p>
     * Remark: the vertices are assumed to have an index, between 0..n-1
     * 
     * @param v  the starting vertex
     * @return the farthest vertex from v
     */		   
    public static Vertex getFarthestVertex(Polyhedron_3 mesh, Vertex v){
    	if(v==null)
    		return null;
    	
    	int[] distances=computeAllDistancesFromVertex(mesh, v); // perform a BFS traversal of the mesh, starting from v
    	
    	int maxDistance=0;
    	Vertex farthest=v;
    	
    	List<Vertex> vertices=mesh.vertices;
    	for(Vertex x: vertices) {
    		if(distances[x.index]>maxDistance) { // update the max distance
    			maxDistance=distances[x.index];
    			farthest=x;
    		}
    	}
    	
    	return farthest;
    }

    /**
     * Return the distance between a vertex 'v' and a set of vertices S
     * <p>
     * Remark: the vertices are assumed to have an index, between 0..n-1
     * 
     * @param v  the starting vertex
     * @param vertexSet  the set of vertices 'S'
     * @return  the minimal distance between v and all vertices in 'S'
     */		   
    public static int getDistanceFromSet(Polyhedron_3 mesh, Vertex v, List<Vertex> vertexSet){
    	if(v==null || vertexSet==null) {
    		throw new Error("Error: input vertex or vertex set not defined");
    	}
    	
    	int[] distances=computeAllDistancesFromVertex(mesh, v); // perform a BFS traversal of the mesh, starting from v
    	
    	int minDistance=Integer.MAX_VALUE;
    	
    	for(Vertex x: vertexSet) {
    		if(distances[x.index]<minDistance) { // update the max distance
    			minDistance=distances[x.index];
    		}
    	}
    	
    	return minDistance;
    }

    /**
     * Compute a (BFS) shortest path 'p2' between two vertices v and w, disjoint from path 'p_1 ... p_k'
     * It performs a BFS visit of the mesh starting from vertex v, avoiding vertices on paths 'p_1 ... p_k'
     * <p>
     * Remark: the vertices are assumed to have an index, between 0..n-1<p>
     * Remark: the first and last vertices on the path 'p1' are assumed to be v and w
     * 
     * @param v  the starting vertex
     * @param w  the destination
     * @param p1 a collection of k vertex paths to avoid
     * @return  the list of vertices on the path between v and w
     */		   
    public static VertexPath computeShortestDisjointPath(Polyhedron_3 mesh, Vertex v, Vertex w, ArrayList<VertexPath> avoidedPaths){
    	if(v==null || w==null)
    		return null;
    	
    	VertexPath path=new VertexPath(); // the shortest path between v and w (to return)
    	int n=mesh.sizeOfVertices(); // number of vertices in the mesh
    	
    	int[] distance=new int[n]; // it stores, for each vertex the distance from the source
    	Vertex[] previous=new Vertex[n];
    	LinkedList<Vertex> queue=new LinkedList<Vertex>(); // stack containing the node to visit
    	boolean reached=false;
    	
    	boolean[] toAvoid=new boolean[n]; // says whether a vertex is to be avoided (during the graph traversal)
    	
    	for(VertexPath p1: avoidedPaths) {
    		for(Vertex x: p1.getVertices())
    			toAvoid[x.index]=true; // the vertex belongs to the path p1, and thus must not be visited
    	}
    	toAvoid[v.index]=true; // vertex u is the source of the path
    	toAvoid[w.index]=false; // vertex w must be reached at the end
    	
    	for(int i=0;i<n;i++)
    		distance[i]=-1; // unvisited vertices have distance -1 from the source vertex 'v'
    	
    	distance[v.index]=0;
    	previous[v.index]=null; // the first vertex has no previous vertex
    	queue.add(v);
    	while(queue.isEmpty()==false && reached==false) {    			
    		Vertex u=queue.poll(); // get and removes the node in the head of the stack
    		//System.out.println("\n current node: "+u.index);

    		if(u==w) {
    			reached=true;
    		}
    		
    		else{
    			List<Halfedge> neighbors=u.getOutgoingHalfedges();
    			for(Halfedge e: neighbors) { // visit incident (outgoing) halfedges around 'u'
    				Vertex neighbor=e.getVertex();
    				if(distance[neighbor.index]==-1 && toAvoid[neighbor.index]==false) { // visit the vertex, only if it does not belong to p1
    					distance[neighbor.index]=distance[u.index]+1; // increment the distance by 1
    					previous[neighbor.index]=u;
    					queue.add(neighbor);
    				}
    			}
    		}
    	}
    	
    	// reconstruct the shortest path, from distances
    	Vertex u=w; // start from last vertex 'w'
    	while(previous[u.index]!=null) { // traverse the path in reverse order
    		path.addFirst(u);
    		u=previous[u.index];
    	}
    	path.addFirst(u); // add the first vertex 'v'
    	
    	return path;
    }
    
	/**
	 * Reset indices for all: vertices, faces and halfedges
	 */	
	public static void resetMeshIndices(Polyhedron_3 mesh) {
		int count=0;
		for(Vertex v: (List<Vertex>)mesh.vertices) {
			v.index=count;
			count++;
		}
		
		count=0;
		for(Face f: (List<Face>)mesh.facets) {
			f.index=count; // standard face color
			count++;
		}
		
		count=0;
		for(Halfedge e: (List<Halfedge>)mesh.halfedges) {
			e.index=count;
			count++;
		}
	}

	/**
	 * Return the index of the two farthest (triangle) faces, according to euclidean distance<p>
	 * The computing distance is between the barycenters of the faces
	 * 
	 * @param faces  collection of triangle faces
	 * @return the two farthest faces
	 */
	public static Face[] getFarthestFaces(ArrayList<Face<Point_3>> faces) {
		Face<Point_3>[] result=new Face[2];

		double distance=0.0;		
		int i=0;
		for(Face<Point_3> f1: faces) {
			Point_3 bar1=(Point_3)f1.getBarycenter();
			for(int j=i+1;j<faces.size();j++) {
				Face<Point_3> f2=faces.get(j);
				Point_3 bar2=(Point_3)f2.getBarycenter();
				if(bar1.squareDistance(bar2).doubleValue()>distance) {
					result[0]=f1;
					result[1]=f2;
					distance=bar1.squareDistance(bar2).doubleValue();
				}
			}
		}
		return result;
	}
	
	/**
	 * Return the index of the two farthest faces in the mesh, according to euclidean distance<p>
	 * The computing distance is between the barycenters of the faces.
	 * 
	 * The computation is based on a random sampling of faces and returns an approximate result
	 * 
	 * @param mesh  input triangulation (triangle mesh)
	 * @param r  size of the random sampling (a number in [0, 1])
	 * @return  the two farthest faces
	 */
	public static Face[] getFarthestFacesGeometric(Polyhedron_3<Point_3> mesh, double r) {
		System.out.print("Computing two farthest faces...");
		if(r<0 || r>1 || mesh==null)
			throw new Error("Error: wrong input parameters ");
		
		ArrayList<Face<Point_3>> randomSample=new ArrayList<Face<Point_3>>(); // random sample of the faces
		for(Face<Point_3> f: mesh.facets) {
			double randomNumber=generator.nextDouble();
			if(randomNumber<=r)
				randomSample.add(f);
		}
		
		Face[] result=getFarthestFaces(randomSample);
		System.out.println("done");
		int v1=result[0].getEdge().getVertex().index;
		int v2=result[0].getEdge().getNext().getVertex().index;
		int v3=result[0].getEdge().getNext().getNext().getVertex().index;
		System.out.println("f1= "+v1+", "+v2+", "+v3);
		v1=result[1].getEdge().getVertex().index;
		v2=result[1].getEdge().getNext().getVertex().index;
		v3=result[1].getEdge().getNext().getNext().getVertex().index;
		System.out.println("f2= "+v1+", "+v2+", "+v3);
		
		boolean adjacent=areAdjacent(result[0], result[1], mesh);
		if(adjacent)
			System.out.println("The two faces are adjacent");
		else
			System.out.println("The two faces are not adjacent");
		
		return result;
	}

	/**
	 * Check whether two faces are adjacent in the mesh (whether they share a vertex)
	 * 
	 * @param f1  first triangle face
	 * @param f2  second triangle face
	 * @param mesh  input triangle mesh
	 * @return true if and only if f1 and f2 share a vertex
	 */
	public static boolean areAdjacent(Face f1, Face f2, Polyhedron_3 mesh) {
		int[] v1=f1.getVertexIndices(mesh); // vertices of f1
		int[] v2=f2.getVertexIndices(mesh); // vertices of f2
		
		if(v1[0]==v2[0] || v1[0]==v2[1] || v1[0]==v2[2])
			return true;
		if(v1[1]==v2[0] || v1[1]==v2[1] || v1[1]==v2[2])
			return true;
		if(v1[2]==v2[0] || v1[2]==v2[1] || v1[2]==v2[2])
			return true;
		return false; // no vertex is shared
	}

	/**
	 * Check whether two faces f1 and f2 share an edge: in the positive, return the common half-edge inside f1
	 * 
	 * @param f1  first triangle face
	 * @param f2  second triangle face
	 * @return true if and only if f1 and f2 share a vertex
	 */
	public static Halfedge getCommonHalfedge(Face f1, Face f2) {
		if(f1==null) {
			System.err.println("Error: face 1 is not defined");
			return null;
		}
		if(f2==null) {
			System.err.println("Error: face 2 is not defined");
			return null;
		}
		
		Halfedge e1=f1.getEdge();
		Halfedge e2=f2.getEdge();
		
		// find the common pair of half-edges (shared by the pair of faces)
		if(e1==e2.getOpposite())
			return e1;
		else if(e1==e2.getNext().getOpposite())
			return e1;
		else if(e1==e2.getPrev().getOpposite())
			return e1;
		
		else if(e1.getNext()==e2.getOpposite())
			return e1.getNext();
		else if(e1.getNext()==e2.getNext().getOpposite())
			return e1.getNext();
		else if(e1.getNext()==e2.getPrev().getOpposite())
			return e1.getNext();
		
		else if(e1.getPrev()==e2.getOpposite())
			return e1.getPrev();
		else if(e1.getPrev()==e2.getNext().getOpposite())
			return e1.getPrev();
		else if(e1.getPrev()==e2.getPrev().getOpposite())
			return e1.getPrev();
		
		else
			return null; // the two faces do not share an edge
	}

	/**
	 * Return the indices of the three vertices in a face <p>
	 * Warning: it only works for triangle meshes
	 * 
	 * @param mesh  input triangulation (triangle mesh)
	 * @param face  index of a face, in [0..f-1]
	 * @return the three indices of the vertices of the face
	 */
	public int[] getVerticesInFace(Polyhedron_3<Point_3> mesh, int face) {
		if(face>=mesh.sizeOfFacets())
			throw new Error("Error: wrong face index "+face);
		
		int v1=mesh.facets.get(face).getEdge().getVertex().index;
		int v2=mesh.facets.get(face).getEdge().getNext().getVertex().index;
		int v3=mesh.facets.get(face).getEdge().getNext().getNext().getVertex().index;
		
		return new int[]{v1, v2, v3};
	}

	/**
	 * Return the three 3d points of a triangle face
	 * 
	 * @param face  3d triangle face
	 * @return the three points (corresponding to the vertices of the triangle face)
	 */
	public Point_3[] getVertexCoordinatesInFace(Face<Point_3> face) {
		if(face==null)
			throw new Error("Error: face not defined");
		
		Point_3 p1=face.getEdge().getVertex().getPoint();
		Point_3 p2=face.getEdge().getNext().getVertex().getPoint();
		Point_3 p3=face.getEdge().getNext().getNext().getVertex().getPoint();
		
		return new Point_3[]{p1, p2, p3};
	}
	
	/**
	 * Given a vertex, return the barycenter of its three neighbors <p>
	 * <p>
	 * Warning: it only works for a vertex of degree 3
	 * 
	 * @param v  a vertex having degree 3
	 * @return barycenter (3d point) of the three neighbors of the input vertex
	 */
	public static Point_3 getNeighborsBarycenter(Vertex<Point_3> v) {
		if(v==null)
			throw new Error("Error: vertex not defined");
		
		List<Halfedge> neighbors=v.getOutgoingHalfedges();
		
		Vertex v0=neighbors.get(0).getVertex();
		Vertex v1=neighbors.get(1).getVertex();
		Vertex v2=neighbors.get(2).getVertex();
		
		Point_3[] points=new Point_3[3];
		points[0]=(Point_3)v0.getPoint();
		points[1]=(Point_3)v1.getPoint();
		points[2]=(Point_3)v2.getPoint();
		
		Point_3 bar=new Point_3(0., 0., 0.);
		bar.barycenter(points);
		return bar;
	}

    /**
     * Output a 3D (surface) triangulation to OFF file
     * 
     * @param mesh  3D triangle mesh (Half-edge representation)
     * 
     * warning: only for triangle meshes (faces must be triangular)
     */
    public static void writePolyhedronToOFF(Polyhedron_3 mesh, String filename, boolean withEdgeColors, boolean withFaceColors){
		Color[] colors=new Color[12];
		colors[0]=new Color(240, 240, 240, 255); // no color defined: gray by default
		colors[1]=Color.red;
		colors[2]=Color.blue;
		colors[3]=Color.black;
		colors[4]=Color.green;
		colors[5]=Color.orange;
		colors[6]=Color.green;
		colors[7]=Color.yellow;
		colors[8]=Color.magenta;
		colors[9]=new Color(240, 240, 240, 255); // standard color for rendering faces
		colors[10]=new Color(0, 128, 0, 255); // dark green
		colors[11]=new Color(100, 100, 100, 255); // dark gray

    	System.out.print("Creating OFF file from Polyhedron...");
    	Jcg.io.IO.writeNewTextFile(filename);
    	
    	int nVertices=mesh.vertices.size();
    	int nFaces=mesh.facets.size();
    	int nEdges=mesh.sizeOfHalfedges()/2;
    	
    	int nElements=0;
    	if(withEdgeColors==true)
    		nElements=nFaces+nEdges;
    	else
    		nElements=nFaces;
    	
    	// file header
    	Jcg.io.IO.println("OFF");
    	Jcg.io.IO.println(nVertices+" "+nElements+" "+0);
    	    	
    	// writing vertex coordinates
    	int i=0;
    	for(Object o: mesh.vertices) { // iterate over all vertices
    		Vertex<Point_> v=(Vertex<Point_>)o;
    		v.index=i;
    		Point_ p=v.getPoint();
    		double x, y, z;
    		x=p.getCartesian(0).doubleValue();
    		y=p.getCartesian(1).doubleValue();
    		z=p.getCartesian(2).doubleValue();
    		
    		Jcg.io.IO.println(""+x+" "+y+" "+z);
    		i++;
    	}
    	
      	// write face/vertex incidence relations
    	for(Object o: mesh.facets){ // iterate over all faces
      		Face<Point_> f=(Face<Point_>)o;
    		int index0=f.getEdge().vertex.index;
    		int index1=f.getEdge().getNext().vertex.index;
    		int index2=f.getEdge().getNext().getNext().vertex.index;
    		String faceColor="";
    		if(f.tag>=0 && withFaceColors==true)
    			faceColor=""+colors[f.tag].getRed()/255.+" "+colors[f.tag].getGreen()/255.+" "+colors[f.tag].getBlue()/255.+" "+colors[f.tag].getAlpha()/255.;
    		Jcg.io.IO.println(3+" "+index0+" "+index1+" "+index2+" "+faceColor);
    		//System.out.println("f"+f.index+", tag "+f.tag);
    	}
    	
    	if(withEdgeColors==true) { // write edge colors (if defined) to OFF file
        	for(Object o: mesh.halfedges){ // iterate over all halfedges
          		Halfedge e=(Halfedge)o;
        		int index0=e.getVertex().index;
        		int index1=e.getOpposite().getVertex().index;
        		
        		int eIndex=e.index;
        		String edgeColor=""+colors[e.tag].getRed()/255.+" "+colors[e.tag].getGreen()/255.+" "+colors[e.tag].getBlue()/255.+" "+colors[e.tag].getAlpha()/255.;
        		if(index0<index1) // each edge is written only once
        			Jcg.io.IO.println(2+" "+index0+" "+index1+" "+edgeColor);
        	}
    	}
    	
    	Jcg.io.IO.writeStandardOutput();
    	System.out.println("done ("+filename+")");
    }

    /**
     * Output a 3D (surface) triangulation to colored OFF file: both edges and faces can be provided with colors
     * 
     * @param mesh  3D triangle mesh (Half-edge representation)
     * 
     * warning: only for triangle meshes (faces must be triangular)
     */
    public static void writeColoredPolyhedronToOFF(Polyhedron_3 mesh, String filename){
		String [] colors=new String[12];
		colors[0]=null; // no color defined: gray by default
		colors[1]=" 1 0 0 1"; // red
		colors[2]=" 0 0 1 1"; // blue
		colors[3]=" 0 0 0 1"; // black
		colors[4]=" 0 1 0 1"; // green
		colors[5]=" 1 0.6 0 1"; // orange
		colors[6]=" 0 1 0 1"; // green
		colors[7]=" 1 1 0 1"; // yellow
		colors[8]=" 0.5 0 0.5 1"; // magenta
		colors[9]=" 0.9 0.9 0.9 1"; // standard color for rendering faces
		colors[10]=" 0 0.5 1 1"; // dark green
		colors[11]=" 0.4 0.4 0.4 1"; // dark gray
		
    	int nVertices=mesh.vertices.size();
    	int nFaces=mesh.facets.size();
    	int nEdges=mesh.sizeOfHalfedges()/2;
    	    	
		// store all the lines of the header (OFF format for colored meshes)
		String headerFile="off/header.off"; // file containing the header for the OFF format with colors
		ArrayList<String> header=new ArrayList();
		TC.lectureDansFichier(headerFile);
		while(TC.finEntree()==false) {
			header.add(TC.lireLigne());
		}

		// store all the lines of the end-of-file (OFF format for colored meshes)
		String endOfFile="off/end.off"; // file containing the header for the OFF format with colors
		ArrayList<String> eof=new ArrayList();
		TC.lectureDansFichier(endOfFile);
		while(TC.finEntree()==false) {
			eof.add(TC.lireLigne());
		}
		
		TC.lectureEntreeStandard();

    	System.out.print("Creating colored OFF file from 3D triangle mesh...");
    	Jcg.io.IO.writeNewTextFile(filename);
    	
    	// write the header to the new output file
    	for(String s: header)
    		TC.println(s);
    	
    	// file header
    	TC.println();
    	TC.println("(new-geometry \"[g1]\"	# polyhedron0");
    	TC.println("{ NOFF");
    	TC.println(nVertices+" "+nFaces+" "+nEdges);
    	    	
    	// writing vertex coordinates
    	int i=0;
    	for(Object o: mesh.vertices) { // iterate over all vertices
    		Vertex<Point_> v=(Vertex<Point_>)o;
    		v.index=i;
    		Point_ p=v.getPoint();
    		double x, y, z;
    		x=p.getCartesian(0).doubleValue();
    		y=p.getCartesian(1).doubleValue();
    		z=p.getCartesian(2).doubleValue();
    		
    		TC.println(""+x+" "+y+" "+z+" 1 0 0"); // Warning: vertex normals are not used (set by default fo '1 0 0'
    		i++;
    	}
    	
      	// write face/vertex incidence relations
    	for(Object o: mesh.facets){ // iterate over all faces
      		Face<Point_> f=(Face<Point_>)o;
    		int index0=f.getEdge().vertex.index;
    		int index1=f.getEdge().getNext().vertex.index;
    		int index2=f.getEdge().getNext().getNext().vertex.index;
    		String faceColor="";
    		if(f.tag>0)
    			faceColor=""+colors[f.tag];
    		Jcg.io.IO.println(3+" "+index0+" "+index1+" "+index2+" "+faceColor);
    		//System.out.println("f"+f.index+", tag "+f.tag);
    	}
    	
    	// write the last lines describing the polyhedron
    	TC.println("}");
    	TC.println(") # end geometry \"[g1]\" polyhedron0");
    	TC.println("(bbox-draw \"[g1]\" off)");
    	TC.println("(bbox-color \"[g1]\" 0.000000 0.000000 0.000000)");
    	TC.println("(normalization \"[g1]\" none)");
    	TC.println("(name-object \"[g1]\" \"polyhedron0\")");
    	
    	// write edge colors (if defined) to OFF file
    	for(Object o: mesh.halfedges){ // iterate over all halfedges
    		Halfedge e=(Halfedge)o;
    		int index0=e.getVertex().index;
    		int index1=e.getOpposite().getVertex().index;
    		Point_3 p=(Point_3)e.getVertex().getPoint();
    		Point_3 q=(Point_3)e.getOpposite().getVertex().getPoint();

    		int eIndex=e.index;
    		int edgeCounter=0;
    		String edgeColor=""+colors[e.tag];
    		if(e.tag!=0 && index0<index1) { // each edge is written only once. The color is not defined if 'tag=0'
    			TC.println("(new-geometry \"[g"+(edgeCounter+2)+"]\"	# Seg"+edgeCounter);
    			TC.println("{ appearance {");
    			TC.println("}");
    			TC.println("VECT");
    			TC.println("1 2 1");
    			TC.println("2");
    			TC.println();
    			TC.println("1");
    			TC.println();
    			TC.println(p.getX().doubleValue()+" "+p.getY().doubleValue()+" "+p.getZ().doubleValue());
    			TC.println(q.getX().doubleValue()+" "+q.getY().doubleValue()+" "+q.getZ().doubleValue());
    			TC.println();
    			TC.println(edgeColor);
    			TC.println("}");
    			TC.println(") # end geometry \"[g"+(edgeCounter+2)+"]\" Seg"+edgeCounter);
    			TC.println("(bbox-draw \"[g"+(edgeCounter+2)+"]\" off)");
    			TC.println("(bbox-color \"[g"+(edgeCounter+2)+"]\" 0.000000 0.000000 0.000000)");
    			TC.println("(normalization \"[g"+(edgeCounter+2)+"]\" none)");
    			TC.println("(name-object \"[g"+(edgeCounter+2)+"]\" \"Seg"+edgeCounter+"\")");
    			
    			edgeCounter++;
    		}
    	}
    	
    	// write the header to the new output file
    	TC.println();
    	for(String s: eof)
    		TC.println(s);
    	
    	Jcg.io.IO.writeStandardOutput();
    	System.out.println("done ("+filename+")");
    }

    /**
     * Output a 3D (surface) triangulation to OFF file, using as vertex locations the 3d coordinates of the nodes in the input 'graph'. <p>
     * <p>
     * Warning: <p>
     * -) the mesh and the graph are assumed to have the same number of vertices and the same combinatorics (not tested) <p>
     * -) the vertices (in the mesh) and the nodes (in the graph) are assumed to have the same ordering (not tested)
     * 
     * @param mesh  3D triangle mesh (Half-edge representation)
     * @param graph  the 1-skeleton graph underlying the mesh (vertex coordinates could be different from the ones in the mesh)
     * 
     * warning: only for triangle meshes (faces must be triangular)
     */
    public static void writePolyhedronToOFF(Polyhedron_3 mesh, AdjacencyListGraph graph, String filename, boolean withEdgeColors, boolean withFaceColors){
    	if(mesh.sizeOfVertices()!=graph.sizeVertices())
    		throw new Error("Error: the mesh and corresponding graph do tno have the same number of vertices");
    	
		Color[] colors=new Color[12];
		colors[0]=new Color(240, 240, 240, 255); // no color defined: gray by default
		colors[1]=Color.red;
		colors[2]=Color.blue;
		colors[3]=Color.black;
		colors[4]=Color.green;
		colors[5]=Color.orange;
		colors[6]=Color.green;
		colors[7]=Color.yellow;
		colors[8]=Color.magenta;
		colors[9]=new Color(240, 240, 240, 255); // standard color for rendering faces
		colors[10]=new Color(0, 128, 0, 255); // dark green
		colors[11]=new Color(100, 100, 100, 255); // dark gray

    	System.out.print("Creating OFF file from 1-skeleton (using node locations in the graph)...");
    	Jcg.io.IO.writeNewTextFile(filename);
    	
    	int nVertices=mesh.vertices.size();
    	int nFaces=mesh.facets.size();
    	int nEdges=mesh.sizeOfHalfedges()/2;
    	
    	int nElements=0;
    	if(withEdgeColors==true)
    		nElements=nFaces+nEdges;
    	else
    		nElements=nFaces;
    	
    	// file header
    	Jcg.io.IO.println("OFF");
    	Jcg.io.IO.println(nVertices+" "+nElements+" "+0);
    	    	
    	// writing vertex coordinates, using the locations of nodes in the graph
    	for(Node u: graph.vertices) { // iterate over all nodes in the graph
    		Point_3 p=u.getPoint();
    		double x, y, z;
    		x=p.getCartesian(0).doubleValue();
    		y=p.getCartesian(1).doubleValue();
    		z=p.getCartesian(2).doubleValue();
    		
    		Jcg.io.IO.println(""+x+" "+y+" "+z);
    	}
    	
      	// write face/vertex incidence relations
    	for(Object o: mesh.facets){ // iterate over all faces
      		Face<Point_> f=(Face<Point_>)o;
    		int index0=f.getEdge().vertex.index;
    		int index1=f.getEdge().getNext().vertex.index;
    		int index2=f.getEdge().getNext().getNext().vertex.index;
    		String faceColor="";
    		if(f.tag>=0 && withFaceColors==true)
    			faceColor=""+colors[f.tag].getRed()/255.+" "+colors[f.tag].getGreen()/255.+" "+colors[f.tag].getBlue()/255.+" "+colors[f.tag].getAlpha()/255.;
    		Jcg.io.IO.println(3+" "+index0+" "+index1+" "+index2+" "+faceColor);
    		//System.out.println("f"+f.index+", tag "+f.tag);
    	}
    	
    	if(withEdgeColors==true) { // write edge colors (if defined) to OFF file
        	for(Object o: mesh.halfedges){ // iterate over all halfedges
          		Halfedge e=(Halfedge)o;
        		int index0=e.getVertex().index;
        		int index1=e.getOpposite().getVertex().index;
        		
        		int eIndex=e.index;
        		String edgeColor=""+colors[e.tag].getRed()/255.+" "+colors[e.tag].getGreen()/255.+" "+colors[e.tag].getBlue()/255.+" "+colors[e.tag].getAlpha()/255.;
        		if(index0<index1) // each edge is written only once
        			Jcg.io.IO.println(2+" "+index0+" "+index1+" "+edgeColor);
        	}
    	}
    	
    	Jcg.io.IO.writeStandardOutput();
    	System.out.println("done ("+filename+")");
    }

    /**
     * Output the edges of a graph to a file (.edges format). <p>
     * <p>
     * @param graph  the 1-skeleton graph underlying the mesh (vertex coordinates could be different from the ones in the mesh)
     * 
     */
    public static void exportEdgesToFile(AdjacencyListGraph graph, String filename){    	
    	System.out.print("Writing a graph to file (edge list)...");
    	Jcg.io.IO.writeNewTextFile(filename);
    	
    	int nVertices=graph.sizeVertices();
    	int countEdges=0;
    	    	    	
    	// writing vertex coordinates, using the locations of nodes in the graph
    	for(Node u: graph.vertices) { // iterate over all nodes in the graph
    		List<Node> neighbors=u.neighbors;
    		int index1=u.index;
    		for(Node v: neighbors) {
    			int index2=v.index;
    			if(index1<index2) {
    				Jcg.io.IO.println(""+index1+" "+index2);
    				countEdges++;
    			}
    		}
    	}
    	
    	Jcg.io.IO.writeStandardOutput();
    	System.out.println("done ("+countEdges+" edges, "+nVertices+" vertices)");
    }

}
