package sd.util;

import java.util.LinkedList;

import Jcg.geometry.Point_3;
import Jcg.geometry.Vector_3;
import jdg.graph.AdjacencyListGraph;
import jdg.graph.Node;

public class Octree {
	public LinkedList<Node> l;  //List of vertices in the Octree if it is a leaf
	
	public Point_3 min;	//Point of minimum coordinates (lexicographic order) delimiting the Octree
	public Point_3 max;	//Point of maximal coordinates (lexicographic order) delimiting the Octree
	public Point_3 bary;	//barycentre of the Octree
	
	public Octree[] children; //table of Octrees splitting the vertices into sub-Octrees
	int depth;  //profondeur de l'Octree par rapport � la racine
	public int nb; // number of Nodes in the Octree(including its children)
	
	static double theta=1.2;
	static int maxDepth=8; //profondeur maximale de l'Octree
	
	/**
     *Build an Octree from a graph
     */	
	public Octree(AdjacencyListGraph g1){
		Point_3[] bounds=boundBox(g1); 
		min=bounds[0];
		max=findCube(min,bounds[1]);
		
		depth=0;
		
		for (Node n: g1.vertices){
			addNode(n);
		}
		
		//calcul du barycentre de chaque Octree ainsi que nb (nombre de Nodes contenus dans l'Octree)
		this.computeBary();
	}
	
	/**
     *Build an Octree from a list of vertices (Node), a bounding box (p1 and p2) and an int corresponding to the depth
     */
		
	Octree(Node n, Point_3 p1, Point_3 p2,int dept){
		min=p1;
		max=p2;
		depth=dept;
		addNode(n);
	}
	
	public double getTheta() {
		return theta;
	}
	
	/**
     * compute and return the bounding box for the points set.
     * The result is stored in a table.
     */	
	
	private Point_3[] boundBox(AdjacencyListGraph g2) { 
		double 	xmin=Double.MAX_VALUE, xmax=Double.MIN_VALUE, 
    			ymin=Double.MAX_VALUE, ymax=Double.MIN_VALUE,
    			zmin=Double.MAX_VALUE, zmax=Double.MIN_VALUE;
    	
    	double x, y, z;
    	for(Node u: g2.vertices) {
    		x=u.getPoint().getX().doubleValue();
    		y=u.getPoint().getY().doubleValue();
    		z=u.getPoint().getZ().doubleValue();
    		if (x<xmin)
    			xmin = x;
    		if (x>xmax)
    			xmax = x;
    		if (y<ymin)
    			ymin = y;
    		if (y>ymax)
    			ymax = y;
    		if (z<zmin)
    			zmin = z;
    		if (z>ymax)
    			zmax = z;
    	}
    	Point_3 p=new Point_3(xmin, ymin, zmin);
    	Point_3 q=new Point_3(xmax, ymax, zmax);
    	return new Point_3[]{p, q};
	}

	/**
     * compute and return the maximal point of a bounding cube containing the points p1 and p2 (p1 < p2).
     */

	private Point_3 findCube(Point_3 p1, Point_3 p2) {
		double dx=p2.x-p1.x;
		double dy=p2.y-p1.y;
		double dz=p2.z-p1.z;
		double max = Math.max(Math.max(dx,dy),dz);
		double m=1.01*max;
		return p1.sum(new Vector_3(m,m,m));
	}
	
	/**
     * compute recursively the barycenter of the Octrees
     */
	
	private void computeBary() {
		//if the Octree is a leaf, the barycenter is the barcycenter of the Nodes it contains
		if(isLeaf()){
			int nbNodes=l.size();
			double x=0;
			double y=0;
			double z=0;
			for(Node n:l){
				x+=n.p.x;
				y+=n.p.y;
				z+=n.p.z;
			}
			bary=new Point_3(x/nbNodes,y/nbNodes,z/nbNodes);
			nb=nbNodes;
			return;
		}
		
		//if it is not a leaf, then we compute the barycenter of its children to determine its own barycenter through a linear combination
		nb=0;
		LinkedList<WeightedPoint> wp=new LinkedList<WeightedPoint>();
		for (int i=0;i<8;i++){
			if (children[i]!=null){
				children[i].computeBary();
				nb+=children[i].nb;
				wp.add(new WeightedPoint(children[i].bary,children[i].nb));
			}
		}
		bary=WeightedPoint.linComb(wp);
	}
	

	/**
     * Method to add a Node to a current Octree
     */
	
	private void addNode(Node n) {
		//if depth==depth_max : we add n to l (or create a LinkedList only containing n)
		if(depth==maxDepth) {
			if(l==null){
				l=new LinkedList<Node>();
			}
			l.add(n);
			return;
		}
		//if is a leaf and l is not defined (it means that it is the first time this level of Octree is explored) : create a LinkedList only containing n
		if (isLeaf() && l==null){
			l=new LinkedList<Node>();
			l.add(n);
			return;
		}
		//if l already contains a Node (and since depth<depth_max l contains only one Node, otherwise children would have been created to divide the Nodes) 
			//-> we create children
		if(l!=null){
			children=new Octree[8];
			l.add(n);
			for (Node n2:l){
				int i = findChild(n2);  //determine in which child n2 belongs
				if(children[i]==null){
					Point_3[] boundingPoints=findBoundBox(i);
					children[i]=new Octree(n2,boundingPoints[0],boundingPoints[1],depth+1);
				} else {
					children[i].addNode(n2);
				}
				
			}
			l=null; 
			return;
		}
		
		//l=null and this is not a leaf so we add n to the child it belongs to
		int i = findChild(n);
		if(children[i]==null){
			Point_3[] boundingPoints=findBoundBox(i);
			children[i]=new Octree(n,boundingPoints[0],boundingPoints[1],depth+1);
			return;
		}
		children[i].addNode(n);	
	}
	
	/**
     * compute and return the min and max Points corresponding to the Bounding Box of the Octree this.children[i]
     */	
	
	private Point_3[] findBoundBox(int i) {
		double cote=max.x-min.x;
		int x=i/4;
		i-=4*(i/4);
		int y=i/2;
		i-=2*(i/2);
		int z=i/2;
		Point_3 Pmin=new Point_3(min.x+x*cote/2,min.y+y*cote/2,min.z+z*cote/2);
		Point_3 Pmax=new Point_3(min.x+(x+1)*cote/2,min.y+(y+1)*cote/2,min.z+(z+1)*cote/2);
		return new Point_3[]{Pmin,Pmax};
	}



	/**
     * compute and return the index of the Octree stored in children to which the Node n belongs to
     */	
	
	private int findChild(Node n) {
		double cote=max.x-min.x;
		int x=(n.p.x<min.x+cote/2)? 0 :1;
		int y=(n.p.y<min.y+cote/2)? 0 :1;
		int z=(n.p.z<min.z+cote/2)? 0 :1;
		return 4*x+2*y+z;
	}

	
	/**
     * return true if p is in the cube bounded by p1 and p2, false otherwise
     */
	
	public static boolean isInCube(Point_3 p, Point_3 p1, Point_3 p2){
		return (p.x<p2.x && p.x>=p1.x && p.y<p2.y && p.y>=p1.y && p.z<p2.z && p.z>=p1.z);
	}
	

	/**
	 * find and return the list of weighted points (corresponding to barycenters of Octrees and their weights) that are far enough of u
	 *  to be considered as a whole in the calculation of the repulsive force
	 *  It is a DFS
	 */
	
	public LinkedList<WeightedPoint> explore(Node u){
		LinkedList<WeightedPoint> clusters=new LinkedList<WeightedPoint>();
		//	if the octree is a leaf not corresponding to Node u then add this Octree (we cannot dig deeper)
		if(isLeaf()){
			// if u is not in the Octree
			if(!isInCube(u.p, min, max)){
				clusters.add(new WeightedPoint(bary,nb));
				return clusters;
			}
			// if u is in the Octree
			else{
				if(l.size()==1) return null;
				Point_3 newBary=new Point_3((nb*bary.x-u.p.x)/(nb-1),(nb*bary.y-u.p.y)/(nb-1),(nb*bary.z-u.p.z)/(nb-1));
				clusters.add(new WeightedPoint(newBary,nb-1));
				return clusters;
			}
			
		}
		double ds=max.x-min.x;
		double norm=Math.sqrt((double) bary.squareDistance(u.p));
		
		//check if the Octree is "far enough" of u
		if(ds/norm<=theta){
			clusters.add(new WeightedPoint(bary,nb));
			return clusters;
		}
		
		//if its not far enough then check the children
		LinkedList<WeightedPoint> aux;
		for(int i=0;i<8;i++){
			if(children[i]!=null){
				aux=children[i].explore(u);
				if(aux!=null){  //aux can be null even though children[i] is not (if children[i] the leaf containing u)
					clusters.addAll(aux);	
				}
			}
		}
		return clusters;
	}
	
	/**
     * return true if the octree has no child, false otherwise
     */
	
	public boolean isLeaf(){
		return (children==null);
	}
	
	private static class WeightedPoint{
		Point_3 p;
		int w;
		WeightedPoint(Point_3 p1,int weight){
			p=p1;
			w=weight;
		}
		
		/**
	     * compute and return the linear combination of the List of WeightedPoint WP.
	     */
		
		static Point_3 linComb(LinkedList<WeightedPoint> WP) {
			double x=0;
			double y=0;
			double z=0;
			int n=0;
			for(WeightedPoint wp:WP){
				x+=wp.p.getX().doubleValue() * wp.w;
				y+=wp.p.getY().doubleValue() * wp.w;
				z+=wp.p.getZ().doubleValue() * wp.w;
				n+=wp.w;
			}
			return new Point_3(x/n,y/n,z/n);
		}
	}
}
