package sd.util;

import Jcg.geometry.Point_3;
import Jcg.geometry.Vector_3;

/**
 * This class allows to compute the affine transformation between a plane rectangular frame and a 3D rectangular frame
 * 
 * @author Luca Castelli Aleardi (Ecole Polytechnique, 2017)
 *
 */
public class RectangleFrame {

	/** The three 3D points defining a rectangular frame in 3D (an hyperplane):<p>
	 * P is the left bottom corner, <p>
	 * Q is the left top corner, <p>
	 * R is the right top corner <p>*/
	public Point_3 P, Q, R;
	
	/** the two vectors (P-Q) and (R-Q)*/
	private Vector_3 vectorPQ, vectorQR;
	
	/** 
	 * The three 3D points defining a rectangular frame in 3D (an hyperplane):<p>
	 * <p>
	 * @param P  is the left bottom corner
	 * @param Q  is the left top corner
	 * @param R  is the right top corner
	 */
	public RectangleFrame(Point_3 P, Point_3 Q, Point_3 R) {
		this.P=P;
		//System.out.println("\nFrame : P"+P+", Q="+Q+", R="+R);
		this.vectorPQ=new Vector_3(P, Q);
		this.vectorQR=new Vector_3(Q, R);

		//System.out.println("Frame : P"+P+", Q="+Q+", R="+R);
		//System.out.println("\t PQ="+this.vectorPQ+", QR="+this.vectorQR);
	}
	
	/**
	 * Given an input point p=(s, t) lying the rectangle [0,1]x[0,h], it computes the affine transformation mapping
	 * p to its image on the 3D plane defined by P, Q, R
	 * 
	 * @param s  x-coordinate of the input point
	 * @param t  y-coordinate of the input point
	 * @param h  height of the rectangle
	 * 
	 * @return a 3D point lying on the 3D plane defined by the points P, Q, R
	 */	
	public Point_3 mapTo3DPlane(double s, double t, double h) {
		double x=P.getX().doubleValue()+(s*this.vectorQR.getX().doubleValue());
		double y=P.getY().doubleValue()+(s*this.vectorQR.getY().doubleValue());
		double z=P.getZ().doubleValue()+(s*this.vectorQR.getZ().doubleValue());
		
		double t1=t/h; // t1 belongs to [0, 1] 
		
		x=x+(t1*this.vectorPQ.getX().doubleValue());
		y=y+(t1*this.vectorPQ.getY().doubleValue());
		z=z+(t1*this.vectorPQ.getZ().doubleValue());
		
		return new Point_3(x, y, z);
	}

}
