package sd.util;

import java.util.LinkedList;
import Jama.*;

import Jcg.geometry.GeometricOperations_2;
import Jcg.geometry.GeometricOperations_3;
import Jcg.geometry.Point_;
import Jcg.geometry.Point_3;
import Jcg.geometry.Segment_3;
import Jcg.geometry.Triangle_3;
import Jcg.geometry.Vector_3;
import Jcg.polyhedron.Face;
import Jcg.polyhedron.Polyhedron_3;
import jdg.graph.AdjacencyListGraph;

/**
 * This class provides primitives for dealing with spherical geometry
 * 
 * @author Gaspard Denis (2017)
 *
 */
public class SphericalGeometry {
	
	/**
	 * Method which finds the geodesic distance between two points q and p of the sphere (radius = 1.0)
	 */
	public static double geodesicDistance(Point_3 p,Point_3 q){
		double long1=Math.atan2(p.y/Math.sqrt(p.x*p.x+p.y*p.y+p.z*p.z), p.x/Math.sqrt(p.x*p.x+p.y*p.y+p.z*p.z));
		double lat1=Math.PI/2.0-Math.acos(p.z/Math.sqrt(p.x*p.x+p.y*p.y+p.z*p.z));
		double long2=Math.atan2(q.y/Math.sqrt(q.x*q.x+q.y*q.y+q.z*q.z), q.x/Math.sqrt(q.x*q.x+q.y*q.y+q.z*q.z));
		double lat2=Math.PI/2.0-Math.acos(q.z/Math.sqrt(q.x*q.x+q.y*q.y+q.z*q.z));
        double dLon = long2 - long1;
        double res=Math.cos(lat1)*Math.cos(lat2)*Math.cos(dLon)+Math.sin(lat1)*Math.sin(lat2);
        double a=0.;
        if(res>=1.){
        	a=Math.acos(1.0);
        	return Math.min(2*Math.PI-a, a);
        }
        if(res<=-1.){
        	a=Math.acos(-1.0);
        	return Math.min(2*Math.PI-a, a);
        }
        a=Math.acos(res);
        return Math.min(2*Math.PI-a, a);
		
	}
	
	/**
	 * Check whether two geodesic arcs are crossings
	 */
	@Deprecated
	public static boolean ntersectSegments(Point_3 p1, Point_3 q1, Point_3 p2, Point_3 q2){
		throw new Error("Old implementation: not safe");
		// Hint: make use of the method: public static boolean doIntersect(Segment_3 s, Triangle_3 t)
		/*Triangle_3 t1=new Triangle_3(p1,q1,new Point_3(0.,0.,0.));
		Segment_3 s1=new Segment_3(p2,q2);
		
		Triangle_3 t2=new Triangle_3(p2,q2,new Point_3(0.,0.,0.));
		Segment_3 s2=new Segment_3(p1,q1);
		
		if (GeometricOperations_3.doIntersect(s1,t1) || GeometricOperations_3.doIntersect(s2,t2)){
			return true;
		}
		else{
			return false;
			
		}*/
	}
	
	/**
	 * Check whether two geodesic arcs are crossing
	 */
	public static boolean geodesicIntersection(Point_3 p1, Point_3 p2, Point_3 q1, Point_3 q2) {
		if(p1.equals(q1) || p2.equals(q2))
			return false;
		if(p1.equals(q2) || p2.equals(q1))
			return false;
		
		Point_3 origin=new Point_3(0., 0., 0.);
		
    	int orientation1=GeometricOperations_3.orientation(origin, p1 ,p2 ,q1);
    	int orientation2=GeometricOperations_3.orientation(origin, p1, p2, q2);
    	
    	if(orientation1*orientation2>=0) {// both points q1 and q2 belong to the same half-space delimited by the half-plane defined by (O, p1, p2)
    		//System.out.println("intersection: 1");
    		return false;
    	}
    	
    	int orientation3=GeometricOperations_3.orientation(origin, q1, q2, p1);
    	int orientation4=GeometricOperations_3.orientation(origin, q1, q2, p2);
    	
    	if(orientation3*orientation4>=0) { // both points p1 and p2 belong to the same half-space delimited by the half-plane containing q1 and q2
    		//System.out.println("Intersection: 2");
    		return false;
    	}
    	
    	System.out.println("crossing: ("+p1+", "+p2+") - ("+q1+", "+q2+")");
    	
    	if(p1.distanceFrom(q1).doubleValue()>1.5 && p2.distanceFrom(q2).doubleValue()>1.5)
    		return false;
    	
    	return true;
	}
	
	/**
	 * Method which takes two faces as parameters and that returns a boolean:
	 * true if both faces are situated in the same side of their common edge
	 * false if the two faces are situated in the two opposite side of their common edge
	 */
	public static boolean inSameSide(Face<Point_3> f1, Face<Point_3>f2, Polyhedron_3 mesh,AdjacencyListGraph g){
		LinkedList<Integer>commonIndices=new LinkedList<Integer>();
		
		LinkedList<Integer>l1=new LinkedList<Integer>();
		int[] t1=f1.getVertexIndices(mesh);
		for(int i=0;i<f1.degree();i++){
			l1.add(t1[i]);
		}

		LinkedList<Integer>l2=new LinkedList<Integer>();
		int[] t2=f2.getVertexIndices(mesh);
		for(int i=0;i<f2.degree();i++){
			l2.add(t2[i]);
		}
		
		int r1=0;
		int r2=0;
		for(int i=0;i<l1.size();i++){
			if(l2.contains(l1.get(i))){
				commonIndices.add(l1.get(i));
			}
			else{
				r1=l1.get(i);
			}
		}
		
		for(int i=0;i<l2.size();i++){
			if(!commonIndices.contains(l2.get(i))){
				r2=l2.get(i);
			}
		}
		Vector_3 v= new Vector_3(g.getNode(commonIndices.get(0)).getPoint(),g.getNode(commonIndices.get(1)).getPoint());
		Vector_3 v1=new Vector_3(g.getNode(commonIndices.get(1)).getPoint(),g.getNode(r1).getPoint());
		Vector_3 v2=new Vector_3(g.getNode(commonIndices.get(1)).getPoint(),g.getNode(r2).getPoint());
		
		return v.crossProduct(v1).innerProduct(v.crossProduct(v2)).doubleValue()>=0;
		
		
	}
	
	/**
	 * Method which takes two faces as parameters and that returns a boolean:
	 * true if the two faces are adjacent (ie they got a common edge)
	 * false if they are not adjacent (ie they have no common edge)
	 */
	public static boolean areAdjacent(Face<Point_3> f1, Face<Point_3>f2, Polyhedron_3 mesh){
		LinkedList<Integer>commonIndices=new LinkedList<Integer>();
		
		LinkedList<Integer>l1=new LinkedList<Integer>();
		int[] t1=f1.getVertexIndices(mesh);
		for(int i=0;i<f1.degree();i++){
			l1.add(t1[i]);
		}
		LinkedList<Integer>l2=new LinkedList<Integer>();
		int[] t2=f2.getVertexIndices(mesh);
		for(int i=0;i<f2.degree();i++){
			l2.add(t2[i]);
		}
		
		for(int i=0;i<l1.size();i++){
			if(l2.contains(l1.get(i))){
				commonIndices.add(l1.get(i));
			}
		}
		return commonIndices.size()==2;
	}
	
	/**
	 * Recursive method that returns a list of faces which are all situated in the same side of a graph
	 * (no fold)
	 * It acts as a spread in order to find all the faces in the same side.
	 */
	public static LinkedList<Face<Point_3>> facesinsameside(LinkedList<Face<Point_3>>l,Polyhedron_3 mesh,AdjacencyListGraph g){
		int length=l.size();
		for(int i=0;i<length;i++){
			for(int j=0;j<mesh.sizeOfFacets();j++){
				if(!l.contains(mesh.facets.get(j))){
					if(areAdjacent((Face<Point_3>)mesh.facets.get(j), l.get(i),  mesh)){ 
						if(!inSameSide((Face<Point_3>)mesh.facets.get(j),l.get(i),mesh,g)){
							l.add((Face<Point_3>)mesh.facets.get(j));
						}
					}
				}
			}
		}
		if(l.size()==length){
			return l;
		}
		else{
			return facesinsameside(l,mesh,g);
		}
	}
	
	
	/**
	 * Method that takes in argument a point p on the sphere and a vector_3d v and returns the symetric point
	 * compared to the plan orthogonal to v and including the origin.
	 */
	public static Point_3 computeSymetricPointtoPlane(Point_3 p, Vector_3 v){
		
		double norm=Math.sqrt(v.squaredLength().doubleValue());
		v=v.divisionByScalar(norm);
		
		Vector_3 vortho1=new Vector_3(v.y*v.y,-v.x*v.y-v.z,v.y);
		double norm1=Math.sqrt(vortho1.squaredLength().doubleValue());
		vortho1=vortho1.divisionByScalar(norm1);
		
		Vector_3 vortho2=v.crossProduct(vortho1);
		double norm2=Math.sqrt(vortho2.squaredLength().doubleValue());
		vortho2=vortho2.divisionByScalar(norm2);
		
		Vector_3 vp=new Vector_3(new Point_3(0.,0.,0.),p);
		double X=vp.innerProduct(vortho1).doubleValue();
		double Y=vp.innerProduct(vortho2).doubleValue();
		double Z=vp.innerProduct(v).doubleValue();
		
		Z=-Z;
		Point_3 q=new Point_3 (X,Y,Z);
		
		double[][]tab=new double[][]{{vortho1.x,vortho2.x,v.x},{vortho1.y,vortho2.y,v.y},{vortho1.z,vortho2.z,v.z}};
		Matrix mat=new Matrix(tab);
		double[][]vecq=new double[][]{{q.x},{q.y},{q.z}};
		Matrix res= mat.times(new Matrix(vecq));
		return new Point_3(res.get(0, 0),res.get(1,0),res.get(2,0));
	}
	
    /**
     * Method which computes the spherical angle between the vectors (V-U, W-U)
     */
    public static double sphericalAngle(Point_3 pU, Point_3 pV, Point_3 pW){
		Vector_3 ex=new Vector_3(1.,0.,0.);
		Vector_3 ey=new Vector_3(0.,1.,0.);
		Vector_3 ez=new Vector_3(0.,0.,1.);

		Vector_3[]tangentialVectors=SphericalGeometry.getTangentialVector(pU,0.01);
		Vector_3 v1=tangentialVectors[0];
		Vector_3 v2=tangentialVectors[1];
		
		Vector_3 deltaA=new Vector_3(pU,pV);
		deltaA=new Vector_3(v1.innerProduct(deltaA).doubleValue()*v1.innerProduct(ex).doubleValue()+v2.innerProduct(deltaA).doubleValue()*v2.innerProduct(ex).doubleValue(),deltaA.innerProduct(v1).doubleValue()*v1.innerProduct(ey).doubleValue()+deltaA.innerProduct(v2).doubleValue()*ey.innerProduct(v2).doubleValue(),deltaA.innerProduct(v1).doubleValue()*v1.innerProduct(ez).doubleValue()+deltaA.innerProduct(v2).doubleValue()*v2.innerProduct(ez).doubleValue());
		
		Vector_3 deltaB=new Vector_3(pU,pW);
		deltaB=new Vector_3(v1.innerProduct(deltaB).doubleValue()*v1.innerProduct(ex).doubleValue()+v2.innerProduct(deltaB).doubleValue()*v2.innerProduct(ex).doubleValue(),deltaB.innerProduct(v1).doubleValue()*v1.innerProduct(ey).doubleValue()+deltaB.innerProduct(v2).doubleValue()*ey.innerProduct(v2).doubleValue(),deltaB.innerProduct(v1).doubleValue()*v1.innerProduct(ez).doubleValue()+deltaB.innerProduct(v2).doubleValue()*v2.innerProduct(ez).doubleValue());
		
    	double normA=Math.sqrt(deltaA.squaredLength().doubleValue());
    	double normB=Math.sqrt(deltaB.squaredLength().doubleValue());
    	
    	double ip=deltaA.innerProduct(deltaB).doubleValue();

    	if(normA*normB==0.0)
    		System.err.println("Warning: division by zero (Spherical Angle)");
    	
    	double costheta=ip/(normA*normB);
    	if(costheta>1.0){
    		return 0.;
    	}
    	if(costheta<-1.0){
    		return Math.PI;
    	}
    	
    	double result=Math.acos(ip/(normA*normB));
    	if(Double.isInfinite(result))
    		System.err.println("Warning: infinite numbers");
    	if(Double.isNaN(result)) {
    		System.err.println("Warning: NaN number");
    		System.out.println("Ip= "+ip+", normA= "+normA+", normB= "+normB);
    	}
    	
    	return result;
    }
    
    /**
     * Method that computes the rotation matrix around the vector_3 u and of angle "angle"
     */
    public static Matrix rotationMatrix(Vector_3 u, double angle){
		double c=Math.cos(angle);
		double s=Math.sin(angle);
		double tab[][]=new double[][] {{u.x*u.x*(1.-c)+c,u.x*u.y*(1.-c)-u.z*s,u.x*u.z*(1.-c)+u.y*s},{u.x*u.y*(1.-c)+u.z*s,u.y*u.y*(1.-c)+c,u.y*u.z*(1.-c)-u.x*s},{u.x*u.z*(1.-c)-u.y*s,u.y*u.z*(1.-c)+u.x*s,u.z*u.z*(1.-c)+c}};		
		Matrix J=new Matrix(tab);
		return J;
    }
    
    /**
     * Method that takes in argument a point_3 p and that returns a array of two vector_3.
     * These two vectors are orthogonal and they generate the plan orthogonal to vector_3(origin,p).
     */
    public static Vector_3[] getTangentialVector(Point_3 p, double tolerance){
    	Vector_3[] TangentialVectors= new Vector_3 [2];
    	double[]coord=coord(p);
    	double theta=coord[0];
    	Vector_3 v1=new Vector_3();
    	Vector_3 v2=new Vector_3();
    	/*if(Math.abs(1-p.z)<tolerance){
    		v1=new Vector_3(1.0,0.,0.);
    		v2=new Vector_3(0.,1.0,0.);
    	}
    	else if(Math.abs(1+p.z)<tolerance){
    		v1=new Vector_3(-1.0,0.,0.);
    		v2=new Vector_3(0.,-1.,0.);
    	}*/

    	//else{
    		v1=new Vector_3(Math.cos(theta+Math.PI/2.0),Math.sin(theta+Math.PI/2.0),0.);
    		v1=v1.multiplyByScalar(1.0/Math.sqrt(v1.squaredLength().doubleValue()));
    		v2=v1.crossProduct(new Vector_3 (new Point_3(0.,0.,0.),p));
    		v2=v2.multiplyByScalar(1.0/Math.sqrt(v2.squaredLength().doubleValue()));
    	//}
		TangentialVectors[0]=v1;
		TangentialVectors[1]=v2;
		return TangentialVectors;
    }
    
    /**
     * Method that takes in argument a point p whose coordinates are cartesian coordinates and that
     * returns an array of two double: theta and phi, the two spherical coordinates of that point p 
     * (rho=1.0 because we are on the unit sphere).
     */
    public static double[] coord(Point_3 p){
		double[]result=new double[2];
		double norm=Math.sqrt(p.x*p.x+p.y*p.y+p.z*p.z);
		double norm2=Math.sqrt(p.x*p.x+p.y*p.y);
		
		double phi=Math.acos(p.z/norm);
		double theta=0.;
		if(p.y>=0.0){
			theta=Math.acos(p.x/norm2);
		}
		else{
			theta=2.0*Math.PI-Math.acos(p.x/norm2);
		}
		result[0]=theta;
		result[1]=phi;
		return result;
	}
    
    /**
     * Method that takes in argument a point p whose coordinates are cartesian coordinates and that
     * returns an array of two double: theta and phi, the two spherical coordinates of that point p 
     * (rho=1.0 because we are on the unit sphere).
     */
    public static double[] coordbisbis(Point_3 p, double tol){
		double[]result=new double[2];
		double norm=Math.sqrt(p.x*p.x+p.y*p.y+p.z*p.z);
		double norm2=Math.sqrt(p.x*p.x+p.y*p.y);
		
		if(norm2<tol){
			if(p.z>0.){
				result[0]=0.0;
				result[1]=0.0;
			}
			else{
				result[0]=0.;
				result[1]=Math.PI;
			}
			return result;
		}
		else{
			double phi=Math.acos(p.z/norm);
			double theta=0.;
			if(p.y>=0.0){
				theta=Math.acos(p.x/norm2);
			}
			else{
				theta=2.0*Math.PI-Math.acos(p.x/norm2);
			}
			result[0]=theta;
			result[1]=phi;
			return result;
		}
	}
    
    /**
     * Method that takes in argument a point p whose coordinates are cartesian coordinates and that
     * returns an array of two double: theta and phi, the two spherical coordinates of that point p 
     * (rho=1.0 because we are on the unit sphere).
     */
    public static double[] coordbis(Point_3 p, double tolcoord){
		double[]result=new double[2];
		
		if(Math.abs(1.0-p.z)<tolcoord){
			result[0]=0.;
			result[1]=0.;
			return result;
		}
		else if(Math.abs(-1.0-p.z)<tolcoord){
			result[0]=0.;
			result[1]=Math.PI;
			return result;
		}
		
		if(Math.abs(1.0-p.y)<tolcoord){
			result[0]=Math.PI/2.0;
			result[1]=Math.PI/2.0;
			return result;
		}
		else if(Math.abs(-1.0-p.y)<tolcoord){
			result[0]=3.0*Math.PI/2.0;
			result[1]=Math.PI/2.0;
			return result;
		}
		
		if(Math.abs(1.0-p.x)<tolcoord){
			result[0]=0.;
			result[1]=Math.PI/2.0;
			return result;
		}
		else if(Math.abs(-1.0-p.x)<tolcoord){
			result[0]=Math.PI;
			result[1]=Math.PI/2.0;
			return result;
		}
		
		double norm=Math.sqrt(p.x*p.x+p.y*p.y+p.z*p.z);
		double norm2=Math.sqrt(p.x*p.x+p.y*p.y);
		double phi=Math.acos(p.z/norm);
		double theta=0.;
		if(p.y>=0.0){
			theta=Math.acos(p.x/norm2);
		}
		else{
			theta=2.0*Math.PI-Math.acos(p.x/norm2);
		}
		result[0]=theta;
		result[1]=phi;
		return result;
	}
  
}
