package sd.viewer;


import java.awt.Color;
import java.math.BigDecimal;
import java.util.ArrayList;
import java.util.LinkedList;
import java.util.List;
import java.util.Random;

import Jcg.geometry.Point_2;
import Jcg.geometry.Point_3;
import jdg.graph.AdjacencyListGraph;
import jdg.graph.Node;
import sd.layout.SphericalLayout;
import tc.TC;

/**
 * Class for exporting the spherical layout of a graph into a file, encoded in Ipe format (.ipe)
 * 
 * @author Luca Castelli Aleardi (Ecole Polytechnique, august 2017)
 */
public class ExportLayoutToIpe {
	// constants defining the 3D bounding box containing the projections of all points on the unit sphere (using the Processing perspective projection)
	// these constants are valid only for a given value of the scale factor: they can be computed using the method 'getSphericalFactor()'
	private float xMin=402.272f, yMin=202.277f, zMin=0.9218f;
	private float xMax=797.785f, yMax=597.779f, zMax=0.9598f;
	private final float width=10.f; // width of the drawing area (for the Ipe drawing)
	public float scalingFactor; // scaling factor for resizing the drawing (in the Ipe code)
	
	private int resolution=15;
	private int prec=2;
	
	private String[] colors=new String[] {"darkgray", "red", "blue", "black", "green", "orange", "lightcyan", "yellow", "darkmagenta", "gray", "darkgreen", "darkgray"};
	
	public SurfaceMeshRendering meshRendering; // 3d render (for meshes and spherical layouts)
	
	/** distance threshold: draw/write only edges whose distance from the '3d camera' is below a given threshold */
	public double threshold;
	
	/** 
	 * For exporting 3D graph layouts to IPE format
	 **/
	public ExportLayoutToIpe(SurfaceMeshRendering meshRendering) {
		this.meshRendering=meshRendering;
		this.threshold=zMin+(zMax-zMin)/2.f;
		//this.threshold=zMax; // for showing all geodesics (even those who are 'hidden' behind the sphere surface
		this.scalingFactor=0.5f;
	}

	/** 
	 * For exporting 2D graph layouts to IPE format
	 **/
	public ExportLayoutToIpe() {
		this.meshRendering=null;
		this.threshold=0;
		this.xMin=0;
		this.xMax=0;
		this.yMin=0;
		this.yMax=0;
		//this.threshold=zMax; // for showing all geodesics (even those who are 'hidden' behind the sphere surface
		this.scalingFactor=10;
	}

    /**
     * Export the graph layout to Ipe format
     * <p>
     * Remark: <p>
     * vertices must be endowed with geometric coordinates. Vertex colors are allowed
     * 
     */		   
    public void export(AdjacencyListGraph g, float[][] colors, String filename, String layoutType) {
    	System.out.print("Exporting graph layout to file in Ipe format ("+filename+")...");
    	this.meshRendering.view.noLoop(); // stop the loop in the main drawing method
    	this.getSphericalFactor();
    	
    	TC.ecritureDansNouveauFichier(filename);
    	this.writeHeader("ipe/sample.ipe"); // this file contains a basic template for the encoding of graphic objects in IPE format

    	if(layoutType.equals("spherical layout")) {
    		System.out.print("[spherical layout]");
    		this.writeGeodesicEdges(g);
    	}
    	else if(layoutType.equals("3D layout")) {
    		System.out.print("[euclidean layout]");
    		this.writeStraightEdges(g);
    	}
    	else {
    		System.err.println("Warning: layout type not supported: only 'spherical layout' and '3D layout' are supported");
    	}
    	//this.writeVertices(g, colors);

    	this.writeEnd();
    	TC.ecritureSortieStandard();
    	this.meshRendering.view.loop(); // restart the drawing loop
    	System.out.println("done");
    }
    
	  /**
	   * Write a colored 3D straight-line edge segment (u, v) to file
	   */
    private void writeEdge(Node u, Node v, int r, int g, int b) {
    	float s=(float)this.meshRendering.scaleFactor;

    	String color="black";
    	String line="<path layer=\"alpha\" stroke=\""+color+"\">";
    	TC.println(line);

    	float x1=this.meshRendering.view.screenX((float)u.getPoint().getX().doubleValue()*s, (float)u.getPoint().getY().doubleValue()*s, (float)u.getPoint().getZ().doubleValue()*s);
    	float y1=this.meshRendering.view.screenY((float)u.getPoint().getX().doubleValue()*s, (float)u.getPoint().getY().doubleValue()*s, (float)u.getPoint().getZ().doubleValue()*s);

    	double coordX=this.roundNumber(x1, this.prec);
    	double coordY=this.roundNumber(y1, this.prec);
    	line=""+coordX+" "+coordY+" m";
    	TC.println(line);

    	float x2=this.meshRendering.view.screenX((float)v.getPoint().getX().doubleValue()*s, (float)v.getPoint().getY().doubleValue()*s, (float)v.getPoint().getZ().doubleValue()*s);
    	float y2=this.meshRendering.view.screenY((float)v.getPoint().getX().doubleValue()*s, (float)v.getPoint().getY().doubleValue()*s, (float)v.getPoint().getZ().doubleValue()*s);
    	coordX=this.roundNumber(x2, this.prec);
    	coordY=this.roundNumber(y2, this.prec);
    	line=""+coordX+" "+coordY+" l";
    	TC.println(line);
    	TC.println("</path>");
    }
	  
		/**
		 * Draw a geodesic arc (p, q) on the unit sphere <p>
		 * 
		 * Geodesic arcs are represented by polylines: edges are subdivided into step+1 segments
		 * 
		 * @param threshold  distance threshold: draw/write only edges whose distance from the '3d camera' is below a given threshold  
		 */
		private void writeGeodesicArcOld(Point_3 p, Point_3 q) {
			float s=(float)this.meshRendering.scaleFactor;
			int prec=6; // numeric precision (for rounding vertex coordinates)
			
			float x1=(float)p.getX().doubleValue()*s;
			float y1=(float)p.getY().doubleValue()*s;
			float z1=(float)p.getZ().doubleValue()*s;
			float x2=(float)q.getX().doubleValue()*s;
			float y2=(float)q.getY().doubleValue()*s;
			float z2=(float)q.getZ().doubleValue()*s;
			//System.out.println("("+x1+", "+y1+", "+z1+") - "+" ("+x2+", "+y2+", "+z2+" )");
			
			Point_3[] vertices=new Point_3[resolution+1];
			
			vertices[0]=SphericalLayout.linearInterpolation(p, q, 0.); // coincides with p
			for(int i=0;i<this.resolution;i++) { // adding subdivision points
				double t=1./(this.resolution+1);
				vertices[i+1]=SphericalLayout.projectOnSphere(SphericalLayout.linearInterpolation(p, q, (i+1)*t));
			}
			vertices[this.resolution]=SphericalLayout.linearInterpolation(p, q, 1.); // coincides with q

			// draw all the pieces of the polyline
			for(int i=0;i<this.resolution;i++) {
				x1=this.meshRendering.view.screenX((float)vertices[i].getX().doubleValue()*s, (float)vertices[i].getY().doubleValue()*s, (float)vertices[i].getZ().doubleValue()*s);
				y1=this.meshRendering.view.screenY((float)vertices[i].getX().doubleValue()*s, (float)vertices[i].getY().doubleValue()*s, (float)vertices[i].getZ().doubleValue()*s);
				z1=this.meshRendering.view.screenZ((float)vertices[i].getX().doubleValue()*s, (float)vertices[i].getY().doubleValue()*s, (float)vertices[i].getZ().doubleValue()*s);
				
				x2=this.meshRendering.view.screenX((float)vertices[i+1].getX().doubleValue()*s, (float)vertices[i+1].getY().doubleValue()*s, (float)vertices[i+1].getZ().doubleValue()*s);
				y2=this.meshRendering.view.screenY((float)vertices[i+1].getX().doubleValue()*s, (float)vertices[i+1].getY().doubleValue()*s, (float)vertices[i+1].getZ().doubleValue()*s);
				z2=this.meshRendering.view.screenZ((float)vertices[i+1].getX().doubleValue()*s, (float)vertices[i+1].getY().doubleValue()*s, (float)vertices[i+1].getZ().doubleValue()*s);
				
				if(x1<0 || x2<0 || y1<0 || y2<0 || z1<0 || z2<0) {
					System.err.println("Warning: wrong spherical coordinates");
					return;
				}
				// draw/write the segment if and only if its two endpoints are closer (to the 3D camera) than a given threshold
				if(z1<threshold && z2<this.threshold) {
					this.writeSegment(x1, y1, x2, y2, 0, 0, 0);
				}
			}

		}
		
		private void updateResolution(Point_3 p, Point_3 q) {
			float s=(float)this.meshRendering.scaleFactor;
			
			float x1=this.meshRendering.view.screenX((float)p.getX().doubleValue()*s, (float)p.getY().doubleValue()*s, (float)p.getZ().doubleValue()*s);
			float y1=this.meshRendering.view.screenY((float)p.getX().doubleValue()*s, (float)p.getY().doubleValue()*s, (float)p.getZ().doubleValue()*s);
			float x2=this.meshRendering.view.screenX((float)q.getX().doubleValue()*s, (float)q.getY().doubleValue()*s, (float)q.getZ().doubleValue()*s);
			float y2=this.meshRendering.view.screenY((float)q.getX().doubleValue()*s, (float)q.getY().doubleValue()*s, (float)q.getZ().doubleValue()*s);
			
			float dist=(float)Math.sqrt((x1-x2)*(x1-x2)+(y1-y2)*(y1-y2)); // euclidean length of the edge (p, q)
			
			/*if(dist<2) {
				this.resolution=5;
				return;
			}
			else if(dist<10) {
				this.resolution=6;
				return;
			}
			else if(this.resolution<15)
				this.resolution=8;
			else
				this.resolution=15;
			*/	
			
		}

		/**
		 * Draw a geodesic arc (p, q) on the unit sphere <p>
		 * 
		 * Geodesic arcs are represented by polylines: edges are subdivided into step+1 segments
		 * 
		 * @param threshold  distance threshold: draw/write only edges whose distance from the '3d camera' is below a given threshold  
		 */
		private void writeGeodesicArc(Point_3 p, Point_3 q) {
			float s=(float)this.meshRendering.scaleFactor;
			
			float x1=(float)p.getX().doubleValue()*s;
			float y1=(float)p.getY().doubleValue()*s;
			float z1=(float)p.getZ().doubleValue()*s;
			float x2=(float)q.getX().doubleValue()*s;
			float y2=(float)q.getY().doubleValue()*s;
			float z2=(float)q.getZ().doubleValue()*s;
			//System.out.println("("+x1+", "+y1+", "+z1+") - "+" ("+x2+", "+y2+", "+z2+" )");
			
			Point_3[] vertices=new Point_3[resolution+1];
			
			vertices[0]=SphericalLayout.linearInterpolation(p, q, 0.); // coincides with p
			for(int i=0;i<this.resolution;i++) { // adding subdivision points
				double t=1./(this.resolution+1);
				vertices[i+1]=SphericalLayout.projectOnSphere(SphericalLayout.linearInterpolation(p, q, (i+1)*t));
			}
			vertices[this.resolution]=SphericalLayout.linearInterpolation(p, q, 1.); // coincides with q

			LinkedList<Double> coordX=new LinkedList<Double>();
			LinkedList<Double> coordY=new LinkedList<Double>();
			
			// draw all the pieces of the polyline
			for(int i=0;i<=this.resolution;i++) {
				x1=this.meshRendering.view.screenX((float)vertices[i].getX().doubleValue()*s, (float)vertices[i].getY().doubleValue()*s, (float)vertices[i].getZ().doubleValue()*s);
				y1=this.meshRendering.view.screenY((float)vertices[i].getX().doubleValue()*s, (float)vertices[i].getY().doubleValue()*s, (float)vertices[i].getZ().doubleValue()*s);
				z1=this.meshRendering.view.screenZ((float)vertices[i].getX().doubleValue()*s, (float)vertices[i].getY().doubleValue()*s, (float)vertices[i].getZ().doubleValue()*s);
				if(x1<0 || x1<0 || y1<0) {
					System.err.println("Warning: wrong spherical coordinates");
					return;
				}
				// draw/write the segment if and only if its two endpoints are closer (to the 3D camera) than a given threshold
				if(z1<threshold) {
					coordX.add((double)x1);
					coordY.add((double)y1);
					//this.writeSegment(x1, y1, x2, y2, 0, 0, 0);
				}
			}
			
			double[] x=new double[coordX.size()];
			int count=0;
			for(Double valueX: coordX) {
				x[count]=valueX;
				count++;
			}				
			double[] y=new double[coordY.size()];
			count=0;
			for(Double valueY: coordY) {
				y[count]=valueY;
				count++;
			}				
				
			this.updateResolution(p, q); // update the resolution to adapt the drawing to the scale
			//this.writePolyline(x, y, 0, 0, 0, 0.5);
			this.writeCurve(x, y, 0, 0, 0, this.scalingFactor);
		}

	  /**
	   * Write a colored 2D straight-line segment from P1 to P2
	   */
	  private void writeSegment(double x1, double y1, double x2, double y2, int r, int g, int b) {
		  String color="black";
		  String line="<path layer=\"alpha\" stroke=\""+color+"\">";
		  TC.println(line);
		  
		  //x1=this.roundNumber((x1-this.xMin)*this.scalingFactor, prec);
		  //y1=this.roundNumber((y1-this.yMin)*this.scalingFactor, prec);
		  x1=this.roundNumber((x1-this.xMin)/2., prec);
		  y1=this.roundNumber((y1-this.yMin)/2., prec);
		  
		  line=""+x1+" "+y1+" m"; // first endpoint
		  TC.println(line);

		  //x2=this.roundNumber((x2-this.xMin)*this.scalingFactor, prec);
		  //y2=this.roundNumber((y2-this.yMin)*this.scalingFactor, prec);
		  x2=this.roundNumber((x2-this.xMin)/2.f, prec);
		  y2=this.roundNumber((y2-this.yMin)/2.f, prec);

		  line=""+x2+" "+y2+" l"; // second endpoint
		  TC.println(line);
		  TC.println("</path>");
	  }

	  /**
	   * Write a colored 2D straight-line segment from P1 to P2
	   */
	  private void writeSegment(double x1, double y1, double x2, double y2, int c) {
		  String color=this.colors[c];
		  String line="<path layer=\"alpha\" stroke=\""+color+"\">";
		  TC.println(line);
		  
		  //x1=this.roundNumber((x1-this.xMin)*this.scalingFactor, prec);
		  //y1=this.roundNumber((y1-this.yMin)*this.scalingFactor, prec);
		  x1=this.roundNumber((x1-this.xMin)/2., prec);
		  y1=this.roundNumber((y1-this.yMin)/2., prec);
		  
		  line=""+x1+" "+y1+" m"; // first endpoint
		  TC.println(line);

		  //x2=this.roundNumber((x2-this.xMin)*this.scalingFactor, prec);
		  //y2=this.roundNumber((y2-this.yMin)*this.scalingFactor, prec);
		  x2=this.roundNumber((x2-this.xMin)/2.f, prec);
		  y2=this.roundNumber((y2-this.yMin)/2.f, prec);

		  line=""+x2+" "+y2+" l"; // second endpoint
		  TC.println(line);
		  TC.println("</path>");
	  }

	  /**
	   * Write a colored 2D polyline
	   */
	  private void writePolyline(double x[], double y[], int r, int g, int b, double scale) {
		  if(x==null || y==null || x.length<=1 || y.length<=1)
			  return;
		  String color;
		  if(r==0 && g==0 && b==0)
			  color="black";
		  else {
			  System.err.println("Warning: arbitrary colors not supported");
			  return;
		  }
		  String line="<path layer=\"alpha\" stroke=\""+color+"\">";
		  TC.println(line);
		  
		  int prec=4;
		  double coordX, coordY;

		  // write first endpoint
		  coordX=this.roundNumber((x[0]-this.xMin)*scale, prec);
		  coordY=this.roundNumber((y[0]-this.yMin)*scale, prec);
		  line=""+coordX+" "+coordY+" m"; // first endpoint
		  TC.println(line);

		  for(int i=1;i<x.length;i++) {
			  coordX=this.roundNumber((x[i]-this.xMin)*scale, prec);
			  coordY=this.roundNumber((y[i]-this.yMin)*scale, prec);
			  line=""+coordX+" "+coordY+" l"; // second endpoint
			  TC.println(line);
		  }		  
		  
		  TC.println("</path>");
	  }

	  /**
	   * Write a colored 2D curve
	   */
	  private void writeCurve(double x[], double y[], int r, int g, int b, double scale) {
		  if(x==null || y==null || x.length==0 || y.length==0)
			  return;
		  String color;
		  if(r==0 && g==0 && b==0)
			  color="black";
		  else {
			  System.err.println("Warning: arbitrary colors not supported");
			  return;
		  }
		  String line="<path layer=\"alpha\" stroke=\""+color+"\">";
		  TC.println(line);
		  
		  int prec=2;
		  double coordX, coordY;

		  // write first endpoint
		  coordX=this.roundNumber((x[0]-this.xMin)*scale, prec);
		  coordY=this.roundNumber((y[0]-this.yMin)*scale, prec);
		  line=""+coordX+" "+coordY+" m"; // first endpoint
		  TC.println(line);

		  for(int i=1;i<x.length-1;i++) {
			  coordX=this.roundNumber((x[i]-this.xMin)*scale, prec);
			  coordY=this.roundNumber((y[i]-this.yMin)*scale, prec);
			  line=""+coordX+" "+coordY; // second endpoint
			  TC.println(line);
		  }		  
		  coordX=this.roundNumber((x[x.length-1]-this.xMin)*scale, prec);
		  coordY=this.roundNumber((y[x.length-1]-this.yMin)*scale, prec);
		  line=""+coordX+" "+coordY+" c"; // second endpoint
		  TC.println(line);
		  
		  TC.println("</path>");
	  }

	  /**
	   * Write a colored edge (u, v) to file <p>
	   * 
	   * Colors are encoded in rgb format: three float values in [0..255]
	   * 
	   * @param u  vertex to write
	   * @param r  red component of the vertex color
	   */
	  private void writeVertex(Node u, float r, float g, float b) {
		  double x=u.getPoint().getX().doubleValue();
		  double y=u.getPoint().getY().doubleValue();
		  String pos=x+" "+y;
		  
		  String color=round((r/255),3)+" "+round((g/255),3)+" "+round((b/255),3);
		  
		  String line="<use name=\"mark/disk(sx)\" pos=\""+pos+"\" size=\"large\" stroke=\""+color+"\"/>";
		  TC.println(line);
	  }

	  /**
	   * Write the entire graph layout to file (only edges)
	   */
	  private void writeStraightEdges(AdjacencyListGraph graph) {
		  if(graph==null) // if the graph is not defined exit
			  return;
		  
			for(Node u: graph.vertices) { // write the edges of g
				for(Node v: u.neighbors) {
					if(v!=null && v.index>u.index) { // write only directed edges (u, v) such that u<v
						this.writeEdge(u, v, 0, 0, 0); // write a black edge
					}
				}
			}
	  }

	  /**
	   * Write a list of edge segments to a file in IPE format
	   */
	  public void exportEdgesToFile(List<Point_2[]> edges, String filename) {
		  if(edges==null) // if the graph is not defined exit
			  return;

		  System.out.print("Exporting 2D edge segments to file in Ipe format ("+filename+")...");

		  TC.ecritureDansNouveauFichier(filename);
		  writeHeader("ipe/sample.ipe"); // this file contains a basic template for the encoding of graphic objects in IPE format
		  //this.writeVertices(g, colors);

		  double s=this.scalingFactor;
		  for(Point_2[] segment: edges) {
			  if(segment!=null) {
				  double x1=segment[0].getX().doubleValue();
				  double y1=segment[0].getY().doubleValue();
				  double x2=segment[1].getX().doubleValue();
				  double y2=segment[1].getY().doubleValue();
				  writeSegment(x1*s, y1*s, x2*s, y2*s, 0, 0, 0); // write a black edge
			  }
		  }

		  this.writeEnd();
		  TC.ecritureSortieStandard();
		  System.out.println("done");

	  }

	  /**
	   * Write a list of edge segments to a file in IPE format
	   */
	  public void exportEdgesToFile(List<Point_2[]> edges, ArrayList<Integer> colors, String filename) {
		  if(edges==null) // if the graph is not defined exit
			  return;

		  System.out.print("Exporting 2D edge segments to file in Ipe format ("+filename+")...");

		  TC.ecritureDansNouveauFichier(filename);
		  writeHeader("ipe/sample.ipe"); // this file contains a basic template for the encoding of graphic objects in IPE format
		  //this.writeVertices(g, colors);

		  double s=this.scalingFactor;
		  int counter=0;
		  for(Point_2[] segment: edges) {
			  if(segment!=null) {
				  double x1=segment[0].getX().doubleValue();
				  double y1=segment[0].getY().doubleValue();
				  double x2=segment[1].getX().doubleValue();
				  double y2=segment[1].getY().doubleValue();
				  int c=colors.get(counter);
				  writeSegment(x1*s, y1*s, x2*s, y2*s, c); // write a black edge
				  counter++;
			  }
		  }

		  this.writeEnd();
		  TC.ecritureSortieStandard();
		  System.out.println("done");

	  }

	  /**
	   * Write the entire graph layout to file (only edges)
	   */
	  private void writeGeodesicEdges(AdjacencyListGraph graph) {
		  if(graph==null) // if the graph is not defined exit
			  return;
		  
			for(Node u: graph.vertices) { // write the edges of g
				for(Node v: u.neighbors) {
					if(v!=null && v.index>u.index) { // write only directed edges (u, v) such that u<v
						Point_3 p=u.getPoint();
						Point_3 q=v.getPoint();
						this.writeGeodesicArc(p, q);
					}
				}
			}
	  }

	  /**
	   * Write the entire graph layout to file (only edges)
	   */
	  private void writeVertices(AdjacencyListGraph graph, float[][] colors) {
		  if(graph==null) // if the graph is not defined exit
			  return;
		  
		  int i=0;
		  for(Node u: graph.vertices) { // write the vertices of g
			  this.writeVertex(u, colors[i][0], colors[i][1], colors[i][2]);
			  i++;
		  }
			  
	  }

	  /**
	   * Get the bounding box containing the projection of all points on the sphere
	   */
	  private float[] getSphericalFactor() {
		  Random generator = new Random(1);
		  int n=20000; // sample points
		  float[] result=new float[3];
		  float xmin=Float.MAX_VALUE, ymin=Float.MAX_VALUE, zmin=Float.MAX_VALUE;
		  float xmax=Float.MIN_VALUE, ymax=Float.MIN_VALUE, zmax=Float.MIN_VALUE;

		  float s=(float)this.meshRendering.scaleFactor;
		  Point_3 p; 
		  double w1=1./2., h1=1./2., d1=1./2.;
		  for(int i=0;i<n;i++){
			  double n1=w1-2*w1*generator.nextDouble();
			  double n2=h1-2*h1*generator.nextDouble();
			  double n3=d1-2*d1*generator.nextDouble();
			  p=SphericalLayout.projectOnSphere(new Point_3 (n1, n2, n3));
			  
			  float x=this.meshRendering.view.screenX((float)p.getX().doubleValue()*s, (float)p.getY().doubleValue()*s, (float)p.getZ().doubleValue()*s);
			  float y=this.meshRendering.view.screenY((float)p.getX().doubleValue()*s, (float)p.getY().doubleValue()*s, (float)p.getZ().doubleValue()*s);
			  float z=this.meshRendering.view.screenZ((float)p.getX().doubleValue()*s, (float)p.getY().doubleValue()*s, (float)p.getZ().doubleValue()*s);
			  
			  xmin=Math.min(xmin, x);
			  ymin=Math.min(ymin, y);
			  zmin=Math.min(zmin, z);
			  xmax=Math.max(xmax, x);
			  ymax=Math.max(ymax, y);
			  zmax=Math.max(zmax, z);
		  }		
		  
		  System.out.print("("+xmin+", "+ymin+", "+zmin+")");
		  System.out.println(" - ("+xmax+", "+ymax+", "+zmax+")");
		  
		  result[0]=xmin;
		  result[1]=ymin;
		  result[2]=zmin;
		  return result;
	  }

    /**
     * Write the header lines (stored in the input file)
     * 
     */		   
    private static void writeHeader(String inputFile) {
    	TC.lectureDansFichier(inputFile);
    	
    	while(TC.finEntree()==false) {
    		String line=TC.lireLigne();
    		if(line.contains("</page>")==false && line.contains("</ipe>")==false) // write all lines 1.. n-2
    			TC.println(line);
    	}
    	
    	TC.lectureEntreeStandard();
    }

    /**
     * Write the last lines
     * 
     */		   
    private void writeEnd() {
    	TC.println("</page>");
    	TC.println("</ipe>");
    }
    
    private static BigDecimal round(float d, int decimalPlace) {
        BigDecimal bd = new BigDecimal(Float.toString(d));
        bd = bd.setScale(decimalPlace, BigDecimal.ROUND_HALF_UP);       
        return bd;
    }
    
	/** Return the value after truncation */
	public static double roundNumber(double x, int prec) {
		double precision=Math.pow(10, prec);
		return ((int)(x*precision)/((double)precision));
	}

}
