package sd.viewer;
import java.util.LinkedList;

import Jcg.geometry.Point_3;
import Jcg.mesh.MeshLoader;
import Jcg.polyhedron.Face;
import Jcg.polyhedron.Halfedge;
import Jcg.polyhedron.Polyhedron_3;
import Jcg.polyhedron.Vertex;
import jdg.graph.AdjacencyListGraph;
import jdg.io.GraphReader;
import jdg.io.GraphReader_Edges;
import jdg.io.GraphReader_MTX;
import jdg.io.GraphReader_OFF;
import sd.graphalgorithms.DisjointPaths;
import sd.graphalgorithms.EdgePath;
import sd.graphalgorithms.RectangularCanonicalLabeling;
import sd.graphalgorithms.BalancedRectangularCanonicalLabeling;
import sd.graphalgorithms.RiverComputation;
import sd.graphalgorithms.VertexPath;
import sd.layout2D.RectangularFPPLayout;
import sd.util.MeshAlgorithms;

/**
 * A class storing all information concerning the input mesh and corresponding graph
 * 
 * @author Luca Castelli Aleardi (INF562, 2017)
 *
 */
public class MeshRepresentation {
	
	// input parameters
	static double tolerance; // numeric tolerance (for convergence)
	static double threshold; // numeric threshold (for convergence of iterative spherical Tutte methods)
	static int vertexSeed; // vertex seed for the computation of shortests paths
	static int southFaceIndex, northFaceIndex; // indices of the south and north pole faces
	static int minDistance; // parameters for the computation of the separating cycle
	static int maxIterations; // parameters for the computation of the separating cycle
	static String linearSolver; // choice of the solver for lienar systems (Jama, MTJ, ParallelColt)
	static int FPPsteps=0, FPPavoid=0; // input parameters for the computation of the canonical labeling
	static boolean useEqualAngles=false; // say whether to use 'equal' angles of 'adapted' angles (for the prism)
	static String inputFile; // name of the file storing the input graph/mesh
	
    /** input graph to draw */
    public AdjacencyListGraph graph;
    
    /** input 3D mesh to draw */
    public Polyhedron_3<Point_3> mesh;

    /** north and south 'pole' faces (a pair of farthest faces in the mesh) */
    Face<Point_3> northFace=null, southFace=null;

    /** north and south 'pole' vertices (a pair of farthest vertices in the mesh) */
    public Vertex<Point_3> northPole=null, southPole=null;

    public MeshRepresentation(AdjacencyListGraph graph, Polyhedron_3<Point_3> mesh) {
    	this.graph=graph;
    	this.mesh=mesh;
    }

    public MeshRepresentation(String filename) {
    	this.inputFile=filename;
		GraphReader reader=null;

		if(filename.endsWith(".txt") || filename.endsWith(".edges"))
			reader=new GraphReader_Edges();
		else if(filename.endsWith(".off") || filename.endsWith(".OFF")) {
    		mesh=MeshLoader.getSurfaceMesh(filename);
    		mesh.isValid(false); // check combinatorial validity

			reader=new GraphReader_OFF();
		}
		else if(filename.endsWith(".mtx"))
			reader=new GraphReader_MTX();
		else {
			System.out.println("Error: input format not supported");
			System.out.println("Supported formats are: edge list, OFF, MTX");
			System.exit(0);
		}
		
		if(reader==null) {
			System.err.println("Error: input files cannot be loaded");
			System.exit(0);
		}
		
		this.graph=reader.read(filename); // read input network from file
    }
    
    public void reloadInputGraph() {
    	String filename=inputFile;
		GraphReader reader=null;

		if(filename.endsWith(".txt") || filename.endsWith(".edges"))
			reader=new GraphReader_Edges();
		else if(filename.endsWith(".off") || filename.endsWith(".OFF")) {
    		mesh=MeshLoader.getSurfaceMesh(filename);
    		mesh.isValid(false); // check combinatorial validity

			reader=new GraphReader_OFF();
		}
		else if(filename.endsWith(".mtx"))
			reader=new GraphReader_MTX();
		else {
			System.out.println("Error: input format not supported");
			System.out.println("Supported formats are: edge list, OFF, MTX");
			System.exit(0);
		}
		
		if(reader==null) {
			System.err.println("Error: input files cannot be loaded");
			System.exit(0);
		}
		
		this.graph=reader.read(filename); // read input network from file
    }

	/**
	 * Get two farthest faces (south and north poles)
	 */	
	public void computeFarthestFaces(int vertexIndex) {
		// choose the seed vertices (a pair of vertices, far from each other)
		System.out.println("Seed vertex: v"+vertexIndex);
		this.southPole=this.mesh.vertices.get(vertexIndex);
		this.northPole=MeshAlgorithms.getFarthestVertex(this.mesh, this.southPole); // farthest vertex from v0
		
		// retrieve two faces incident to v0 and v1
		this.southFace=southPole.getHalfedge().getFace();
		this.northFace=northPole.getHalfedge().getFace();
		System.out.println("Two farthest faces computed: f"+this.southFace.index+", f"+this.northFace.index);
	}

	/**
	 * Get two farthest faces (south and north pole faces)
	 */	
	public void computePoleFaces(int face1, int face2) {
		this.southFace=this.mesh.facets.get(face1);
		this.northFace=this.mesh.facets.get(face2);
		System.out.println("Setting the two pole faces: f"+this.southFace.index+", f"+this.northFace.index);
	}

}
