package sd.viewer;
import java.awt.Color;
import java.util.ArrayList;
import java.util.LinkedList;
import java.util.List;

import Jcg.geometry.*;
import Jcg.polyhedron.*;
import jdg.graph.AdjacencyListGraph;
import jdg.graph.Node;
import processing.core.PApplet;
import sd.layout.SphericalLayout;


/**
 * Class for rendering a 3D mesh (a 3d polyhedral surface) using Processing 
 * 
 * @author Luca Castelli Aleardi (INF421, 2016)
 *
 */
public class SurfaceMeshRendering {
	
	// different layout modes
	public static final int render3DMesh=0;
	public static final int renderSphericalLayout=1;
	public static final int renderEuclidean3DLayout=2;
	
	/** choose between: <p> 3d mesh rendering, spherical layout, euclidean 3d layout */
	public int layoutRenderingType;
	
	public boolean pause=false;
	
	double scaleFactor=60; // scaling factor: useful for 3d rendering
	public double zoom=1.; // for scaling the drawing
	private int geodesicArcResolution=100; // for drawing geodesic arcs using polylines
	private int renderType=1; // choice of type of rendering
	private int layoutType=0; // choice of the Layout algorithm

	public double sphereSize; // scale factor for drawing vertices
	public Color[] colors; // for drawing faces, edges and vertices

	public PApplet view; // Processing 3d frame (where meshes are rendered)

	/** store all information concerning the input graph and input mesh*/
	static public MeshRepresentation input;
	
	/** select a vertex (on a 3d mesh) on the screen (using mouse click) */
	public Vertex<Point_3> selectedVertex=null;

	/** select a vertex (in a graph) on the screen (using mouse click) */
	public Node selectedNode=null;

	/**
	 * Create a surface mesh from an OFF file
	 */	
	public SurfaceMeshRendering(PApplet view, MeshRepresentation input) {
		this.colors=new Color[12];
		this.colors[0]=null; // no color defined
		this.colors[1]=Color.red;
		this.colors[2]=Color.blue;
		this.colors[3]=Color.black;
		this.colors[4]=Color.green;
		this.colors[5]=Color.orange;
		this.colors[6]=Color.cyan;
		this.colors[7]=Color.yellow;
		this.colors[8]=Color.magenta;
		this.colors[9]=new Color(240, 240, 240, 255); // standard color for rendering faces
		this.colors[10]=new Color(0, 128, 0, 255); // dark green
		this.colors[11]=new Color(100, 100, 100, 255); // dark gray
		
		this.view=view;

    	this.input=input; // input 3d mesh (half-edge representation)
    	
    	if(this.input.mesh!=null) {
    		double averageEdgeLength=0;
    		for(Halfedge e: input.mesh.halfedges) {
    			double dist=Math.sqrt(e.getVertex().getPoint().squareDistance(e.getOpposite().getVertex().getPoint()).doubleValue());
    			//minEdgeLength=Math.min(dist, minEdgeLength);
    			averageEdgeLength=averageEdgeLength+dist;
    		}
    		averageEdgeLength=averageEdgeLength/(input.mesh.sizeOfHalfedges()/2);
    		sphereSize=(averageEdgeLength)/6;
    	
    		//System.out.println(polyhedron3D.verticesToString());   	
    		//System.out.println(polyhedron3D.facesToString());
    		input.mesh.isValid(false);
    	    	
    		this.scaleFactor=this.computeScaleFactor(); // compute the scale factor for the 3d mesh
    		resetMeshColors(input.mesh);
    	}
	}
	
	/** Draw the 3d mesh */	
	public void show3DMesh() {
		this.layoutRenderingType=SurfaceMeshRendering.render3DMesh;
		this.updateScaleFactor();
	}

	/** Draw the spherical layout of the graph (on the unit sphere) */	
	public void showSphericalLayout() {
		this.layoutRenderingType=SurfaceMeshRendering.renderSphericalLayout;
		this.updateScaleFactor();
	}

	/** Draw the euclidean layout of the graph in 3D */	
	public void showEuclidean3DLayout() {
		this.layoutRenderingType=SurfaceMeshRendering.renderEuclidean3DLayout;
		this.updateScaleFactor();
	}
	
	/** Update the scaleFactor, to perform a zoom in/out of the 3d scene */	
	public void zoom(double factor) {
		this.scaleFactor=this.scaleFactor*factor;
	}

	/**
	 * Set the initial tag of all faces, edges and vertices to a given value (used for color rendering)
	 */	
	public static void resetMeshColors(Polyhedron_3 mesh) {
		for(Vertex v: (List<Vertex>)mesh.vertices)
			v.tag=0;
		for(Face f: (List<Face>)mesh.facets) {
			f.tag=0; // standard face color
		}
		for(Halfedge e: (List<Halfedge>)mesh.halfedges) {
			e.tag=0;
		}
	}

	/**
	 * Draw a segment between two points
	 */	
	/*public void drawSegment(Point_3 p, Point_3 q) {
		float s=(float)this.scaleFactor*(float)this.zoom;
		float x1=(float)p.getX().doubleValue()*s;
		float y1=(float)p.getY().doubleValue()*s;
		float z1=(float)p.getZ().doubleValue()*s;
		float x2=(float)q.getX().doubleValue()*s;
		float y2=(float)q.getY().doubleValue()*s;
		float z2=(float)q.getZ().doubleValue()*s;
		this.view.line(	x1, y1, z1, x2, y2, z2 );		
	}*/
	
	/**
	 * Draw a vertex (as a small sphere)
	 */	
	public void drawVertex(Vertex<Point_3> v) {
		if(v==null )
			return;
		
		if(this.colors[v.tag]==null)
			this.drawVertex(v, Color.lightGray); // standard vertices are drawn in gray
		else
			this.drawVertex(v, this.colors[v.tag]); // draw the vertex with a given color
	}

	/**
	 * Draw a colored vertex (as a small sphere)
	 */	
	public void drawVertex(Vertex<Point_3> v, Color c) {
		if(v==null || c==null)
			return;
		
		Point_3 p=v.getPoint();
		float s=(float)this.scaleFactor;
		float x1=(float)p.getX().doubleValue()*s;
		float y1=(float)p.getY().doubleValue()*s;
		float z1=(float)p.getZ().doubleValue()*s;
		
		view.fill((float)c.getRed(), (float)c.getGreen(), (float)c.getBlue());
		view.translate(x1, y1, z1);
		view.sphere((float)(this.sphereSize*s));
		view.translate(-x1, -y1, -z1);
	}

	/**
	 * Draw a colored edge
	 */	
	public void drawEdge(Halfedge<Point_3> e, Color c) {
		if(e==null || c==null)
			return;
		
		Point_3 p=e.getVertex().getPoint();
		Point_3 q=e.getOpposite().getVertex().getPoint();
		
		float s=(float)this.scaleFactor*(float)this.zoom;
		float x1=(float)p.getX().doubleValue()*s;
		float y1=(float)p.getY().doubleValue()*s;
		float z1=(float)p.getZ().doubleValue()*s;
		float x2=(float)q.getX().doubleValue()*s;
		float y2=(float)q.getY().doubleValue()*s;
		float z2=(float)q.getZ().doubleValue()*s;
		
		view.stroke((float)c.getRed(), (float)c.getGreen(), (float)c.getBlue());
		this.view.line(	x1, y1, z1, x2, y2, z2 );		
	}

	/**
	 * Draw a colored edge
	 */	
	public void drawEdge(Halfedge<Point_3> e) {
		if(e==null )
			return;
		
		if(this.colors[e.tag]==null)
			this.drawEdge(e, Color.lightGray); // standard edges are drawn in gray
		else
			this.drawEdge(e, this.colors[e.tag]); // draw the edge with a given color
	}

	/**
	 * Draw a triangle
	 */	
	public void drawTriangle(Point_3 p, Point_3 q, Point_3 r) {
		float s=(float)this.scaleFactor*(float)this.zoom;
		view.vertex( (float)(p.getX().doubleValue()*s), (float)(p.getY().doubleValue()*s), (float)(p.getZ().doubleValue()*s));
		view.vertex( (float)(q.getX().doubleValue()*s), (float)(q.getY().doubleValue()*s), (float)(q.getZ().doubleValue()*s));
		view.vertex( (float)(r.getX().doubleValue()*s), (float)(r.getY().doubleValue()*s), (float)(r.getZ().doubleValue()*s));
	}

	/**
	 * Draw a (triangle or polygonal) face
	 */	
	public void drawFace(Face<Point_3> f) {
		if(f==null )
			return;
		
		if(this.colors[f.tag]==null)
			this.drawFace(f, this.colors[9]); // color for standard faces 
		else
			this.drawFace(f, this.colors[f.tag]); // draw the faces with a given color
	}

	/**
	 * Draw a (triangle or polygonal) face with a given color
	 */	
	public void drawFace(Face<Point_3> f, Color c) {
		Halfedge<Point_3> h=f.getEdge();
		Halfedge<Point_3> pEdge=h.getNext();
		
		Point_3 u=h.getOpposite().getVertex().getPoint();
		view.noStroke();
		view.fill((float)c.getRed(), (float)c.getGreen(), (float)c.getBlue()); // color of the triangle
		
		while(pEdge.getVertex()!=h.getOpposite().getVertex()) {
			Point_3 v=pEdge.getOpposite().getVertex().getPoint();
			Point_3 w=pEdge.getVertex().getPoint();
			this.drawTriangle(u, v, w); // draw a triangle face
			
			pEdge=pEdge.getNext();
		}
	}

	/**
	 * Assign a color to a given collection of vertices
	 */	
	public void colorVertices(List<Vertex> coloredVertices, int colorTag) {
		if(coloredVertices==null)
			return;
		
		for(Vertex v: coloredVertices)
			v.tag=colorTag; // assign to each vertex a given color
	}

	/**
	 * Assign a color to a subset in a collection of vertices
	 */	
	public void colorVertices(List<Vertex<Point_3>> coloredVertices, boolean[] marked, int colorTag) {
		if(coloredVertices==null)
			return;
		
		for(Vertex v: coloredVertices) {
			if(marked[v.index]==true)
				v.tag=colorTag; // assign to each vertex a given color
		}
	}

	/**
	 * Assign a color to a given collection of vertices
	 */	
	public void colorVertex(Vertex vertex, int colorTag) {
		if(vertex==null)
			return;
		
		vertex.tag=colorTag; // assign to the vertex a given color
	}

	/**
	 * Assign a color to a given collection of edges <p>
	 * Remark: each edge and its corresponding halfedge must have the same color
	 */	
	public void colorEdges(List<Halfedge> coloredEdges, int colorTag) {
		if(coloredEdges==null)
			return;
		
		for(Halfedge e: coloredEdges) {
			e.tag=colorTag;
			e.getOpposite().tag=colorTag; // color the opposite halfedge
		}
	}
	
	/**
	 * Draw a geodesic arc (p, q) on the unit sphere <p>
	 * 
	 * Geodesic arcs are represented by polylines: edges are subdivided into step+1 segments
	 * 
	 * @param size  size of the subdivision (number of edges defining the polyline)
	 */
	private void drawGeodesic(Point_3 p, Point_3 q, int size) {
		Point_3[] vertices=new Point_3[size+1];
		
		vertices[0]=SphericalLayout.linearInterpolation(p, q, 0.); // coincides with p
		for(int i=0;i<size;i++) { // adding subdivision points
			double t=1./(size+1);
			vertices[i+1]=SphericalLayout.projectOnSphere(SphericalLayout.linearInterpolation(p, q, (i+1)*t));
		}
		vertices[size]=SphericalLayout.linearInterpolation(p, q, 1.); // coincides with q

		view.strokeWeight(2);
		
		float s=(float)this.scaleFactor;
		//float maxDistance=0.9415f;
		// draw all the pieces of the polyline
		for(int i=0;i<size;i++) {
			float x1=(float)vertices[i].getX().doubleValue()*s;
			float y1=(float)vertices[i].getY().doubleValue()*s;
			float z1=(float)vertices[i].getZ().doubleValue()*s;
			
			/*float depth1=this.view.screenZ(x1, y1, z1);
			if(depth1<maxDistance)
				this.drawSegment(vertices[i], vertices[i+1]);*/
			
			this.drawSegment(vertices[i], vertices[i+1]);
		}
	}
	
	/**
	 * Draw a vertex (as a small sphere)
	 */	
	public void drawVertex(Point_3 p) {
		float s=(float)this.scaleFactor*(float)this.zoom;
		float x1=(float)p.getX().doubleValue()*s;
		float y1=(float)p.getY().doubleValue()*s;
		float z1=(float)p.getZ().doubleValue()*s;
		
		this.view.fill(250f, 0f, 0f);
		view.strokeWeight(1);
		this.view.translate(x1, y1, z1);
		this.view.sphere(s/40f);
		this.view.translate(-x1, -y1, -z1);
	}

	/**
	 * Draw a vertex (as a small sphere)
	 */	
	/*public void drawVertexOnSphere(Point_3 p) {
		System.out.println("Drawing vertex on sphere");
		float s=(float)this.scaleFactor;
		float x1=(float)p.getX().doubleValue()*s;
		float y1=(float)p.getY().doubleValue()*s;
		float z1=(float)p.getZ().doubleValue()*s;
		
		view.strokeWeight(1);
		this.view.translate(x1, y1, z1);
		this.view.sphere(0.1f);
		this.view.translate(-x1, -y1, -z1);
	}*/

	/**
	 * Draw a straight-line segment between two points in 3D
	 */	
	public void drawSegment(Point_3 p, Point_3 q) {
		float s=(float)this.scaleFactor;
		float x1=(float)p.getX().doubleValue()*s;
		float y1=(float)p.getY().doubleValue()*s;
		float z1=(float)p.getZ().doubleValue()*s;
		float x2=(float)q.getX().doubleValue()*s;
		float y2=(float)q.getY().doubleValue()*s;
		float z2=(float)q.getZ().doubleValue()*s;
		this.view.line(	x1, y1, z1, x2, y2, z2 );		
	}

	/**
	 * Assign a color to a given collection of faces
	 */	
	public void colorFaces(List<Face> coloredFaces, int colorTag) {
		if(coloredFaces==null)
			return;
		
		for(Face f: coloredFaces) {
			f.tag=colorTag; // assign to each face a given color
		}
	}

	/**
	 * Draw the entire mesh or graph
	 * 
	 */
	public void draw(int type) {
		// -------------------------------------------------------------
		// ---- draw the spherical or euclidean layout of the graph ----
		// -------------------------------------------------------------
		if(layoutRenderingType==renderSphericalLayout || layoutRenderingType==renderEuclidean3DLayout) {
			this.view.strokeWeight(1); // set the line width (for edges)
			this.view.stroke(20);
			if(this.input==null || this.input.graph==null)
				return;
			
			for(Node u: this.input.graph.vertices) { // draw all arcs
				Point_3 p=u.getPoint();
				for(Node v: u.neighbors) {
					if(v!=null && v.index>u.index) { // draw only directed edges (u, v) such that u<v
						Point_3 q=v.getPoint();
						if(layoutRenderingType==renderEuclidean3DLayout)
							this.drawSegment(p, q); // draw edge (p,q) as a straight line segment in 3D
						else
							this.drawGeodesic(p, q, this.geodesicArcResolution); // draw the edge (p, q) as a geodesic (on the unit sphere)
					}
				}
			}
			//view.strokeWeight(1);
			
			if(this.renderType==0) { // draw
				this.view.noStroke();
				this.view.fill(0f, 0f, 250f);
				for(Node v: this.input.graph.vertices) {
					this.drawVertex(v.getPoint());
				}
			}
			
			if(layoutRenderingType==renderSphericalLayout) { // draw an unit sphere (only when drawing a spherical layout)
				float r=0.99999999f; // radius of the sphere
				this.view.fill((float)0, (float)204, (float)204, (float)255);
				this.view.noStroke();
				this.view.sphere((float)(r*this.scaleFactor));
				
				if(this.selectedNode!=null) { // draw the selected vertex on the unit sphere, if any
					Point_3 p=this.selectedNode.getPoint();
					this.drawVertex(p);
				}

			}
		}
		// -----------------------------------------------
		// ------------- draw the 3D mesh ----------------
		// -----------------------------------------------
		else { // draw the 3d mesh
			// draw all faces
			view.beginShape(view.TRIANGLES);
			for(Face<Point_3> f: this.input.mesh.facets) {
				this.drawFace(f);
			}
			view.endShape();

			view.strokeWeight(2); // line width (for edges)
			if(type!=1) { // draw all edges
				view.stroke(20);
				//if(pause==false) {
					for(Halfedge<Point_3> e: this.input.mesh.halfedges)
						this.drawEdge(e);
				//}
			}
			//view.strokeWeight(1);

			if(type!=0) { // no rendering for vertices
				view.noStroke();
				if(this.pause==false) {
					for(Vertex<Point_3> v: this.input.mesh.vertices)
						this.drawVertex(v);
				}
			}

			//this.drawColoredVertices(view.coloredVertices, view.vertexColors);
			//this.drawColoredEdges(view.coloredEdges, view.edgeColors);

			view.strokeWeight(2);
			
			if(this.selectedVertex!=null) {
				Point_3 p=this.selectedVertex.getPoint();
				this.drawVertex(p);
			}
		}
	}
	
	/**
	 * Return the value after truncation (with a given numeric precision)
	 */
	private static double round(double x, int precision) {
		return ((int)(x*precision)/(double)precision);
	}
	
	/**
	 * Compute the scale factor (depending on the max distance of the point set in the original 3d mesh)
	 */
	private double computeScaleFactor() {
		if(this.input.mesh==null || this.input.mesh.vertices.size()<1)
			return 1;
		double maxDistance=0.;
		Point_3 origin=new Point_3(0., 0., 0.);
		for(Vertex<Point_3> v: this.input.mesh.vertices) {
			double distance=Math.sqrt(v.getPoint().squareDistance(origin).doubleValue());
			maxDistance=Math.max(maxDistance, distance);
		}
		return Math.sqrt(3)/maxDistance*150;
	}
	
	/**
	 * Update the scale factor
	 */
	public void updateScaleFactor() {
		if(this.layoutRenderingType==renderSphericalLayout || this.layoutRenderingType==renderEuclidean3DLayout)
			this.scaleFactor=300; // only for the spherical drawing of the graph (on the unit sphere)
		else
			this.scaleFactor=this.computeScaleFactor(); // for drawing the 3d mesh
	}
	
	  /**
	   * Get the 2D coordinates of the projection of point 'p' on the screen
	   */
	  public float[] getPointOn2DScreen(Point_3 p) {			  
		  double x=p.getX().doubleValue();
		  double y=p.getY().doubleValue();
		  double z=p.getZ().doubleValue();

		  float[] q=new float[2]; // coordinates in the 2D frame
		  q[0]=this.view.screenX((float)x, (float)y, (float)z);
		  q[1]=this.view.screenY((float)x, (float)y, (float)z);

		  return q;
	  }

	  /**
	   * Select the graph node whose 2d projection is the closest to pixel (i, j)
	   */
	  private Node selectNodeOnScreen(AdjacencyListGraph g, float i, float j) {			  
		  Node result=null;
		  float s=(float)this.scaleFactor*(float)this.zoom;
		  
		  double minDist=10.;
		  for(Node u: g.vertices) { // iterate over the vertices of g
			  Point_3 p=u.getPoint(); // 3D coordinates of the node (on the unit sphere)
			  double x=p.getX().doubleValue()*s;
			  double y=p.getY().doubleValue()*s;
			  double z=p.getZ().doubleValue()*s;
			  
			  float[] q=new float[2]; // coordinates in the 2D frame
			  q[0]=this.view.screenX((float)x, (float)y, (float)z);
			  q[1]=this.view.screenY((float)x, (float)y, (float)z);
			  
			  double dist=Math.sqrt((q[0]-i)*(q[0]-i)+(q[1]-j)*(q[1]-j));
			  if(dist<minDist) {
				  minDist=dist;
				  result=u;
			  }
		  }
		  
		  if(result!=null) {
			  //System.out.print("p3d: "+result.getPoint());
			  //float[] q=this.getPointOn2DScreen(this.selectedVertex.getPoint());
			  //System.out.println("p2d: "+q[0]+","+q[1]);
			  double x=result.getPoint().getX().doubleValue()*s;
			  double y=result.getPoint().getY().doubleValue()*s;
			  double z=result.getPoint().getZ().doubleValue()*s;
			  
			  float[] q=new float[3]; // coordinates in the 2D frame
			  q[0]=this.view.screenX((float)x, (float)y, (float)z);
			  q[1]=this.view.screenY((float)x, (float)y, (float)z);
			  q[2]=this.view.screenZ((float)x, (float)y, (float)z);
			  System.out.println("Selected vertex on mesh: v"+result.index+" - ("+q[0]+", "+q[1]+", "+q[2]+")");
		  }
		  else
			  System.out.println("Selected vertex on graph: none");

		  
		  return result;
	  }

	  /**
	   * Select the vertex whose 2d projection is the closest to pixel (i, j)
	   */
	  private Vertex<Point_3> selectVertexOnScreen(Polyhedron_3<Point_3> mesh, float i, float j) {			  
		  Vertex<Point_3> result=null;
		  float s=(float)this.scaleFactor*(float)this.zoom;
		  
		  //System.out.print("mouse location: "+i+", "+j);
		  
		  double minDist=10.;
		  for(Vertex<Point_3> u: mesh.vertices) { // iterate over the vertices of the mesh
			  Point_3 p=u.getPoint(); // 3D coordinates of the node
			  double x=p.getX().doubleValue()*s;
			  double y=p.getY().doubleValue()*s;
			  double z=p.getZ().doubleValue()*s;
			  
			  float[] q=new float[2]; // coordinates in the 2D frame
			  q[0]=this.view.screenX((float)x, (float)y, (float)z);
			  q[1]=this.view.screenY((float)x, (float)y, (float)z);
			  
			  double dist=Math.sqrt((q[0]-i)*(q[0]-i)+(q[1]-j)*(q[1]-j));
			  //System.out.println("\t v"+u.index+" "+p+"\t ("+q[0]+", "+q[1]+") - \t dist="+dist); // for debugging
			  
			  if(dist<minDist) {
				  minDist=dist;
				  result=u;
			  }
			  
		  }
		  
		  if(result!=null) {
			  //System.out.print("p3d: "+result.getPoint());
			  //float[] q=this.getPointOn2DScreen(this.selectedVertex.getPoint());
			  //System.out.println("p2d: "+q[0]+","+q[1]);
			  
			  double x=result.getPoint().getX().doubleValue()*s;
			  double y=result.getPoint().getY().doubleValue()*s;
			  double z=result.getPoint().getZ().doubleValue()*s;
			  
			  float[] q=new float[3]; // coordinates in the 2D frame
			  q[0]=this.view.screenX((float)x, (float)y, (float)z);
			  q[1]=this.view.screenY((float)x, (float)y, (float)z);
			  q[2]=this.view.screenZ((float)x, (float)y, (float)z);
			  System.out.println("Selected vertex on mesh: v"+result.index+" - ("+q[0]+", "+q[1]+", "+q[2]+"), f"+result.getHalfedge().getFace().index);
		  }
		  else
			  System.out.println("Selected vertex on mesh: none");
		  
		  return result;
	  }

	  /**
	   * Select the vertex or node whose 2d projection is the closest to pixel (i, j)
	   */
	  public void selectVertexOnScreen(float i, float j) {
		  if(this.layoutRenderingType==this.render3DMesh) {
			  this.selectedVertex=this.selectVertexOnScreen(this.input.mesh, i, j);
			  this.selectedNode=null;
		  }
		  else {
			  this.selectedVertex=null;
			  this.selectedNode=this.selectNodeOnScreen(this.input.graph, i, j);			  
		  }
	  }

	  /**
	   * Draw a label on the 2D canvas
	   */
	  public void drawLabel(Point_3 p, String label) {
		  float[] q=this.getPointOn2DScreen(p);
		  float i=q[0];
		  float j=q[1];

		  this.view.stroke(200, 200, 200);
		  this.view.fill(200, 200, 200);
		  this.view.rect(i, j, 150, 85); // fill a gray rectangle
		  this.view.fill(0);
		  this.view.text(label, i+5, j+14); // draw the vertex label
	  }

}
